#ifndef MRC_H_OCT_07_2008
#define MRC_H_OCT_07_2008

#include <cfloat>
#include <fstream>
#include <cstring>
#include "util.h"
#include "timer.h"
#include "relation.h"
#include "clusterpair.h"
#include "intdouble.h"
#include "move.h"

typedef RBTree<Move*,MoveComp> MergeTree;
typedef MergeTree::RBNode MoveNode;

const double DEFAULT_FALSE_PROB = 0.99999;


class MRC
{
 public:
  MRC(const char* const& paramsFile)
    : inDir_(""),            outDir_(""),    infoFile_(""),  dbFile_(""),
      betaTrue_(-1),         betaTotal_(-1), lambda_(-1),    loglike_(1.0),
      bestClustFile_(""),    combsFile_(""), print_(false),  iter_(0),
      mergeOccurred_(false), maxIter_(-1),   checkSymValid_(false),
      minSymAppear_(-1),     maxShared_(-1), threshRPerPair_(-1),
      llThreshold_(-1),      printUnit_(false),  saveUnit_(false),
      defaultFalseProb_(DEFAULT_FALSE_PROB), pureGreedy_(false)
  { readParams(paramsFile); init(); }

  MRC()
    : inDir_(""),            outDir_(""),    infoFile_(""), dbFile_(""),
      betaTrue_(-1),         betaTotal_(-1), lambda_(-1),   loglike_(1.0),
      bestClustFile_(""),    combsFile_(""), print_(false), iter_(0),
      mergeOccurred_(false), maxIter_(-1),   checkSymValid_(false),
      minSymAppear_(-1),     maxShared_(-1), threshRPerPair_(-1),
      llThreshold_(-1),      printUnit_(false),  saveUnit_(false),
      defaultFalseProb_(DEFAULT_FALSE_PROB), pureGreedy_(false) {}

  MRC(const string& inDir,          const string& outDir,
      const string& infoFile,       const string& dbFile,
      const double& betaTrue,       const double& betaTotal,
      const double& lambda,         const string& bestClustFile,
      const string& combsFile,      const int&    maxIter,
      const bool&   checkSymValid,  const int&    minSymAppear,
      const int&    maxShared,      const double& threshRPerPair,
      const double& llThreshold,    const bool&   printUnit,
      const bool&   saveUnit,       const bool&   pureGreedy)
    : inDir_(inDir),   outDir_(outDir),     infoFile_(infoFile),
      dbFile_(dbFile), betaTrue_(betaTrue), betaTotal_(betaTotal),
      lambda_(lambda), bestClustFile_(bestClustFile), combsFile_(combsFile),
      print_(false), iter_(0), mergeOccurred_(false), maxIter_(maxIter),
      checkSymValid_(checkSymValid), minSymAppear_(minSymAppear),
      maxShared_(maxShared), threshRPerPair_(threshRPerPair),
      llThreshold_(llThreshold), printUnit_(printUnit), saveUnit_(saveUnit),
      defaultFalseProb_(DEFAULT_FALSE_PROB), pureGreedy_(pureGreedy)
      { init(); }

  ~MRC() { types_.deleteItemsAndClear(); relations_.deleteItemsAndClear(); }

  string inDir()           const { return inDir_;           }
  string outDir()          const { return outDir_;          }
  string infoFile()        const { return infoFile_;        }
  string dbFile()          const { return dbFile_;          }
  double betaTrue()        const { return betaTrue_;        }
  double betaTotal()       const { return betaTotal_;       }
  double lambda()          const { return lambda_;          }
  string bestClustFile()   const { return bestClustFile_;   }
  string combsFile()       const { return combsFile_;       }
  bool   print()           const { return print_;           }
  int    iter()            const { return iter_;            }
  bool   mergeOccurred()   const { return mergeOccurred_;   }
  int    maxIter()         const { return maxIter_;         }
  bool   checkSymValid()   const { return checkSymValid_;   }
  int    minSymAppear()    const { return minSymAppear_;    }
  int    maxShared()       const { return maxShared_;       }
  double threshRPerPair()  const { return threshRPerPair_;  }
  double llThreshold()     const { return llThreshold_;     }
  bool   printUnit()       const { return printUnit_;       }
  bool   saveUnit()        const { return saveUnit_;        }
  double defaultFalseProb()const { return defaultFalseProb_;}
  bool   pureGreedy()      const { return pureGreedy_;      }

  Type* type(const string& name) const
  {
    StringToIntMap::const_iterator it = typeNameToIdMap_.find(name);
    if (it == typeNameToIdMap_.end()) return NULL;
    int id = (*it).second;
    return types_[id];
  }

  Type* type(const int& id) const   { return types_[id]; }

  const Array<Type*>& types() const { return types_;     }

  Relation* relation(const string& name) const
  {
    StringToIntMap::const_iterator it = relNameToIdMap_.find(name);
    if (it == relNameToIdMap_.end()) return NULL;
    int relId = (*it).second;
    return relations_[relId];
  }

  Relation* relation (const int& id) const  { return relations_[id]; }

 private:
  //create types, relations, tuples; update score
  void init()
  {
    readInfoFile();
    readDBFile(false); //first pass, no store tuples, read & add sym

    for (int i = 0; i < types_.size(); i++)
    {
      Util::assertt(types_[i]->numSymbols() != 0,
                    "no constants defined for", types_[i]->name().c_str(), -1);
      types_[i]->initClusters();
    }

    for (int i = 0; i < relations_.size(); i++)
      relations_[i]->createClustIdToCombs();

    readDBFile(true);  //second pass, store tuples, don't add sym
    for (int i=0; i < types_.size(); i++)
    {
      types_[i]->compress();
      types_[i]->setRepeatedInNaryRelation();
      types_[i]->setBinRelationsWithSameTypes();
    }

    //commented out: check clusters are correctly created for symbols
    /*for (int i = 0; i < types_.size(); i++)
    {
      cout << "\t" << types_[i]->name() << endl;
      types_[i]->printSymbols(cout);
    }//*/

    /*for (int i = 0; i < types_.size(); i++)
    {
      cout << "\t" << types_[i]->name() << endl;
      int noop;
      types_[i]->printClusters(cout, true, noop, false, false);
    }//*/

    //commented out: check order of combs in comb trees
    /*for (int i=0; i < relations_.size(); i++)
      relations_[i]->checkOrderInCombTrees();
    exit(-1);//*/
  }//init()

  void readInfoFile()
  {
    string infoFile = inDir_;
    if (infoFile.at( infoFile.length()-1) != '/') infoFile += '/';
    infoFile.append(infoFile_);

    ifstream inInfo(infoFile.c_str());
    Util::assertt(!inInfo.fail(),"bad file name", infoFile, -1);
    bool relFound = false;
    string buf;
    StringHashArray typesRead;

    while (getline(inInfo,buf))
    {
      if (buf.empty()) continue;
      buf = Util::trim(buf);

      if (buf.find("//Relations") == 0) //read the relations and types
      {
        relFound = true;
        createRelsTypes(inInfo);
      }
      else
      if (buf.find("//Types") == 0) //read info about types
        createTypeInfos(inInfo, typesRead);
      else
      if (buf.find("//") == 0) continue;
    }

    //set the default values for the types which are not in info file
    for (int i = 0; i < types_.size(); i++)
    {
      Type* type = types_[i];
      if (!typesRead.contains(type->name()))
      {
        type->setDoCluster(true);
        type->setMinClusters(1);
        type->setMaxClusters(INT_MAX);
      }
    }

    //checks
    Util::assertt(relFound,"no relation declarations", -1);

    //check type ids
    for (int t = 0; t < types_.size(); t++)
      Util::assertt(types_[t]->id() == t, "in MRC::init(). Wrong type id",-1);

    //check relation ids
    for (int r = 0; r < relations_.size(); r++)
      Util::assertt(relations_[r]->id() == r, "in MRC::init(). Wrong relation id",-1);

    cout << "\tTYPES: " << endl;
    for (int t = 0; t < types_.size(); t++)
    {
      cout << *(types_[t]) << endl;
      //types_[t]->printRels(cout); cout << endl;
    }
    cout << "\tRELATIONS: " << endl;
    for (int r = 0; r < relations_.size(); r++)
      cout << *(relations_[r]) << "  " << relations_[r]->betaTrue() << " "
           << relations_[r]->betaTotal() << "  "
           << (relations_[r]->isClosedWorld()?"cw":"ow") << endl;
  }//readInfoFile()

  void createRelsTypes(ifstream& inInfo)
  {
    StringHashArray relNames;
    StringHashArray typeNames;
    string buf;

    while(true)
    {
      getline(inInfo,buf);
      Util::trim(buf);
      if (buf.empty()) break;
      if (buf.find("//") == 0) continue;

      string relStr;
      Array<string> typeStrs;
      double betaTrue, betaTotal;
      bool closedWorld;

      //extract relation, type, beta, and closedWorld property
      readRelString(buf, relStr, typeStrs, closedWorld, betaTrue, betaTotal);

      hash_map<int,Array<int>*> typeIdToIndexes; //maps typeId to args of relation
      IntHashArray uniqueTypeIds;

      //////////////// create types
      Array<Type*> types( typeStrs.size() ); //types in current relation
      for (int t = 0; t < typeStrs.size(); t++)
      {
        Type* type = NULL;
        int id = typeNames.append(typeStrs[t]);

        if (id >= 0) // if type does not already exists
        {
          type = new Type(id, typeStrs[t]);
          types_.append(type);
        }
        else
        { //else type already exists
          id = typeNames.find(typeStrs[t]);
          type = types_[id];
        }
        types.append(type);

        //store the index at which each type is found in relation
        uniqueTypeIds.append( type->id() );
        hash_map<int,Array<int>*>::iterator it = typeIdToIndexes.find(type->id());
        Array<int>* indexes;
        if (it == typeIdToIndexes.end())
        {
          indexes = new Array<int>;
          typeIdToIndexes[ type->id() ] = indexes;
        }
        else
          indexes = (*it).second;
        indexes->append(t);
      }

      //////////////// create relations
      int relId = relNames.append(relStr);
      Relation* relation;
      if (relId >= 0) // if relation has not been declared
      {
        relation = new Relation(relId, relStr, closedWorld, betaTrue, betaTotal,
                                types);
        int rid = relations_.append(relation);
        Util::assertt(rid == relId, "in MRC::init(). rid != relId",-1);
      }
      else
      {
        //relation = relations_[ relNameToIdMap_[relStr] ];
        Util::exit("redeclared relation", relStr, -1);
      }

      //////////////// create Rel object for each type
      for (int i = 0; i < uniqueTypeIds.size(); i++)
      {
        int typeId = uniqueTypeIds[i];
        Array<int>& indexes = *( typeIdToIndexes[ typeId ] );
        Rel* rel = new Rel(relation);
        for (int j = 0; j < indexes.size(); j++)
          rel->appendIndex( indexes[j] );
        rel->compressIndexes();
        Type* type = types_[typeId];
        type->addRel(rel);
        delete &indexes;
      }
    }

    //create map from type/relation names to their ids
    for (int i = 0; i < types_.size(); i++)
      typeNameToIdMap_[ types_[i]->name() ] = types_[i]->id();
    Util::assertt((int)typeNameToIdMap_.size() == types_.size(), "wrong #type",-1);

    for (int i = 0; i < relations_.size(); i++)
      relNameToIdMap_[ relations_[i]->name() ] = relations_[i]->id();
    Util::assertt((int)relNameToIdMap_.size() == relations_.size(), "wrong #rel",-1);
  }//createRelsTypes()

  void readRelString(const string& str, string& relStr,
                     Array<string>& typeStrs, bool& closedWorld,
                     double& betaTrue, double& betaTotal)
  {
    string::size_type ltparen = str.find("(");
    Util::assertt(ltparen != string::npos, "in MRC::reRelTypes. No (",-1);
    relStr = Util::trim( str.substr(0, ltparen) );
    int i = ltparen+1;
    while (true)
    {
      string::size_type comma = str.find(",",i);
      if (comma == string::npos) break;
      string typeStr = Util::trim( str.substr(i, comma-i) );
      typeStrs.append(typeStr);
      i = comma+1;
    }
    string::size_type rtparen = str.find(")",i);
    Util::assertt(rtparen != string::npos,"in MRC::reRelTypes. No )",-1);
    string typeStr = Util::trim( str.substr(i, rtparen-i) );
    typeStrs.append(typeStr);

    //set the value of beta for relation
    betaTrue = -1;
    betaTotal = -1;
    closedWorld = true;
    string paramStr = str.substr(rtparen+1, str.length()-rtparen-1);
    paramStr += " ::";
    istringstream iss(paramStr);
    string buf;
    while (true)
    {
      iss >> buf;
      if (buf.compare("::")==0) break;
      if (isdigit(buf.at(0)))
      {
        if (betaTrue == -1)       betaTrue = atof(buf.c_str());
        else if (betaTotal == -1) betaTotal = atof(buf.c_str());
        else Util::exit("confused beta values",-1);
      }
      else
      {
        if (buf.compare("ow")==0) closedWorld = false;
      }
    }

    if (betaTrue == -1) betaTrue = betaTrue_;
    if (betaTotal == -1) betaTotal = betaTotal_;

    //commented out: check whether relations are correctly read
    cout << relStr << "(";
    for (int i = 0; i < typeStrs.size()-1; i++)
      cout << typeStrs[i] << ",";
    cout << typeStrs.lastItem() << ")" << endl;//*/
  }//readRelString

  void createTypeInfos(ifstream& inInfo, StringHashArray& typesRead)
  {
    string buf;
    string name, toClustStr, minClustStr, maxClustStr;
    StringToIntMap::iterator it;

    //read the values of toCluster, minClusters, maxClusters
    while(true)
    {
      getline(inInfo,buf);
      Util::trim(buf);
      if (buf.empty()) break;
      if (buf.find("//") == 0) continue;
      istringstream iss(buf);
      iss >> name >> toClustStr >> minClustStr >> maxClustStr;
      it = typeNameToIdMap_.find(name);

      typesRead.append(name);
      Util::assertt(it != typeNameToIdMap_.end(), "type not found ", name.c_str(), -1);

      Type* type = types_[ (*it).second ];   assert(type->id() == (*it).second);

      bool toClust = false;
      int minClust = -1, maxClust = -1;

      if (toClustStr.compare("true") == 0)
      {
        minClust = (minClustStr.at(0) == '-') ? -1 : atoi(minClustStr.c_str());
        maxClust = (maxClustStr.at(0) == '-') ? -1 : atoi(maxClustStr.c_str());
        toClust = true;
      }
      else if (toClustStr.compare("false") == 0)
      {
        minClust = -1;
        maxClust = -1;
        toClust = false;
      }
      else
      {
        Util::exit("wrong value for toCluster. Expect true/false", -1);
      }

      Util::assertt(minClust <= maxClust,"expect minClust <= maxClust",-1);

      if (minClust == -1) minClust = 1;
      if (maxClust == -1) maxClust = INT_MAX;

      type->setDoCluster(toClust);
      type->setMinClusters(minClust);
      type->setMaxClusters(maxClust);
    }
  }//createTypeInfos

  void readDBFile(const bool& storeTupleValues)
  {
    string dbFile = inDir_;
    if (dbFile.at( dbFile.length()-1) != '/') dbFile += '/';
    dbFile.append(dbFile_);

    ifstream inDB(dbFile.c_str());
    Util::assertt(!inDB.fail(),"bad file name", dbFile, -1);
    string buf;

    while (getline(inDB,buf))
    {
      buf = Util::trim(buf);
      if (buf.empty() || buf.find("//") == 0) continue;
      if (buf.at(0) == '#') { readConstants(buf); continue; }
      readTuples(this, buf, storeTupleValues, NULL, false);
    }

    //commented out: checks
    /*if (!storeTupleValues)
    {
      for (int i = 0; i < types_.size(); i++) //check symbols stored in types
      {
        cout << "\t" << types_[i]->name() << " constants" << endl;
        types_[i]->printSymbols(cout);
      }
    }
    else
    {
      for (int i = 0; i < relations_.size(); i++) //check relation tuples
      {
        cout << "\t" << relations_[i]->name() << " tuples " << endl;
        relations_[i]->printTupleValues(cout);
      }

      for (int i = 0; i < relations_.size(); i++) //check relation tuples
      {
        cout << "\tcontents of " << relations_[i]->name() << " clustIdToCombs_"
             << endl;
        relations_[i]->printClustIdToCombs(cout);
      }
      //exit(-1);
    }//*/
  }//readDBFile()

  void readConstants(const string& constantsStr)
  {
    istringstream iss(constantsStr);
    string str, typeName;
    iss >> str;
    Util::assertt(str.compare("#CONSTANTS") == 0, "expect #CONSTANTS",-1);
    iss >> typeName;
    Type* type = NULL;
    for (int i = 0; i < types_.size(); i++)
      if (types_[i]->name().compare( typeName ) == 0)
      { type = types_[i]; break; }
    iss >> str >> str; // { constant
    int symId;
    while (str.compare("}") != 0)
    {
      type->addSymbol(str, symId);
      iss >> str;
    }
  }

  static void readTuples(MRC* const& mrcPtr, const string& str,
                         const bool& storeTupleValues,
                         Array<Array<string>*>* const& args,
                         const bool& storeUnknownTuples)
  {
    int tuplesSeen = 0;
    string relStr, argStr;
    string::size_type ltparen, comma, rtparen;
    int startIndex = 0, argIndex, symId;
    Relation* rel = NULL;
    Type* type;

    while (true)
    {
      ltparen = str.find("(", startIndex);

      if (ltparen == string::npos)
      {
        Util::assertt(tuplesSeen != 0, "in MRC::readTuples. No (",-1);
        break;
      }

      relStr = Util::trim( str.substr(startIndex, ltparen-startIndex) );

      double value = -1;
      if (isdigit(relStr.at(0)))
      {
        istringstream iss(relStr);
        iss >> value >> relStr;
        Util::assertt(0.0 <= value && value <= 1.0, "invalid tuple value",
                         value, -1);
      }
      else if (relStr.at(0) == '!')
      {
        assert(value == -1);
        value = FALSE;
        relStr = relStr.substr(1,relStr.length()-1);
      }
      else if (relStr.at(0) == '?')
      {
        assert(value == -1);
        value = UNKNOWN;
        relStr = relStr.substr(1,relStr.length()-1);
      }
      else if (value == -1)
      {
        value = TRUE;
      }

      relStr = Util::trim(relStr);
      if (mrcPtr)
      {
        rel = mrcPtr->relation(relStr); //must always retrieve relation
        Util::assertt(rel != NULL, "undefined relation", relStr.c_str(), -1);
      }

      int i = ltparen+1;
      argIndex = -1;

      Array<string>* curArgs = NULL;
      if (args) { curArgs = new Array<string>; args->append(curArgs); }

      Array<int> symIds;
      while (true)
      {
        comma = str.find(",",i);
        if (comma == string::npos) break;
        rtparen = str.find(")",i);
        if (rtparen < comma) break; //comma belongs to next atom
        argStr = Util::trim( str.substr(i, comma-i) );
        if (curArgs) curArgs->append(argStr);
        argIndex++;
        i = comma+1;
        if (mrcPtr)
        {
          type = rel->type(argIndex);
          type->addSymbol(argStr, symId);
          symIds.append(symId) ;
        }
      }
      rtparen = str.find(")",i);
      Util::assertt(tuplesSeen != 0 || rtparen != string::npos, "in MRC::readTuples. No )",-1);
      argStr = Util::trim( str.substr(i, rtparen-i) );
      if (curArgs) curArgs->append(argStr);
      argIndex++;
      if (mrcPtr)
      {
        type = rel->type(argIndex);
        type->addSymbol(argStr, symId);
        symIds.append(symId);
      }

      if (storeTupleValues && mrcPtr)
      {
        rel->setTupleValue(symIds, value);
        rel->addTupleToCombs(symIds, value, false);
      }

      if (storeUnknownTuples && mrcPtr)
      {
        rel->setCombUnknownTuples( symIds );
        rel->addUnknownTuple( symIds );
      }

      startIndex = rtparen+1;
      tuplesSeen++;
    }
    //cout << "tuplesRead = " << tuplesSeen << endl;
  }//readTuples

 public:
  //returns true if not all symbols of every type in one cluster
  bool search()
  {
    double startSec = timer_.time();
    cout <<"acSearch started..." << endl;
    checkAssumptions();

    iter_ = 0;
    setRelationsParams();
    if (checkSymValid_) setSymbolsValid();

    //A symbol is added to symSets_ if it appears minSymAppear_ times in
    //at least one relation
    populateSymSetsAndCreateCanopies(symSets_);

    Array<Type*> types( types_.size() );
    getTypesToCluster(types);

    double score = 0.0;
    mergeOccurred_ = false;
    bool hasMerge = true;
    while (hasMerge && iter_ < maxIter_)
    {
      double iterBegSec = timer_.time();
      ++iter_;  cout << "ITER " << iter_ << endl;

      hasMerge = false;
      double scoreChange = 0.0;
      for (int t = 0; t < types.size(); t++)
      {
        cout << "------------------ ITER " << iter_ << "  TYPE "<< types[t]->id() << " , " << types[t]->name() << endl;

        MergeTree goodMerges;
        findGoodMerges( types[t], goodMerges );
        //printGoodMerges( goodMerges );
        scoreChange += performMerges( types[t], goodMerges, hasMerge );
      }
      score += scoreChange;

      if (hasMerge)
      {
        mergeOccurred_ = true;
        cout << "============= " << iter_ << ": SCORE " << score<<"  , SC_CHANGE "
             << scoreChange << " , NEW CLUSTERING:" << endl;
        saveBestClusters(iter_, types, saveUnit_,  true); //save then print
        printBestClusters(cout, types, printUnit_, true);
      }

      cout << "ITER " << iter_ << " took ";
      timer_.printTime(cout, timer_.time()-iterBegSec);
      cout << " ; time elapsed = "; timer_.printTime(cout,timer_.time()-startSec);
      cout << endl;
    }
    cout << "ended merge loop because ";
    if (!hasMerge)              cout << "no merges" << endl;
    else if (iter_ >= maxIter_) cout << "reached max iter" << endl;
    else assert(false);

    cout << "============= " << iter_ << " , FINAL CLUSTERING:" << endl;
    //sort syms, clusts, & set clust ids
    for (int i = 0; i < types.size(); i++)
      types[i]->sortClustersAndRenumberIds(true);
    saveBestClusters(iter_, types_, false, false);  //save all types
    printBestClusters(cout, types, printUnit_, true);
    printBestClustersHisto(cout, types);
    printCombs();

    cout << iter_ << ": acSearch took ";
    timer_.printTime(cout, timer_.time()-startSec); cout << endl;

    //checkSearchEnd();

    bool allOneClust = true; //check whether all the types have only one clust
    for (int i = 0; i < types.size(); i++)
      if (types[i]->numClusters() > 1) { allOneClust = false; break; }

    return !allOneClust;
  }//acSearch()

 private:
  void checkAssumptions()
  {
    Util::assertt(inDir_.at( inDir_.length()-1 ) == '/', "expect inDir to end with /", -1);
    Util::assertt(outDir_.at( outDir_.length()-1 ) == '/', "expect outDir to end with /", -1);
    Util::assertt(minSymAppear_ > 0, "expect minSymAppear > 0", -1);

    /*for (int i = 0; i < relations_.size(); i++)
    {
      Relation* relation = relations_[i];
      const Array<Type*>& types = relation->types();
      Util::assertt(types.size() >= 2, "expect all relations to be of arity 2 or more", relation->name(), -1);
      //for (int j = 0; j < types.size(); j++)
      //  for (int k = j+1; k < types.size(); k++)
      //    Util::assertt(types[j] != types[k], "expect a type to appear at most once in a relation", -1);
    }*/
  }

  void setRelationsParams()
  {
    for (int i = 0; i < relations_.size(); i++)
    {
      relations_[i]->setDefaultFalseProb(defaultFalseProb_);
      relations_[i]->setBetaTrue();
      cout << "relation " << relations_[i]->name() << " betaTrue/betaTotal set to "
           << relations_[i]->betaTrue() << " / " << relations_[i]->betaTotal() << endl;
    }
  }

  void setSymbolsValid()
  {
    for (int i = 0; i < types_.size(); i++)
    {
      const Array<Symbol*>& symbols = types_[i]->symbols();
      for (int j = 0; j < symbols.size(); j++)
        symbols[j]->setValid(true);
    }
  }

  //A symbol is added to symSets if it appears minSymAppear_ times in
  //at least one relation
  void populateSymSetsAndCreateCanopies(Array<StringSet>& symSets)
  {
    symSets_.growToSize( types_.size() );

    //for each relation, build a map from symbol to counts for each dim
    //                   get the symbols that appear at least minSymAppear_
    //                   set its number of true tuples
    for (int i = 0; i < relations_.size(); i++)
    {
      Relation* relation = relations_[i];

      cout << "creating symbol to cnt map for relation " << relation->name() << "..." << endl;
      relation->populateSymSets(minSymAppear_, symSets);

      double begSec = timer_.time();
      cout << "creating canopyTree for relation " << relation->name() << "..." << endl;
      relation->createCanopyTree();
      cout << iter_ << ": create clusterSetTree took ";
      timer_.printTime(cout, timer_.time()-begSec); cout << endl;
    }
  }

  void getTypesToCluster(Array<Type*>& types)
  {
    for (int i = 0; i < types_.size(); i++)
      if (types_[i]->doCluster()) types.append( types_[i] );
  }

  void findGoodMerges(const Type* const& type, MergeTree& goodMerges)
  {
    double begSec = timer_.time();
    int typeId = type->id();
    int totalClustPairs = 0;
    ClusterPairTree clustPairs; //keep track of cluster pairs that were compared

    cout << iter_ << ": finding good merges for type " << typeId << "..." << endl;

    //for each relation that type appears in
    for (int i = 0; i < type->numRels(); i++)
    {
      Relation* relation = type->rel(i)->relation();
      const Array<int>& indexes  = type->rel(i)->indexes();
      for (int j = 0; j < indexes.size(); j++)
      {
        int dim = indexes[j];
        CanopyTree* canopyTree = relation->canopyTree( dim );
        if (canopyTree == NULL) continue;
        CanopyNode* node = canopyTree->head();
        while(node)
        {
          Canopy* canopy = node->data();
          ClusterTree& clusters = canopy->clusters_;
          if (1 < clusters.size() && clusters.size() <= maxShared_)
          {
            //for each pair of clusters in non-shared position (e.g. position A)
            //compute score if they are merged
            findGoodMergesInCanopy(type, clusters, totalClustPairs, clustPairs, goodMerges);
          }
          node = node->next();
        }
      }
    }

    int actualClustPairs = clustPairs.size();
    ClusterPairNode* cpNode = clustPairs.head(); //clean up clustPairs
    while(cpNode) { delete cpNode->data(); cpNode = cpNode->next(); }

    cout << iter_ << ": find good merges for type " << typeId << " took ";
    timer_.printTime(cout, timer_.time()-begSec); cout << endl;
    cout << iter_ <<": #cand clust pairs (have dup) = "<< totalClustPairs  <<endl;
    cout << iter_ <<": #actual pairs                = "<< actualClustPairs <<endl;
  }//findGoodMerges()

  void findGoodMergesInCanopy(const Type* const& type,const ClusterTree& canopy,
                              int& totalClustPairs, ClusterPairTree& clustPairs,
                              MergeTree& goodMerges)
  {
    //for each pair of clusters in canopy
    ClusterNode* inode = canopy.head();
    while(inode)
    {
      ClusterNode* jnode = inode->next();
      while(jnode)
      {
        totalClustPairs++;
        Cluster* iclust = inode->data();
        Cluster* jclust = jnode->data();

        if (!canCompare(iclust, jclust, clustPairs, type))
        { jnode = jnode->next(); continue; }

        Move* move;
        double change = computeMergeScore(type, iclust, jclust, move);

        //cout << change << "  " << (*iclust)[0]->name() << "  " << (*jclust)[0]->name() << endl;

        if (change > llThreshold_)
        {
          MoveNode* noop = NULL;
          bool inserted = goodMerges.insert(move, noop);
          if (!inserted) delete move;
        }
        else delete move;

        jnode = jnode->next();
      }

      inode = inode->next();
    }
  }//findGoodMergesInCanopy()

  bool canCompare(const Cluster*const& iclust, const Cluster*const& jclust,
                  ClusterPairTree& clustPairs, const Type* const& type)
  {
    //check whether sym is valid
    if (checkSymValid_ && !checkClustersValid(iclust, jclust)) return false;

    //check whether iclust and jclust appear more than min #times
    if (!checkClustersAppearEnough(iclust, jclust, type)) return false;

    //check whether the cluster pair have been compared; if so, skip it
    ClusterPair* cp;
    if (iclust < jclust)  cp = new ClusterPair(iclust, jclust);
    else                  cp = new ClusterPair(jclust, iclust);
    if (!clustPairs.insert(cp)) { delete cp; return false; }
    return true;
  }

  bool checkClustersValid(const Cluster* const& iclust,
                          const Cluster* const& jclust)
  {
    if (iclust->size() > 1 && jclust->size() > 1)         return true;
    if (( iclust->size()==1 && !(*iclust)[0]->valid() ) ||
        ( jclust->size()==1 && !(*jclust)[0]->valid() ))  return false;
                                                          return true;
  }

  bool checkClustersAppearEnough(const Cluster* const& iclust,
                                 const Cluster* const& jclust,
                                 const Type* const& type)
  {
    StringSet& symSet = symSets_[ type->id() ];
    if (iclust->size() > 1 && jclust->size() > 1)  return true;
    if (( iclust->size()==1 && symSet.find( (*iclust)[0]->name() )==symSet.end() )
        ||
        ( jclust->size()==1 && symSet.find( (*jclust)[0]->name() )==symSet.end()))
      return false;
    return true;
  }

  double computeMergeScore(const Type* const& type, const Cluster* const& c0,
                           const Cluster* const& c1, Move*& move)
  {
    assert(c0 != c1);

    Cluster* clustA, *clustB;
    if (c0->id() < c1->id()) { clustA = (Cluster*) c0; clustB = (Cluster*) c1; }
    else                     { clustA = (Cluster*) c1; clustB = (Cluster*) c0; }

    Array<Move*> mmove(1);
    double change    = 0.0;  //change is bestChange
    double badChange = 0.0;  //badChange <= 0.0

    //cout << "trying merge " << (*clustA)[0]->name() << " , "
    //     << (*clustB)[0]->name() << endl;
    badChange = tmpMergeClusters(type, clustA, clustB, change, mmove);

    if (mmove.empty())
    {
      move = NULL;
      assert(change <= 0.0);
      return badChange;
    }

    assert(mmove.size() == 1);
    move = mmove[0];
    return change;
  }

  double tmpMergeClusters(const Type* const& type,
                          Cluster* const& clust0, Cluster* const & clust1,
                          double& bestChange, Array<Move*>& bestMoves)
  {
    double change = 0.0;

    //stores combs that contain two or more of clust0 or clust1
    Array<CombNode*> repeatedNodes;
    Array<CombTree*> repeatedTrees;

    computeChangeForCombWithRepeatedClusters(type, clust0->id(), clust1->id(),
                                             change, repeatedNodes, repeatedTrees);

    //at this point no comb contains two of clust0 or clust1

    //find the relations for which cluster combinations overlap/don't overlap
    Array<Rel*> overlapRels  ( type->numRels() );
    Array<Rel*> noOverlapRels( type->numRels() );
    getRelsWithOverlapCombs(type, clust0->id(), clust1->id(), overlapRels, noOverlapRels);

    //compute change from overlapping combs
    computeChangeOverlapCombs(overlapRels, clust0, clust1, change, bestChange);

    //threshold per pair of symbols and consider ll change for all tuples
    Util::assertt(threshRPerPair_ >= 0.0, "expect threshRPerPair >= 0.0", -1);
    double thresh = threshRPerPair_ * clust0->size() * clust1->size();
    change -= thresh;

    //change can only decrease from here on

    if (change<bestChange)
    {
      cleanUp(overlapRels,noOverlapRels);
      Relation::restoreCombNodesToTrees(repeatedNodes, repeatedTrees);
      return change;
    }

    //compute change from non-overlapping combs
    computeChangeNonOverlapCombs(noOverlapRels, clust0, clust1, change, bestChange);

    //printMergeScore(cout, " TMP MERGE", type, clust0, clust1, change);

    if (change > bestChange)
    {
      //printMergeScore(cout,"TMP MERGE bestChange",type,clust0,clust1,change);
      if (change > bestChange) bestMoves.deleteItemsAndClear();
      bestChange = change;
      bestMoves.append(new Move(MOVE_MERGE, type->id(), clust0, clust1, change));
    }

    cleanUp(overlapRels, noOverlapRels);
    Relation::restoreCombNodesToTrees(repeatedNodes, repeatedTrees);
    return change;
  }//tmpMergeClusters()

  void computeChangeForCombWithRepeatedClusters(const Type* const& type,
                                                const int& clustId0,
                                                const int& clustId1,
                                                double& change,
                                                Array<CombNode*>& repeatedNodes,
                                                Array<CombTree*>& repeatedTrees)

  {
    for (int relIdx = 0; relIdx < type->numRels(); relIdx++)
    {
      Relation* relation = type->rel(relIdx)->relation();
      const Array<int>& typeIndexes = type->rel(relIdx)->indexes();

      if (typeIndexes.size() > 1) //if type appears at more than one position
      {
        for (int i = 0; i < typeIndexes.size(); i++)
        {
          int dim = typeIndexes[i];
          Array<CombNode*> newRepNodes;

          relation->removeCombsWithTwoOrMoreOfClusts(dim, clustId0, clustId1,
                                                     typeIndexes, repeatedNodes,
                                                     repeatedTrees, newRepNodes);
          int    numOverlapCombs = 0;
          double tmpChange = relation->tmpMergeClustersRepeated(dim, clustId0,
                                                                clustId1,
                                                                newRepNodes,
                                                                numOverlapCombs);
          change += tmpChange;
          change += -numOverlapCombs * lambda_;   assert(lambda_ <= 0.0);
        }
      }
    }
  }

  void getRelsWithOverlapCombs(const Type* const& type,
                               const int& clustId0, const int& clustId1,
                               Array<Rel*>& overlapRels,
                               Array<Rel*>& noOverlapRels)
  {
    for (int relIdx = 0; relIdx < type->numRels(); relIdx++)
    {
      Relation* relation = type->rel(relIdx)->relation();
      const Array<int>& typeIndexes = type->rel(relIdx)->indexes();

      Array<int> overlapIndexes, noOverlapIndexes; //positions of type in relation
      relation->overlapCombs(typeIndexes, clustId0, clustId1, overlapIndexes,
                             noOverlapIndexes);
      if (!overlapIndexes.empty())
        overlapRels.append( new Rel(relation, overlapIndexes) );
      else
      if (!noOverlapIndexes.empty())
        noOverlapRels.append( new Rel(relation, noOverlapIndexes) );
    }
  }

  void computeChangeOverlapCombs(const Array<Rel*>& overlapRels,
                                 const Cluster* const& clust0,
                                 const Cluster* const& clust1,
                                 double& change, const double& bestChange)
  {
    //compute change from overlapping combs
    for (int i = 0; i < overlapRels.size(); i++)
    {
      Relation* relation = overlapRels[i]->relation();
      const Array<int>& indexes = overlapRels[i]->indexes();
      for (int j = 0; j < indexes.size(); j++)
      {
        int dim = indexes[j];
        bool checkChange = (overlapRels.size() == 1 && indexes.size() == 1);
        double tmpChange = 0.0;
        int numOverlapCombs = 0;
        tmpChange = relation->tmpMergeClusters(dim, clust0->id(), clust1->id(),
                                               bestChange, checkChange,
                                               numOverlapCombs);
        /*cout << "\t TMPCHANGE " << tmpChange << "  PRIOR " << change
             << "  " << (*clust0)[0]->name() << "  " << (*clust1)[0]->name() 
             << endl;//*/

        Util::assertt(lambda_ <= 0.0, "expect lambda <= 0", -1);
        tmpChange += (-numOverlapCombs) * lambda_;
        change += tmpChange;

        /*cout << "\t\t numOverlapCombs " << numOverlapCombs << endl;
        cout << "\t\t lambda " << lambda_ << endl;
        //cout << "\t\t thresh " << thresh << endl;
        cout << "\t\t change " << change << endl;//*/
      }
    }
  }//computeChangeOverlapCombs()

  void computeChangeNonOverlapCombs(const Array<Rel*>& noOverlapRels,
                                    const Cluster* const& clust0,
                                    const Cluster* const& clust1,
                                    double& change, double& bestChange)
  {
    for (int i = 0; i < noOverlapRels.size(); i++)
    {
      Relation* relation = noOverlapRels[i]->relation();
      const Array<int>& indexes = noOverlapRels[i]->indexes();
      for (int j = 0; j < indexes.size(); j++)
      {
        int dim = indexes[j];
        double newClustSize = clust0->size() + clust1->size();
        bool checkChange = true;
        double tmpChange;
        tmpChange = relation->tmpMergeClustersNoOverlap(dim, clust0->id(),
                                                        clust1->id(), newClustSize,
                                                        bestChange, checkChange);
        change += tmpChange;
      }
    }
  }

  void cleanUp(Array<Rel*>& overlapRels, Array<Rel*>& noOverlapRels) const
  {
    overlapRels.deleteItemsAndClear();
    noOverlapRels.deleteItemsAndClear();
  }

  double performMerges(const Type* const& type, const MergeTree& goodMerges,
                       bool& hasMerge)
  {
    double begSec = timer_.time();
    int typeId = type->id();
    cout << iter_ << ": performing merges for type " << typeId << "..." << endl;

    //clusters that're to be merged w/h chosenClusters but not b/c latter merged
    ConstClusterSet chosenClusters;
    Array<Move*> chosenMerges(10000);

    bool pureGreedy;
    const Array<Relation*>* binRels = NULL;
    if (pureGreedy_ || type->repeatedInNaryRelation()) pureGreedy = true;
    else  { pureGreedy = false; binRels = type->binRelationsWithSameTypes(); }

    if (pureGreedy)
    {
      MoveNode* node = goodMerges.head();
      while(node)
      {
        Move* move = node->data();
        if (node == goodMerges.head()) chosenMerges.append(move);
        else                           delete move;
        node = node->next();
      }
    }
    else
    { //merge as many pairs of clusters as possible and not just the best one
      MoveNode* node = goodMerges.head();
      while (node) //for each merge moves
      {
        Move* move = node->data();;
        Cluster* clust0 = move->clust0_;
        Cluster* clust1 = move->clust1_;
        assert(move->clust0_->id() < move->clust1_->id());

        if ( canBeMerged(clust0, clust1, chosenClusters, binRels) )
        {
          chosenMerges.append(move);
          chosenClusters.insert(clust0);
          chosenClusters.insert(clust1);
        }
        else
        {
          delete move;
        }

        node = node->next();
      }
    }

    int numMerges = 0;
    double scoreChange = 0;
    for (int i = 0; i < chosenMerges.size(); i++)
    {
      Move* move = chosenMerges[i];
      bool ok = makeMerge(move, numMerges);
      if (ok) numMerges++;
      scoreChange += move->llChange_;
      hasMerge = true;
      delete move;
    }

    cout << iter_ << ": perform merges for type " << typeId << " took ";
    timer_.printTime(cout, timer_.time()-begSec); cout << endl;
    cout << iter_ << ": #merges " << typeId << " = " << numMerges << " / "
         << chosenMerges.size() << " / " << goodMerges.size() << endl << endl;

    return scoreChange;
  }//performMerges()

  bool canBeMerged(const Cluster* const& clust0, const Cluster* const& clust1,
                   const ConstClusterSet& chosenClusters,
                   const Array<Relation*>* const& binRelations)
  {
    if (chosenClusters.find(clust0) != chosenClusters.end()) return false;
    if (chosenClusters.find(clust1) != chosenClusters.end()) return false;
    if (binRelations && appearInCombWithChosenCluster(clust0, clust1, chosenClusters, *binRelations))
      return false;
    return true;
  }

  //returns true if clust0 or clust1 appears with a chosen cluster in
  //a true grounding of binary relation R(type,type)
  bool appearInCombWithChosenCluster(const Cluster* const& clust0,
                                     const Cluster* const& clust1,
                                     const ConstClusterSet& chosenClusters,
                                     const Array<Relation*>& binRelations)
  {
    for (int i = 0; i < binRelations.size(); i++)
    {
      Relation* binRel = binRelations[i];
      assert(binRel->types().size() == 2 && binRel->type(0) == binRel->type(1));
      ConstClusterSet::iterator it = chosenClusters.begin();
      for (; it != chosenClusters.end(); it++)
      {
        Cluster* clust = (Cluster*) (*it);
        if (binRel->appearInComb(clust0, clust) || binRel->appearInComb(clust1, clust))
          return true;
      }
    }
    return false;
  }

  bool makeMerge(const Move* const& move, const int& numMerges)
  {
    cout << "\t" << iter_ << ": making merge " << numMerges << " for type "
         << move->typeId_ << "..." <<endl;
    printMerge(cout, move, iter_, "\t");

    //in clust ids may have changed due to previous merges
    assert(move->clust0_ != move->clust1_);
    if (move->clust0_->id() > move->clust1_->id())
    {
      Move* mmove   = (Move*) move;
      Cluster* tmp  = mmove->clust0_;
      mmove->clust0_ = mmove->clust1_;
      mmove->clust1_ = tmp;
    }

    int typeId      = move->typeId_;
    Cluster* clust0 = move->clust0_;
    Cluster* clust1 = move->clust1_;

    Type* type = types_[typeId];
    for (int i = 0; i < type->numRels(); i++)
    {
      Relation* relation = type->rel(i)->relation();
      const Array<int>& typeIndexes = type->rel(i)->indexes();

      for (int j = 0; j < typeIndexes.size(); j++)
      {
        int dim = typeIndexes[j];
        relation->updateCanopyTree( dim, clust0, clust1, typeIndexes );
      }
    }

    mergeClusters(move);

    return true;
  }//makeMerge()

  void mergeClusters(const Move* const& move)
  {
    assert(move->clust0_->id() < move->clust1_->id());

    Type* type = types_[ move->typeId_ ];  assert(type->id() == move->typeId_);
    Cluster* clust0 = move->clust0_; assert(type->cluster(clust0->id())==clust0);
    Cluster* clust1 = move->clust1_; assert(type->cluster(clust1->id())==clust1);
    loglike_ += move->llChange_;

    //for each relation that has argument with type
    //  remove clust1's combs
    //  for each removed comb of clust1
    //    for each cluster in comb that is not clust 1
    //      remove comb from cluster's comb
    //  merge clust1 into clust0
    //  for each removed comb of clust1 (clust1 is now clust0)
    //    for each cluster in comb
    //      add comb to cluster's comb

    //for each relation that has argument with type
    for (int i = 0; i < type->numRels(); i++)
    {
      Relation* relation = type->rel(i)->relation();
      const Array<int>& typeIndexes = type->rel(i)->indexes();
      for (int j = 0; j < typeIndexes.size(); j++)
      {
        int dim = typeIndexes[j];
        //remove combs containing clust1 but don't remove combs from clust1's tree
        relation->removeCombsExceptClust(dim, clust1);
      }
    }

    clusterAddSymbols(clust0, clust1); //add clust1 symbols to clust0

    //for each relation that has argument with type
    for (int i = 0; i < type->numRels(); i++)
    {
      Relation* relation = type->rel(i)->relation();
      const Array<int>& typeIndexes = type->rel(i)->indexes();
      for (int j = 0; j < typeIndexes.size(); j++)
      {
        int dim = typeIndexes[j];

        //remove clust1's tree, merge clust1 into clust0, update numTrueTuples of
        //combs containing clust0
        relation->removeClust1CombsAndMerge(dim, clust0, clust1, typeIndexes);
      }
    }

    //for each relation that has argument with type
    for (int i = 0; i < type->numRels(); i++)
    {
      Relation* relation = type->rel(i)->relation();
      const Array<int>& typeIndexes = type->rel(i)->indexes();
      for (int j = 0; j < typeIndexes.size(); j++)
      {
        int dim = typeIndexes[j];
        relation->removeNullInClustIdToCombs(dim, clust1->id());
      }
    }

    Cluster* rmClust = type->removeCluster(clust1);  assert(rmClust==clust1);
    delete clust1;

    //NOTE: relation->removeNullInClustIdToCombs() and type->removeCluster()
    //      requires clust1 to be removed and replaced with last cluster


    Util::assertt(rmClust != NULL,"rmClust==NULL!",-1);
    assert(type->numClusters() >= type->minClusters());
    assert(type->numClusters() <= type->maxClusters());
  }//mergeClusters()

  void printGoodMerges(const MergeTree& goodMerges)
  {
    MoveNode* node = goodMerges.head();
    while (node) //for each merge moves
    {
      Move* move = node->data();
      cout << "TOTAL CHANGE = " << move->llChange_  << " "
           << move->clust0_->symbol(0)->name() << " (" << move->clust0_->size() << ")  "
           << move->clust1_->symbol(0)->name() << " (" << move->clust1_->size() << ")" << endl;
      node = node->next();
    }
  }

  int clusterAddSymbol(Cluster* const& cluster, Symbol* const& symbol)
  {
    symbol->setCluster(cluster);
    return cluster->addSymbol(symbol);
  }

  void clusterAddSymbol(Cluster* const& cluster, Symbol* const& symbol,
                        const int& prevSymIdx)
  {
    symbol->setCluster(cluster);
    cluster->addSymbol(symbol, prevSymIdx);
  }

  void clusterAddSymbols(Cluster* const& dstCluster,
                        const Cluster* const& srcCluster)
  {
    Array<Symbol*>& syms = dstCluster->symbols();
    int origSize = syms.size();
    syms.growToSize(origSize + srcCluster->size());
    for (int i = 0; i < srcCluster->size(); i++)
    {
      syms[origSize+i] = srcCluster->symbol(i);
      syms[origSize+i]->setCluster(dstCluster);
    }
  }

  Symbol* clusterRemoveSymbol(Cluster* const& cluster, const int& symbolIdx)
  {
    Symbol* symbol = cluster->removeSymbol(symbolIdx);
    symbol->setCluster(NULL);
    return symbol;
  }

  void clusterRemoveSymbols(Cluster* const& dstCluster,
                            const Cluster* const& srcCluster)
  {
    for (int i = srcCluster->size()-1; i >= 0; i--)
    {
      Symbol* sym = dstCluster->removeLastSymbol();
      assert(sym == srcCluster->symbol(i)); sym = NULL;
    }
  }

  bool checkClusterIds() const
  {
    for (int i = 0; i < types_.size(); i++)
      if (!types_[i]->checkClusterIds()) return false;
    return true;
  }

  void saveBestClusters(const int& iter) { saveBestClusters(iter, types_, true, false); }

  void saveBestClusters(const int& iter, const Array<Type*>& types,
                        const bool& printUnit, const bool& sortClusts)
  {
    for (int t = 0; t < types.size(); t++)
      types[t]->setBestClusters(types[t]->clusters());

    string bestClustFile;
    bestClustFile.append( outDir_ ).append( bestClustFile_ );

    string dst = bestClustFile; dst.append(".old");
    bool openSrc, openDst;
    Util::copyFile(bestClustFile, dst, openSrc, openDst);
    //commented out: may fail to open src because it does not exist
    //Util::assertt(openSrc && openDst, "failed to copy bestClustFile", -1);

    ofstream out(bestClustFile.c_str());
    out << "BEST LOGLIKE = " << loglike_ << endl;
    out << "BEST SEARCH  = " << iter << endl;
    out << "BEST ClUSTERS: " << endl;
    printBestClusters(out, types, printUnit, sortClusts);
    out.close();
  }

  void printBestClusters(ostream& out, const Array<Type*>& types,
                         const bool& printUnit, const bool& sortClusts) const
  { printClusters(true, out, types, printUnit, sortClusts); }

  void printClusters(const bool& best, ostream& out, const Array<Type*>& types,
                     const bool& printUnit, const bool& sortClusts) const
  {
    for (int i = 0; i < types.size(); i++)
    {
      if (types[i]->doCluster())
      {
        out << "\tTYPE " << types[i]->name() << "  " << endl;
        static bool sortLargestFirst = false;
        int numNonUnit = 0;

        if (best) types[i]->printBestClusters(out, printUnit, numNonUnit,
                                              sortClusts, sortLargestFirst);
        else      types[i]->printClusters(    out, printUnit, numNonUnit,
                                              sortClusts, sortLargestFirst);

        out << "\tTYPE " << types[i]->name() << "  ";
        out << "#= nonUnit " << numNonUnit << " / " << types[i]->numClusters()
            << "  Unit " << types[i]->numClusters()-numNonUnit << " / "
            << types[i]->numClusters() << "  sym " << types[i]->numSymbols()
            << endl << endl;
      }
    }
  }

  //ASSUMPTION: clusters have been sorted
  void printBestClustersHisto(ostream& out, const Array<Type*>& types)
  {
    for (int i = 0; i < types.size(); i++)
    {
      cout << "TYPE " << types[i]->name() << "  cluster histogram" << endl;
      const Array<Cluster*>& clusters = types[i]->clusters();
      int cnt = 0;
      int curSize = 0;
      for (int i = 0; i < clusters.size(); i++)
      {
        Cluster& clust = *( clusters[i] );

        if (curSize != clust.size())
        {
          out << "cluster size " << curSize << " : " << cnt << endl;
          curSize = clust.size();
          cnt = 0;
        }
        cnt++;
      }
      out << "cluster size " << curSize << " : " << cnt << endl;
    }
  }

  void printCombs()
  {
    string combsFile;
    combsFile.append( outDir_ ).append( combsFile_ );
    ofstream out(combsFile.c_str());
    int idx = 0;
    for (int i = 0; i < relations_.size(); i++)
      relations_[i]->printCombs(out, idx);
    out.close();
  }

  void readParams(const char* const& paramFile)
  {
    ifstream in(paramFile);
    string buf;
    while (getline(in,buf))
    {
      buf = Util::trim(buf);
      if (buf.empty()) continue;
      if (buf.at(0) == '/' && buf.at(1) == '/') continue;
      istringstream iss(buf); //cout << "buf = " << buf << endl;
      string name, valueStr;
      iss >> name >> valueStr;
      const char* value = valueStr.c_str();
      bool valueIsTrue = strcasecmp(value,"true")==0;

      if      (name.compare("inDir")         ==0)  inDir_         = value;
      else if (name.compare("outDir")        ==0)  outDir_        = value;
      else if (name.compare("infoFile")      ==0)  infoFile_      = value;
      else if (name.compare("dbFile")        ==0)  dbFile_        = value;
      else if (name.compare("betaTrue")      ==0)  betaTrue_      = atof(value);
      else if (name.compare("betaTotal")     ==0)  betaTotal_     = atof(value);
      else if (name.compare("lambda")        ==0)  lambda_        = atof(value);
      else if (name.compare("bestClustFile") ==0)  bestClustFile_ = value;
      else if (name.compare("combsFile")     ==0)  combsFile_     = value;
      else if (name.compare("maxIter")       ==0)  maxIter_       = atoi(value);
      else if (name.compare("checkSymValid") ==0)  checkSymValid_ = valueIsTrue;
      else if (name.compare("minSymAppear")  ==0)  minSymAppear_  = atoi(value);
      else if (name.compare("maxShared")     ==0)  maxShared_     = atoi(value);
      else if (name.compare("threshRPerPair")==0)  threshRPerPair_= atof(value);
      else if (name.compare("llThreshold")   ==0)  llThreshold_   = atof(value);
      else if (name.compare("printUnit")     ==0)  printUnit_     = valueIsTrue;
      else if (name.compare("saveUnit")      ==0)  saveUnit_      = valueIsTrue;
      else if (name.compare("defaultFalseProb")==0)defaultFalseProb_= atof(value);
      else if (name.compare("pureGreedy")    ==0)  pureGreedy_    = valueIsTrue;
      else Util::exit("unknown param", buf.c_str(), -1);
    }
    in.close();

    printParams(cout);
  }

  void printParams(ostream& out)
  {
    out << "---------------- params -------------------" << endl;
    out << "inDir               = " << inDir_                          << endl;
    out << "outDir              = " << outDir_                         << endl;
    out << "infoFile            = " << infoFile_                       << endl;
    out << "dbFile              = " << dbFile_                         << endl;
    out << "betaTrue            = " << betaTrue_                       << endl;
    out << "betaTotal           = " << betaTotal_                      << endl;
    out << "lambda              = " << lambda_                         << endl;
    out << "bestClustFile       = " << bestClustFile_                  << endl;
    out << "combsFile           = " << combsFile_                      << endl;
    out << endl;
    out << "maxIter             = " << maxIter_                        << endl;
    out << "checkSymValid       = " << (checkSymValid_?"true":"false") << endl;
    out << "minSymAppear        = " << minSymAppear_                   << endl;
    out << "maxShared           = " << maxShared_                      << endl;
    out << "threshRPerPair      = " << threshRPerPair_                 << endl;
    out << "llThreshold         = " << llThreshold_                    << endl;
    out << "printUnit           = " << (printUnit_?"true":"false")     << endl;
    out << "saveUnit            = " << (saveUnit_?"true":"false")      << endl;
    out << "defaultFalseProb    = " << defaultFalseProb_               << endl;
    out << "pureGreedy          = " << (pureGreedy_?"true":"false")    << endl;
    out << "-------------------------------------------" << endl;
  }

  void printMerge(ostream& out, const Move* const& move, const int& iter,
                  const string& prepend)
  {
    //print the merge
    out << prepend;
    out << "ITER " << iter << " ";
    Type* type = types_[ move->typeId_ ];  assert(type->id() == move->typeId_);
    Cluster* clust0  = move->clust0_;  assert(type->cluster(clust0->id())==clust0);
    Cluster* clust1  = move->clust1_;  assert(type->cluster(clust1->id())==clust1);
    out << "MAKE MERGE : " << type->name() << " CLUS "
        << clust0->id() << " (sz=" << clust0->size() << ") "
        << (*clust0)[0]->name()
        << " & "
        << clust1->id() << " (sz=" << clust1->size() << ") "
        << (*clust1)[0]->name()
        << "  llChange= " << move->llChange_ << endl;
  }

  void printSymbols(const Type* const& type, const bool& checkValid,
                    const bool& checkAppearEnough, const char* outFile)
  {
    ofstream out(outFile);
    Util::assertt(!out.fail(),"Failed to open", outFile, -1);

    const Array<Symbol*>& symbols = type->symbols();
    for (int i = 0; i < symbols.size(); i++)
    {
      Symbol* sym = symbols[i];
      if (checkValid && !sym->valid()) continue;
      if (checkAppearEnough)
      {
        StringSet& symSet = symSets_[ type->id() ];
        if (symSet.find( sym->name() ) == symSet.end()) continue;
      }
      out << sym->name() << endl;
    }

    out.close();
  }

  void checkSearchEnd()
  {
    cout << endl;
    //print tuples to check
    for (int i = 0; i < relations_.size(); i++)
    {
      cout << "END TUPLE VALUES FOR " << relations_[i]->name() << endl;
      relations_[i]->printTupleValues(cout);
    }

    //print clustIdToCombs to check
    for (int i = 0; i < relations_.size(); i++)
    {
      cout << "CLUST_ID_TO_COMBS FOR " << relations_[i]->name() << endl;
      relations_[i]->printClustIdToCombs(cout);
    }

    //print canopyTree to check
    for (int i = 0; i < relations_.size(); i++)
    {
      cout << "CANOPY_TREE FOR " << relations_[i]->name() << endl;
      relations_[i]->printCanopyTree(cout);
    }
  }

 private:
  string inDir_;   //input directory for DB (not for infoFile)
  string outDir_;  //output directory for best clusters
  string infoFile_;
  string dbFile_;

  double betaTrue_;
  double betaTotal_;
  double lambda_;  //value of exponential prior on #cluster combination
  double loglike_;

  Array<Type*>     types_;
  Array<Relation*> relations_;
  StringToIntMap   typeNameToIdMap_;
  StringToIntMap   relNameToIdMap_;

  string bestClustFile_;
  string combsFile_;
  bool print_;
  Timer timer_;

  int    iter_;
  bool   mergeOccurred_;
  int    maxIter_;
  bool   checkSymValid_;  //check symbol is proper noun /valid reln
  int    minSymAppear_;   //min times symbol appear to be compared
  int    maxShared_;      //max #clusters that share a prop
  double threshRPerPair_; //loglike thresh for R() per pair of objs to be merged
  double llThreshold_;    //loglike improvement threshold
  bool   printUnit_;      //print unit clusters
  bool   saveUnit_;       //save unit clusters in output file
  Array<StringSet> symSets_;
  double defaultFalseProb_;
  bool   pureGreedy_;
};

#endif
