#ifndef RELATION_H_OCT_06_2008
#define RELATION_H_OCT_06_2008

#include "util.h"
#include "array.h"
#include "arraysaccessor.h"
#include "hashint.h"
#include "hashstring.h"
#include "rbtree.h"
#include "type.h"
#include "comb.h"
#include "hashtuple.h"
#include "canopy.h"

const double FALSE   = 0.0;
const double TRUE    = 1.0;
const double UNKNOWN = 2.0;


enum { STATE_NONE=0, STATE_N0_NULL_ONLY=1, STATE_N1_NULL_ONLY=2,
       STATE_N0_N1_NULL=3, STATE_N0_EQ_N1=4,
       STATE_N0_BEFORE_N1=5, STATE_N0_AFTER_N1=6 };

typedef RBTree<Comb*,CombComp> CombTree;
typedef CombTree::RBNode CombNode;
typedef hash_map<string, Array<Comb*>*, HashString, EqualString> StringToCombArrMap;


class Relation
{
 public:
  Relation() : id_(-1), closedWorld_(true), betaTrue_(-1.0), betaTotal_(-1.0),
               defaultFalseProb_(0.99999) {}

  Relation(const int& id, const string& name, const bool& closedWorld,
           const double& betaTrue, const double& betaTotal,
           const Array<Type*>& types)
    : id_(id), name_(name), closedWorld_(closedWorld),
      betaTrue_(betaTrue), betaTotal_(betaTotal), defaultFalseProb_(0.99999)
      { types_.append(types); }

  ~Relation()
  {
    deleteTrueTuples();
    deleteUnknownTuples();
    deleteClustIdToCombs();
    deleteCanopyTree();
  }

  int id() const            { return id_; }
  void setId(const int& id) { id_ = id;   }

  string name() const           { return name_; }
  void setName(const string& n) { name_ = n;    }

  bool isClosedWorld() const         { return closedWorld_; }
  void setClosedWorld(const bool& b) { closedWorld_ = b;    }

  double betaTotal() const           { return betaTotal_; }
  void setBetaTotal(const double& d) { betaTotal_ = d;    }

  Type* type(const int& index) const { return types_[index]; }
  const Array<Type*>& types() const  { return types_; }
  int arity() const                  { return types_.size(); }

  double defaultFalseProb() const    { return defaultFalseProb_; }
  void setDefaultFalseProb(double p) { defaultFalseProb_ = p; }

  double betaTrue() const            { return betaTrue_; }
  void setBetaTrue(const double& d)  { betaTrue_ = d;    }
  void setBetaTrue()
  {
    double total = 1.0;
    for (int i = 0; i < types_.size(); i++)
     total *= ((double)types_[i]->numSymbols());
    betaTrue_ = trueTuples_.size()/total * betaTotal_;
    if (betaTrue_ == betaTotal_) betaTrue_ = 0.99999 * betaTotal_;
  }

  double computeLoglike(const Array<Cluster*>& clusterComb) const
  {
    double numTrue, numFalse, numUnknown;
    computeTrueFalseTuples(clusterComb, numTrue, numFalse, numUnknown);
    return computeLoglike(numTrue,numFalse);
  }

  //compute loglikelihood contributions of all combinations containing clust
  double computeLoglike(const int& dim, const Cluster* const& clust) const
  {
    CombTree* combTree = clustIdToCombs_[dim][ clust->id() ];
    CombNode* node = combTree->head();
    double ll = 0.0;

    while (node)
    {
      Comb& comb = *node->data();
      double total = 1.0;
      for (int i = 0; i < comb.size(); i++)
        total *= comb[i]->size();

      double numTrue = comb.numTrueTuples();
      double numFalse = total - numTrue;
      ll += computeLoglike(numTrue, numFalse);

      node = node->next();
    }

    return ll;
  }

  double computeLoglike(const double& numTrue, const double& numFalse) const
  {
    assert(numTrue  >= 0.0);
    assert(numFalse >= 0.0);

    double total = numTrue + numFalse;
    double betaFalse = betaTotal_ - betaTrue_;
    double ll =  numTrue  * log( (numTrue +betaTrue_) / (total+betaTotal_) )
               + numFalse * log( (numFalse+betaFalse) / (total+betaTotal_) );

    assert(ll <= 0.0);
    return ll;
  }

  void computeTrueFalseTuples(const Array<Cluster*>& clusterComb,
                              double& numTrue, double& numFalse,
                              double& numUnknown) const
  {
    ArraysAccessor<Symbol*> acc;
    for (int i = 0; i < clusterComb.size(); i++)
      acc.appendArray( (Array<Symbol*>*) &(clusterComb[i]->symbols()) );

    numTrue = 0.0, numFalse = 0.0, numUnknown = 0.0;
    while(acc.hasNextCombination())
    {
      Symbol* sym;
      Array<int> symIds( clusterComb.size() );
      while (acc.nextItemInCombination(sym))
        symIds.append(sym->id());
      double value = tupleValue(symIds);

      if (value == TRUE)         numTrue++;
      else if (value == FALSE)   numFalse++;
      else if (value == UNKNOWN) numUnknown++;
      else if (0.0 < value && value < 1.0) { numTrue+=value; numFalse+=1.0-value;}
      else Util::exit("1. undefined tuple value.",-1);
    }
  }

  void deleteTrueTuples()
  {
    HashTuple::setSize( types_.size() );
    TupleMap::iterator it = trueTuples_.begin();
    while (it != trueTuples_.end())
    {
      const int* tuple = (*it).first;
      it++;
      delete [] tuple;
    }
  }

  void deleteUnknownTuples()
  {
    HashTuple::setSize( types_.size() );
    TupleMap::iterator it = unknownTuples_.begin();
    while (it != unknownTuples_.end())
    {
      const int* tuple = (*it).first;
      it++;
      delete [] tuple;
    }
  }

  double tupleValue(const Array<int>& symIds) const
  {
    HashTuple::setSize( types_.size() );
    TupleMap::const_iterator it = trueTuples_.find( symIds.getItems() ) ;
    if (it == trueTuples_.end()) return FALSE;
    return (*it).second;
  }

  void setTupleValue(const Array<int>& symIds, double& value)
  { bool noop; setTupleValue(symIds, value, noop); }

  void setTupleValue(const Array<int>& symIds, double& value,
                     bool& trueTupleExists)
  {
    assert(symIds.size() == types_.size());
    assert(0.0 <= value && value <= 1.0);
    trueTupleExists = false;

    HashTuple::setSize( symIds.size() );
    TupleMap::iterator it = trueTuples_.find( symIds.getItems() );
    if (value == FALSE)
    {
      if (it != trueTuples_.end())
      {
        const int* symbolIds = (*it).first;
        trueTuples_.erase(it);
        delete [] symbolIds;
        trueTupleExists = true;
      }
    }
    else //0.0 < value <= 1.0
    {
      if (it != trueTuples_.end())
      {
        (*it).second = value;
        trueTupleExists = true;
      }
      else
      {
        int* tuple = new int[ symIds.size() ];
        for (int i = 0; i < symIds.size(); i++) tuple[i] = symIds[i];
        trueTuples_[tuple] = value;
      }
    }
  }

  bool isUnknownTuple(const Array<int>& symIds)
  {
    HashTuple::setSize( types_.size() );
    TupleMap::const_iterator it = unknownTuples_.find( symIds.getItems() ) ;
    if (it == unknownTuples_.end()) return false;
    return true;
  }

  bool addTupleToCombs(const Array<int>& symIds, const double& value,
                       const bool& addTupleIfCombExists)
  {
    return addTupleToCombs(symIds.getItems(), symIds.size(), value,
                           addTupleIfCombExists);
  }

  //returns whether tuple is inserted
  bool addTupleToCombs(const int* const& symIds, const int& size,
                       const double& value, const bool& addTupleIfCombExists)
  {
    if (value == FALSE || value == UNKNOWN) return false;
    assert(0.0 <= value && value <= 1.0);

    //get cluster combination of symbols in tuple
    assert(size == types_.size());
    Array<Cluster*> clustComb( types_.size() );
    for (int d = 0; d < types_.size(); d++)
    {
      const Symbol* sym = types_[d]->symbol( symIds[d] );
      Cluster* clust = sym->cluster();
      clustComb.append(clust);
    }

    //add cluster combination to comb trees and update number of true tuples
    Comb* comb = new Comb(clustComb, value);

    Array<const Symbol*> symbols(types_.size());
    for (int d = 0; d < types_.size(); d++)
      symbols.append( types_[d]->symbol( symIds[d] ) );

    bool inserted = false;
    CombNode* existingNode = NULL;
    for (int d = 0; d < clustIdToCombs_.size(); d++)
    {
      Array<CombTree*>& clustIdToCombs = clustIdToCombs_[d];
      Cluster* clust = clustComb[d];
      if (clustIdToCombs.size() <= clust->id())
        clustIdToCombs.growToSize(clust->id()+1, NULL);
      CombTree*& combTree = clustIdToCombs[ clust->id() ];
      if (combTree == NULL) combTree = new CombTree;

      if (combTree->insert(comb, existingNode)) //if comb inserted into tree
        inserted = true;
      else
      { Util::assertt(d==0, "combTree d != 0", -1); break; }
    }

    //if comb is not inserted into any tree because it is already in tree
    if (!inserted)
    {
      /*cout << "duplicate tuple " << name_ << "(";
      for (int i = 0; i < comb->size(); i++)
      {
        cout << (*comb)[i]->symbol(0)->name();
        if (i == comb->size()-1) cout << ")";
        else                     cout << ",";
      }
      cout << endl;//*/

      delete comb;
      if (addTupleIfCombExists)
      {
        Comb* comb = existingNode->data();
        comb->addNumTrueTuples(value);
      }
    }

    return inserted;
  }//addTupleToCombs()

  void setCombUnknownTuples(const Array<int>& symIds)
  {
    bool clustCombExists = true;
    Array<Cluster*> clustComb( types_.size() );
    for (int d = 0; d < types_.size(); d++)
    {
      const Symbol* sym = types_[d]->symbol( symIds[d] );
      Cluster* clust = sym->cluster();
      if (clust == NULL) { clustCombExists = false; break; }
      clustComb.append(clust);
    }
    if (!clustCombExists) return;

    Comb unknownComb(clustComb, UNKNOWN);
    Array<CombTree*>& clustIdToCombs = clustIdToCombs_[0];
    CombTree* combTree = clustIdToCombs[ unknownComb[0]->id() ];
    if (combTree == NULL) return;
    Comb* foundComb = NULL;
    combTree->find((Comb*)&unknownComb, foundComb);
    if (foundComb)
    {
      foundComb->setNumUnknown( 1.0 );
      Util::assertt(foundComb->numTrueTuples()==1.0,
                    "foundComb->numTrueTuples() != 1.0",
                    foundComb->numTrueTuples(), -1);
      //cout << "setting unknownComb...";
      //for (int i = 0; i < unknownComb.size(); i++)
      //  cout << unknownComb[i]->member(0)->name() << " ";
      //cout << endl;
    }
  }

  void addUnknownTuple(const Array<int>& symIds)
  {
    HashTuple::setSize( symIds.size() );
    TupleMap::iterator it = unknownTuples_.find( symIds.getItems() );
    if (it == trueTuples_.end())
    {
      int* tuple = new int[ symIds.size() ];
      for (int i = 0; i < symIds.size(); i++) tuple[i] = symIds[i];
      unknownTuples_[tuple] = UNKNOWN;
    }
  }

  void printTupleValues(ostream& out) const
  {
     HashTuple::setSize( types_.size() );
     TupleMap::const_iterator it = trueTuples_.begin();
     for (; it != trueTuples_.end(); it++)
     {
       cout << name_ << "(";
       const int* tuple = (*it).first;
       double value     = (*it).second;
       for (int i = 0; i < types_.size(); i++)
       {
         cout << types_[i]->symbol( tuple[i] )->name();
         cout << (i<types_.size()-1?",":")");
       }
       if (value == TRUE) cout << "  TRUE" << endl;
       else               cout << "  " << value << endl;
     }

     it = unknownTuples_.begin();
     for (; it != unknownTuples_.end(); it++)
     {
       cout << name_ << "(";
       const int* tuple = (*it).first;
       double value     = (*it).second;
       for (int i = 0; i < types_.size(); i++)
       {
         cout << types_[i]->symbol( tuple[i] )->name();
         cout << (i<types_.size()-1?",":")");
       }
       if (value == UNKNOWN) cout << "  UNKNOWN" << endl;
       else               cout << "  " << value << endl;
     }
   }

  void checkNumTrueTuples() const
  {
    for (int d = 0; d < clustIdToCombs_.size(); d++)
    {
      Array<CombTree*>& clustIdToCombs = clustIdToCombs_[d];
      for (int c = 0; c < clustIdToCombs.size(); c++)
        checkNumTrueTuples(d, c);
    }
  }

  void checkNumTrueTuples(const int& dim, Cluster* const& cluster) const
  { checkNumTrueTuples(dim, cluster->id()); }

  void checkNumTrueTuples(const int& dim, const int& clustId) const
  {
    CombTree* combTree = clustIdToCombs_[dim][clustId];

    CombNode* node = combTree->head();
    while(node)
    {
      Comb& comb = *node->data();
      double total = 1.0;
      ArraysAccessor<Symbol*> acc;
      for (int d = 0; d < comb.size(); d++)
      {
        acc.appendArray( (Array<Symbol*>*) &( comb[d]->symbols()));
        total *= comb[d]->size();
      }

      double numTrue  = comb.numTrueTuples();
      double numFalse = total - numTrue;

      double nTrue = 0.0, nFalse = 0.0;
      countNumTuples(acc, nTrue, nFalse);

      if (numTrue != nTrue || numFalse != nFalse)
      {
        cout << "ERROR" << endl;
        cout << "comb=" << &comb << endl;
        comb.printFull(cout); cout << endl;
        cout << "dim = " << dim << endl;
        Cluster& cluster = *comb[dim];
        cout << "cluster = ";
        for (int i = 0; i < cluster.size(); i++)
          cout << cluster[i]->name() << " ";
        cout << endl;
        cout << "numTrue =" << numTrue << endl;
        cout << "nTrue   =" << nTrue << endl;
        cout << "numFalse=" << numFalse << endl;
        cout << "nFalse  =" << nFalse << endl;
        Util::assertt(numTrue == nTrue && numFalse == nFalse, "wrong num true/false tuples",-1);
      }

      node = node->next();
    }
  }

  void countNumTuples(ArraysAccessor<Symbol*>& acc,
                      double& numTrue, double& numFalse) const
  {
    while(acc.hasNextCombination())
    {
      Symbol* sym;
      Array<int> symIds( acc.numItemsInCombination() );
      while (acc.nextItemInCombination(sym))
        symIds.append(sym->id());

      double value = tupleValue(symIds);
      if (value == TRUE)                 numTrue  += 1.0;
      else if (value == FALSE)           numFalse += 1.0;
      else if (0.0<value && value<1.0) { numTrue += value; numFalse += 1.0-value;}
      else Util::exit("2. undefined tuple value.",-1);
    }
  }

  void createClustIdToCombs()
  {
    assert(types_.size() > 0);
    clustIdToCombs_.growToSize( types_.size() );
    for (int i = 0; i < types_.size(); i++)
      clustIdToCombs_[i].growToSize( types_[i]->numClusters(), NULL );
  }

  void deleteClustIdToCombs()
  {
    //delete the combs in the trees of the first dimension
    Array<CombTree*>& clustIdToCombs = clustIdToCombs_[0];
    for (int c = 0; c < clustIdToCombs.size(); c++)
    {
      CombTree* combTree = clustIdToCombs[c];
      if (combTree == NULL) continue; // NULL because size before reinit is larger
      CombNode* node = combTree->head();
      while (node)
      {
        Comb* comb = node->data();
        node = node->nextNode(true);
        delete comb;
      }
    }

      //delete comb trees in all dimensions
    for (int d = 0; d < clustIdToCombs_.size(); d++)
    {
      Array<CombTree*>& clustIdToCombs = clustIdToCombs_[d];
      for (int c = 0; c < clustIdToCombs.size(); c++)
      {
        CombTree* combTree = clustIdToCombs[c];
        delete combTree;
        clustIdToCombs[c] = NULL;
      }
      clustIdToCombs.clear();
    }
  }

  void printClustIdToCombs(ostream& out) const
  {
    for (int d = 0; d < clustIdToCombs_.size(); d++)
    {
      Array<CombTree*>& clustIdToCombs = clustIdToCombs_[d];
      for (int c = 0; c < clustIdToCombs.size(); c++)
      {
        out << endl << "\tDIM " << d << " , CLUST " << c << endl;
        CombTree* combTree = clustIdToCombs[c];
        if (combTree == NULL) { out << "NULL" << endl; continue; }
        CombNode* node = combTree->head();
        int cnt = 0;
        while (node)
        {
          Comb* comb = node->data();
          const Array<Cluster*>& clustComb = comb->comb();
          out << "\tCLUSTER COMBINATION " << cnt++
               << "  numTrueTuples=" << comb->numTrueTuples() << endl;
          for (int i = 0; i < clustComb.size(); i++)
          {
            Cluster* clust = clustComb[i];
            out << "clust " << clust->id() << " ";
            out << endl;
            for (int j = 0; j < clust->size(); j++)
              out << "\t" << j << " : " << (*clust)[j]->name() << endl;
          }
          node = node->nextNode(true);
        }
      }
    }
  }

  //idx is the index of cluster combinations
  void printCombs(ostream& out, int& idx)
  {
    out << "#START_RELATION " << name_ << endl;
    Array<CombTree*>& clustIdToCombs = clustIdToCombs_[0];
    for (int c = 0; c < clustIdToCombs.size(); c++)
    {
      CombTree* combTree = clustIdToCombs[c];
      if (combTree == NULL) continue;
      CombNode* node = combTree->head();
      while (node)
      {
        Comb* comb = node->data();
        const Array<Cluster*>& clustComb = comb->comb();
        out << "  #START_COMB"
            << "  #IDX " << idx++
            << "  #CLUSTERS " << clustComb.size()
            << "  #NUM_TRUE_TUPLES " << comb->numTrueTuples() << endl;
        for (int i = 0; i < clustComb.size(); i++)
        {
          Cluster* clust = clustComb[i];
          out << "    #START_CLUSTER " << i << "  CLUSTER_ID " << clust->id()
              << endl;
          for (int j = 0; j < clust->size(); j++)
            out << "    " << (*clust)[j]->name() << endl;
          out << "    #END_CLUSTER" << endl;
        }
        out << "  #END_COMB" << endl;


        out << "  #START_TRUE_TUPLES" << endl;
        ArraysAccessor<Symbol*> acc;
        for (int i = 0; i < clustComb.size(); i++)
          acc.appendArray(&(clustComb[i]->symbols()));
        while (acc.hasNextCombination())
        {
          Array<int> symIds( types_.size() );
          Symbol* sym;
          while (acc.nextItemInCombination(sym))
            symIds.append(sym->id());

          if (tupleValue(symIds) == TRUE)
          {
            out << "  " << name_ << "(" << types_[0]->symbol(symIds[0])->name();
            for (int j = 1; j < types_.size(); j++)
              out << "," << types_[j]->symbol(symIds[j])->name();
            out << ")" << endl;
          }
        }
        out << "  #END_TRUE_TUPLES"  << endl;

        node = node->nextNode(true);
      }
    }
    out << "#END_RELATION " << endl;
  }

  void checkOrderInCombTrees()
  {
    for (int d = 0; d < clustIdToCombs_.size(); d++)
    {
      Array<CombTree*>& clustIdToCombs = clustIdToCombs_[d];
      for (int c = 0; c < clustIdToCombs.size(); c++)
      {
        CombTree* tree = clustIdToCombs[c];

        Array<CombNode*> nodes(  tree->numItems() );
        inorderCombTree(nodes, tree->root());
        Util::assertt(nodes.size() == tree->numItems(),
                      "nodes.size!=tree.numItems", nodes.size(),
                      tree->numItems(), -1);
        for (int i = 0; i < nodes.size(); i++)
          if (i > 0) assert(compare(nodes[i-1], nodes[i], d) <= 0);//check nodes

        assert(nodes.size() == tree->size());

        int idx = -1;
        CombNode* node = tree->head();
        while(node)
        {
          ++idx;
          assert(node == nodes[idx]);
          if (node->next()) assert(compare(node, node->next(), d) <= 0);
          node = node->next();
        }
      }
    }
  }

  private: void inorderCombTree(Array<CombNode*>& nodes, CombNode* const& node)
  {
    if (node)
    {
      inorderCombTree(nodes, node->link_[0]);
      nodes.append(node);
      inorderCombTree(nodes, node->link_[1]);
    }
  }

 public:
  const Array<Array<CombTree*> >& clustIdToCombs() const
  { return clustIdToCombs_; }

  Array<CombTree*>& clustIdToCombs(const int& idx) const
  { return clustIdToCombs_[idx]; }

  void printCombTree(const CombTree* const & combTree, ostream& out) const
  {
    CombNode* node = combTree->head();
    int cnt = -1;
    while (node)
    {
      Comb& comb = *node->data();

      out << "COMB " << ++cnt << " : ";
      for (int d = 0; d < comb.size(); d++)
        out << comb[d] << " , ";
      out << endl;

      node = node->next();
    }
  }

  ostream& print(ostream& out) const
  {
    out << id_ << "  ";
    out << name_ << "(";
    for (int t = 0; t < types_.size()-1; t++)
      out << types_[t]->name() << ",";
    out << types_.lastItem()->name() << ")";

    return out;
  }

  void printComb(ostream& out, const char* nodeName, CombNode* const & node)
    const { printComb(out, nodeName, *node->data()); }

  void printComb(ostream& out, const char* nodeName, const Comb& comb) const
  {
    out << nodeName << ": ";
    for (int i = 0; i < comb.size(); i++)
    {
      Cluster& clust = *(comb[i]);
      for (int j = 0; j < clust.size(); j++)
        out << clust[j]->name() << " ";
      out << " | ";
    }
    out << endl;
  }

  ///////////////////////////////// MERGE ///////////////////////////////////

 public:
  //overlapIndexes: type indexes for which combinations containing clustId0
  //                and clustId1 overlap
  void overlapCombs(const Array<int>& typeIndexes,
                    const int& clustId0, const int& clustId1,
                    Array<int>& overlapIndexes, Array<int>& noOverlapIndexes)
  {
    //for each position type appears
    for (int i = 0; i < typeIndexes.size(); i++)
    {
      int idx = typeIndexes[i];
      CombTree* combTree0 = clustIdToCombs_[idx][clustId0];
      CombTree* combTree1 = clustIdToCombs_[idx][clustId1];
      if (combTree0 && combTree1)
      {
        if (overlap(combTree0, combTree1, idx))  overlapIndexes.append(idx);
        else                                     noOverlapIndexes.append(idx);
      }
      else if (combTree0 || combTree1)           noOverlapIndexes.append(idx);
    }
  }

  bool overlap(const CombTree* const& ct0, const CombTree* const& ct1,
              const int& ignoreDim)
  {
    if (ct0 == NULL || ct1 == NULL) return false;
    if (ct0->head() == NULL || ct1->head() == NULL) return false;
    if (   (compare(ct0->tail(), ct1->head(), ignoreDim) <0)
        || (compare(ct1->tail(), ct0->head(), ignoreDim) <0) )
      return false;
    return true;
  }

  //returns the change in loglikelihood in merging clusters
  //the trees of clustId0 and clust1 in dim must overlap
  //numTrue is the number of common true tuples
  //NOTE: if checkChange and returns early, numTrue may be incorrect
  double tmpMergeClusters(const int& dim, int clustId0, int clustId1,
                          const double& bestChange, const bool& checkChange,
                          int& numOverlapCombs)
  {
    double change = 0.0, tmpChange;
    numOverlapCombs = 0;

    CombTree* combTree0 = clustIdToCombs_[ dim ][ clustId0 ];
    CombTree* combTree1 = clustIdToCombs_[ dim ][ clustId1 ];

    assert(overlap(combTree0, combTree1, dim));
    bool forward; //direction to iterate through combTree0's and 1's lists
    bool sswitch = switchTrees(combTree0, combTree1, dim, forward);
    if (sswitch) {int tmpp = clustId0;  clustId0 = clustId1;  clustId1 = tmpp; }

    CombNode* n0 = forward ? combTree0->head() : combTree0->tail();  assert(n0);
    CombNode* n1 = forward ? combTree1->head() : combTree1->tail();  assert(n1);
    if (forward) assert( compare(n0->data(), n1->data(), dim) <= 0 );
    else         assert( compare(n0->data(), n1->data(), dim) >= 0 );

    Comb*    comb0 = n0->data(),          * comb1    = n1->data();
    Cluster* clust0 = (*comb0)[dim],      * clust1   = (*comb1)[dim];
    double   clust0Size = clust0->size(), clust1Size = clust1->size();//*/

    int state = getState(n0, n1, dim, forward);

    while (state != STATE_N0_N1_NULL)
    {
      //if (n0 != NULL) printComb(cout, "N0", n0);
      //if (n1 != NULL) printComb(cout, "N1", n1);

      switch (state)
      {
        case STATE_N0_EQ_N1:
        {
          tmpChange = loglikeChange(n0->data(), n1->data(), clust0Size,
                                    clust1Size, dim, numOverlapCombs);
          change += tmpChange;

          //cout << "\tOVERLAP: tmpChange = " << tmpChange << " "
          //     << "change = " << change << endl;

          //assert(change > 0); //commented out: not true if prob of tuple <= 0.5
          n0 = n0->nextNode(forward);
          n1 = n1->nextNode(forward);
          state = getState(n0, n1, dim, forward);
          break;
        }

        case STATE_N0_BEFORE_N1:
        case STATE_N1_NULL_ONLY:
        {
          //tmpChange=loglikeChange(n0->data(), clust0Size,
          //                        clust0Size+clust1Size, dim, ntrue);
          tmpChange = computeLoglikeChange(n0->data(), clust1, clust0Size,
                                           clust1Size, dim);
          change += tmpChange;

          //cout << "\tNO-OVERLAP: tmpChange = " << tmpChange << " "
          //     << "change = " << change << endl;;

          n0 = n0->nextNode(forward);
          state = getState(n0, n1, dim, forward);
          break;
        }

        case STATE_N0_AFTER_N1:
        case STATE_N0_NULL_ONLY:
        {
          //tmpChange=loglikeChange(n1->data(), clust1Size,
          //                        clust0Size+clust1Size, dim, ntrue);
          tmpChange = computeLoglikeChange(n1->data(), clust0, clust1Size,
                                           clust0Size, dim);
          change += tmpChange;

          //cout << "\tNO-OVERLAP: tmpChange = " << tmpChange << " "
          //     << "change = " << change << endl;;

          n1 = n1->nextNode(forward);
          state = getState(n0, n1, dim, forward);
          break;
        }
      }

      //check if change can be better than bestChange; if not, return
      if (checkChange && change < bestChange)
      {
        if (state==STATE_N0_NULL_ONLY || state==STATE_N1_NULL_ONLY)
          break;

        if (forward)
        {
          if (   (n0==combTree0->tail() && compare(n0,combTree1->tail(),dim)>0)
              || (n1==combTree1->tail() && compare(n1,combTree0->tail(),dim)>0))
            break;
        }
        else //backward
        {
          if (   (n0==combTree0->head() && compare(n0,combTree1->head(),dim)<0)
              || (n1==combTree1->head() && compare(n1,combTree0->head(),dim)<0))
            break;
        }
      }

    }//while(state != STATE_N0_N1_NULL)

    return change;
  }//tmpMergeClusters()

  //remove combs containing clust0 or clust1 in more than one position
  void removeCombsWithTwoOrMoreOfClusts(const int& dim, const int& clustId0,
                                        const int& clustId1,
                                        const Array<int>& typeDims,
                                        Array<CombNode*>& repeatedNodes,
                                        Array<CombTree*>& repeatedTrees,
                                        Array<CombNode*>& newRepNodes)
  {
    assert(typeDims.size() > 0);
    CombTree* combTree0 = clustIdToCombs_[dim][clustId0];
    CombTree* combTree1 = clustIdToCombs_[dim][clustId1];

    if (combTree0)
      removeCombsWithTwoOrMoreClustsHelper(dim, combTree0, clustId0, clustId1,
                                           typeDims, repeatedNodes, repeatedTrees,
                                           newRepNodes);
    if (combTree1)
      removeCombsWithTwoOrMoreClustsHelper(dim, combTree1, clustId0, clustId1,
                                           typeDims, repeatedNodes, repeatedTrees,
                                           newRepNodes);
  }

  void removeCombsWithTwoOrMoreClustsHelper(const int& dim,
                                            CombTree* const& combTree,
                                            const int& clustId0,
                                            const int& clustId1,
                                            const Array<int>& typeDims,
                                            Array<CombNode*>& repeatedNodes,
                                            Array<CombTree*>& repeatedTrees,
                                            Array<CombNode*>& newRepNodes)
  {
    CombNode* node = combTree->head();
    while (node)
    {
      CombNode* curNode = node;
      node = node->next();

      Comb& comb = *curNode->data();

      int firstRepeatedDim = -1;
      for (int i = 0; i < typeDims.size(); i++)
      {
        int d = typeDims[i]; assert(types_[dim] == types_[d]);
        if (d == dim) continue;
        if (comb[d]->id() == clustId0 || comb[d]->id() == clustId1)
        {
          firstRepeatedDim = d;
          break;
        }
      }
      if (firstRepeatedDim >= 0) //if has two or more clusters of clust0/clust1
      {
        combTree->tmpRemoveFromList(curNode);
        repeatedNodes.append(curNode);
        repeatedTrees.append(combTree);

        if (firstRepeatedDim > dim)  newRepNodes.append(curNode);
      }
    }
  }

  //returns the change due to combs with two or more of clust0 or clust1
  double tmpMergeClustersRepeated(const int& dim, const int& clustId0,
                                  const int& clustId1,
                                  const Array<CombNode*>& repeatedNodes,
                                  int& numOverlapCombs)
  {
    //hash combs by their non-dim clustIds,
    StringToCombArrMap idsToCombsMap;
    for (int i = 0; i < repeatedNodes.size(); i++)
    {
      Comb& comb = *repeatedNodes[i]->data();
      assert(comb.size() == types_.size());
      string idsStr = "";
      for (int d = 0; d < comb.size(); d++)
      {
        if (d == dim) continue;
        if (types_[d] == types_[dim] && (comb[d]->id() == clustId0 || comb[d]->id() == clustId1))
          idsStr += Util::intToString(clustId0) + " ";
        else
          idsStr += Util::intToString(comb[d]->id()) + " ";
      }

      //cout << "idsStr " << idsStr << endl;
      StringToCombArrMap::iterator it = idsToCombsMap.find(idsStr);
      Array<Comb*>* combArr;
      if (it == idsToCombsMap.end()) { combArr = new Array<Comb*>; idsToCombsMap[idsStr] = combArr; }
      else                           combArr = (*it).second;
      combArr->append(&comb);
    }

    //compute the changes for each group of combs
    double newClustSize =  types_[dim]->cluster(clustId0)->size()
                         + types_[dim]->cluster(clustId1)->size();
    double change = 0.0;
    numOverlapCombs = 0;
    double prevTotalFalse = 0.0;
    double decrDefFalseTup = 0.0;

    StringToCombArrMap::iterator it = idsToCombsMap.begin();
    for (; it != idsToCombsMap.end(); it++)
    {
      double prevLoglike = 0.0;
      double newTrue     = 0.0;
      double newTotal    = 1.0;

      Array<Comb*>& combArr = *((*it).second);   assert(combArr.size() > 0);
      numOverlapCombs += combArr.size() - 1;

      for (int i = 0; i < combArr.size(); i++)
      {
        Comb& comb = *(combArr[i]);

        if (i == 0) //if first comb
        {
          for (int d = 0; d < comb.size(); d++)
          {
            if (types_[d] == types_[dim] && (comb[d]->id() == clustId0 || comb[d]->id() == clustId1))
              newTotal *= newClustSize;
            else
              newTotal *= comb[d]->size();
          }
        }

        double numTrue  = comb.numTrueTuples();
        double total    = comb.numTotalTuples();
        double numFalse = total - numTrue;
        assert(numTrue > 0.0 && numFalse >= 0.0);
        double tmpPrevLoglike = computeLoglike(numTrue, numFalse);
        prevLoglike += tmpPrevLoglike;
        newTrue += numTrue;
        prevTotalFalse += numFalse;
      }

      double newFalse = newTotal - newTrue;
      double newLoglike = computeLoglike(newTrue, newFalse);  assert(newTrue > 0.0 && newFalse >= 0.0);
      double tmpChange = newLoglike - prevLoglike;
      change += tmpChange;
      decrDefFalseTup += newFalse - prevTotalFalse;

      delete &combArr;
    }

    change -= decrDefFalseTup * log(defaultFalseProb_);
    return change;
  }//tmpMergeClustersRepeated()

  static void restoreCombNodesToTrees(const Array<CombNode*>& repeatedNodes,
                                      const Array<CombTree*>& repeatedTrees)
  {
    assert(repeatedNodes.size() == repeatedTrees.size());
    for (int i = repeatedNodes.size()-1; i >= 0; i--)
      repeatedTrees[i]->restoreToList( repeatedNodes[i] );
  }

  //returns the change in loglikelihood in merging clusters
  //NOTE: if checkChange and returns early then numTrue not correct
  double tmpMergeClustersNoOverlap(const int& dim,
                                   const int& clustId0,
                                   const int& clustId1,
                                   const double& newClustSize,
                                   const double& bestChange,
                                   const bool& checkChange)
  {
    //change can only decrease
    double tmpChange, change = 0.0;

    tmpChange = tmpMergeClustersNoOverlapHelper(dim, clustId0, clustId1,
                                                newClustSize, bestChange,
                                                checkChange);
    change += tmpChange;

    if (checkChange && change < bestChange) return change;

    tmpChange = tmpMergeClustersNoOverlapHelper(dim, clustId1, clustId0,
                                                newClustSize, bestChange,
                                                checkChange);
    change += tmpChange;

    return change;
  }

  //remove combs containing clust but don't remove the combs from clust's combTree
  void removeCombsExceptClust(const int& dim, Cluster* const & clust)
  {
    Array<CombTree*>& clustIdToCombs = clustIdToCombs_[dim];
    CombTree* combTree = clustIdToCombs[ clust->id() ];
    if (combTree == NULL) return;
    CombNode* node = combTree->head();
    while(node) //for each comb containing clust
    {
      Comb& comb = *node->data();

      for (int d = 0; d < comb.size();d++) //for each cluster in comb (not clust)
      {
        if (d != dim)
        {
          //remove comb from cluster's combTree
          Cluster* cclust = comb[d];
          Array<CombTree*>& clustIdToCombs = clustIdToCombs_[d];
          CombTree* tree = clustIdToCombs[ cclust->id() ];
          if (tree) tree->remove(&comb);
        }
      }

      node = node->next();
    }
  }

  //remove clust1's tree, merge clust1 into clust0, update numTrueTuples of
  //combs containing clust0.
  //update the numTrueTuples and numUnknownTuples in comb containing merged cluster
  void removeClust1CombsAndMerge(const int& dim, Cluster* const& clust0,
                                 Cluster* const& clust1,
                                 const Array<int>& typeDims)
  {
    Array<CombTree*>& clustIdToCombs = clustIdToCombs_[dim];
    CombTree* combTree1 = clustIdToCombs[ clust1->id() ];
    if (combTree1 == NULL) return;
    clustIdToCombs[ clust1->id() ] = NULL; //remove clust1's combs

    CombNode* node = combTree1->head();
    while(node) //for each comb containing clust1
    {
      Comb& comb = *node->data();  assert(comb[dim] == clust1);
      for (int i = 0; i < typeDims.size(); i++)
      {
        int d = typeDims[i];
        if (comb[d]==clust1) comb.setCluster(d, clust0);
      }

      //for each cluster in comb
      for (int d = 0; d < comb.size(); d++)
      {
        //add comb to cluster's comb
        Cluster* clust = comb[d];
        Array<CombTree*>& clustIdToCombs = clustIdToCombs_[d];
        if (clustIdToCombs[ clust->id() ] == NULL)
          clustIdToCombs[ clust->id() ] = new CombTree;
        CombTree* tree = clustIdToCombs[ clust->id() ];

        CombNode* node;
        bool inserted = tree->insert(&comb, node);
        if (!inserted) //if comb already in clust's tree (overlapping combs)
        {
          Util::assertt(d==0, "expect detect overlapping comb in dim 0", -1);
          Comb* existingComb = node->data();
          existingComb->addNumTrueTuples( comb.numTrueTuples() );
          existingComb->addNumUnknown( comb.numUnknown() );

          Util::assertt(existingComb->numTrueTuples() >=
                        existingComb->numUnknown(),
                        "numTrueTuples(0) < numUnknown",
                        existingComb->numTrueTuples(),
                        existingComb->numUnknown(), -1);
          delete &comb;
          break; //because comb also found in all other clusters
        }
      }//for each cluster in comb

      node = node->next();
    }
    delete combTree1;
  }//removeClust1CombsAndMerge()

  void removeNullInClustIdToCombs(const int& dim, const int& clustId)
  {
    Array<CombTree*>& clustIdToCombs = clustIdToCombs_[dim];
    assert( clustIdToCombs[ clustId ] == NULL );

    bool remLastClust = (clustId == clustIdToCombs.size()-1);
    CombTree* lastTree = clustIdToCombs.lastItem();
    clustIdToCombs.removeItemFastDisorder( clustId );

    //change the id of the cluster at position dim of each comb in lastTree
    if (lastTree != NULL)
    {
      CombNode* node = lastTree->head();
      while(node)
      {
        Comb& comb = *node->data();  
        Cluster* cluster = comb[dim];
        cluster->setId(clustId);
        node = node->next();
      }
    }

    assert(remLastClust || clustIdToCombs[ clustId ] == lastTree);
    remLastClust = false; lastTree = NULL;
  }

 private:
  //combTree0 should embed combTree1 or if intersect, combTree0 shorter
  bool switchTrees(CombTree*& combTree0, CombTree*& combTree1, const int& dim,
                   bool& forward)
  {
    bool sswitch;
    if (combTree0==NULL || combTree1==NULL || !overlap(combTree0, combTree1, dim))
    {
      sswitch = false;
      forward = true;
    }
    else //if combTree0's list is embedded in combTree1's
    if ( (compare(combTree0->head(), combTree1->head(), dim) >= 0) &&
         (compare(combTree0->tail(), combTree1->tail(), dim) <= 0) )
    {
      sswitch = true;
      forward = true;
    }
    else //if combTree1's list is embedded in combTree0's
    if ( (compare(combTree1->head(), combTree0->head(), dim) >= 0) &&
         (compare(combTree1->tail(), combTree0->tail(), dim) <= 0) )
    {
      sswitch = false;
      forward = true;
    }
    else //lists intersect
    if (combTree0->size() <= combTree1->size())
    {
      sswitch = false;
      forward = ( compare(combTree0->head(), combTree1->head(), dim) < 0 );
      if (!forward) assert(compare(combTree0->tail(),combTree1->tail(),dim)>=0);
    }
    else
    if (combTree0->size() > combTree1->size())
    {
      sswitch = true;
      forward = ( compare(combTree1->head(), combTree0->head(), dim) < 0 );
      if (!forward) assert(compare(combTree1->tail(),combTree0->tail(),dim)>=0);
    }
    else
    {
      sswitch = false;
      cout << "ERROR: unexpected trees configuration. " << endl; exit(-1);
    }

    if (sswitch) { CombTree* tmp=combTree0; combTree0=combTree1; combTree1=tmp;}
    return sswitch;
  }

  int getState(const CombNode* const& n0, const CombNode* const& n1,
               const int& dim, const bool& forward)
  {
    if (n0 == NULL && n1 != NULL) return STATE_N0_NULL_ONLY;
    if (n0 != NULL && n1 == NULL) return STATE_N1_NULL_ONLY;
    if (n0 == NULL && n1 == NULL) return STATE_N0_N1_NULL;

    int cmp = compare(n0, n1, dim);
    if (forward)
    {
      if (cmp == 0) return STATE_N0_EQ_N1;
      if (cmp < 0)  return STATE_N0_BEFORE_N1;
                    return STATE_N0_AFTER_N1;
    }
    else
    {
      if (cmp == 0) return STATE_N0_EQ_N1;
      if (cmp < 0)  return STATE_N0_AFTER_N1;
                    return STATE_N0_BEFORE_N1;
    }
  }

  //a subclass should redefine this function
  double computeLoglikeChange(const Comb* const& comb0,
                              const Cluster* const& clust1,
                              const double& clust0Size,
                              const double& clust1Size,
                              const int& dim)
  {
    return loglikeChange(comb0, clust0Size, clust0Size+clust1Size, dim);
  }

  //for overlapping cluster comb
  double loglikeChange(const Comb* const& comb0, const Comb* const& comb1,
                       const double& clust0Size, const double& clust1Size,
                       const int& dim, int& numOverlapCombs)
  {
    double numTotal = 1.0;
    for (int d = 0; d < comb0->size(); d++)
      if (d != dim) numTotal *= (*comb0)[d]->size();

    double total0   = numTotal * clust0Size;
    double total1   = numTotal * clust1Size;
    double newTotal = numTotal * (clust0Size + clust1Size);

    double true0   = comb0->numTrueTuples();
    double true1   = comb1->numTrueTuples();
    double newTrue = true0 + true1;

    double unknown0 = comb0->numUnknown();
    double unknown1 = comb1->numUnknown();
    double newUnknown = unknown0 + unknown1;

    true0   -= unknown0;
    true1   -= unknown1;
    newTrue -= newUnknown;

    if (true0 > 0.0 && true1 > 0.0) numOverlapCombs++;

    double false0   = total0   - true0   - unknown0;
    double false1   = total1   - true1   - unknown1;
    double newFalse = newTotal - newTrue - newUnknown;

    double llNew = computeLoglike(newTrue, newFalse);
    double ll0   = computeLoglike(true0, false0);
    double ll1   = computeLoglike(true1, false1);
    double change = llNew - ll0 - ll1;

    return change;
  }//loglikeChange()

  //for non-overlapping cluster comb
  double loglikeChange(const Comb* const& comb, const double& clustSize,
                       const double& newClustSize, const int& dim)
  {
    double numTotal = 1.0;
    for (int d = 0; d < comb->size(); d++)
      if (d != dim) numTotal *= (*comb)[d]->size();

    double ttotal   = numTotal * clustSize;
    double newTotal = numTotal * newClustSize;

    double ttrue   = comb->numTrueTuples();
    double newTrue = ttrue;

    double unknown = comb->numUnknown();
    double newUnknown = unknown;
    ttrue   -= unknown;
    newTrue -= newUnknown;

    double ffalse   = ttotal   - ttrue   - unknown;
    double newFalse = newTotal - newTrue - newUnknown;

    double llNew  = computeLoglike(newTrue, newFalse);
    double ll     = computeLoglike(ttrue, ffalse);
    double change = llNew - ll;

    double decrNumDefaultFalseTuples = newFalse - ffalse;

    //change -= decrNumDefaultFalseTuples * log(0.9999);
    //change -= decrNumDefaultFalseTuples * log(1.0 - 1e-15);
    change -= decrNumDefaultFalseTuples * log(defaultFalseProb_);

    return change;
  }//loglikeChange()

  //NOTE: if checkChange and returns early, numTrue may be incorrect
  double tmpMergeClustersNoOverlapHelper(const int& dim,
                                         const int& clustId0,
                                         const int& clustId1,
                                         const double& newClustSize,
                                         const double& bestChange,
                                         const bool& checkChange)
  {
    double change = 0.0, tmpChange;

    CombTree* combTree = clustIdToCombs_[ dim ][ clustId0 ];
    if (combTree == NULL) return change;
    CombNode* node = combTree->head();
    if (node == NULL) return change;

    Cluster* clust1 = types_[dim]->cluster(clustId1);
    double clust1Size = clust1->size();

    Comb* comb = node->data();
    int clust0Size = (*comb)[dim]->size();
    assert(newClustSize == clust0Size + clust1Size);

    while (node)
    {
      //tmpChange = loglikeChange(node->data(), clust0Size, newClustSize, dim);
      tmpChange = computeLoglikeChange(node->data(), clust1, clust0Size,
                                       clust1Size, dim);
      //assert(tmpChange <= 0.0); //commented out: may be very small +ve number
      change += tmpChange;

      node = node->next();
      if (checkChange && change < bestChange) return change;
    }

    return change;
  }

  int compare(const CombNode* const& cn0, const CombNode* const& cn1,
              const int& ignoreDim)
  { return Comb::compare(cn0->data(), cn1->data(), ignoreDim); }

  int compare(const Comb* const& c0, const Comb* const& c1,const int& ignoreDim)
  { return Comb::compare(c0, c1, ignoreDim); }

  //////////////////////////// AGGLOMERATIVE CLUSTERING ////////////////////////

  public: void createCanopyTree()
  {
    if (canopyTree_.size() > 0) return;
    if (types_.size() == 1) return;

    canopyTree_.growToSize( types_.size() , NULL );
    assert(canopyTree_.size() == clustIdToCombs_.size());

    //for each dim
    for (int dim = 0; dim < clustIdToCombs_.size(); dim++)
    {
      if (!types_[dim]->doCluster())
      {
        cout << "not creating canopy for type " << types_[dim]->name()
             << " because its symbols are not clustered" << endl;
        continue;
      }

      canopyTree_[dim] = new CanopyTree;
      cout << "creating canopies for dim " << dim << "..." << endl;
      createCanopyTreeHelper(dim);
    }
  }

  private: void createCanopyTreeHelper(const int& dim)
  {
    CanopyTree& canopyTree = *( canopyTree_[dim] );
    Array<CombTree*>& clustIdToCombs = clustIdToCombs_[dim];

    for (int clustId = 0; clustId < clustIdToCombs.size(); clustId++)
    {
      CombTree* combTree = clustIdToCombs[ clustId ];
      if (combTree == NULL) continue;
      CombNode* node = combTree->head();
      while(node)
      {
        Comb& comb = *node->data();
        Canopy* canopy = getCanopy(canopyTree, comb, dim);
        ClusterTree& clusters = canopy->clusters_;
        clusters.insert( comb[dim] );   assert(comb[dim]->id() == clustId);

        node = node->next();
      }
    }

    //commented out: should not remove such canopies because after clusters are
    //merged, the size of the canopies can grow
    //remove canopies that are of size 1 or less
    /*for (int d = 0; d < canopyTree_.size(); d++)
    {
      Array<Canopy*> remCanopies;
      CanopyTree* canopyTree = canopyTree_[d];
      if (canopyTree == NULL) continue;
      CanopyNode* node = canopyTree->head();
      while(node)
      {
        Canopy* canopy = node->data();
        if (canopy->clusters_.size() <= 1) remCanopies.append(canopy);
        node = node->next();
      }

      for (int i = 0; i < remCanopies.size(); i++)
      {
        Canopy* canopy = remCanopies[i];
        bool removed = canopyTree->remove(canopy);
        delete canopy;
        Util::assertt(removed,"canopy failed to be removed",-1);
      }
      if (remCanopies.size() > 0)
        cout << "removed " << remCanopies.size() << " canopies of size <= 1 for relation "
             << name_ << " dim " << d << endl;
    }//*/
  }

 public:
  Canopy* getCanopy(CanopyTree& canopyTree, const Comb& comb, const int& dim)
  {
    Canopy* newCanopy = new Canopy(comb, dim);
    CanopyNode* existingNode = NULL;
    bool inserted = canopyTree.insert( newCanopy, existingNode );
    if (!inserted) delete newCanopy;
    return existingNode->data() ;
  }

  void deleteCanopyTree()
  {
    for (int d = 0; d < canopyTree_.size(); d++)
    {
      CanopyTree* canopyTree = canopyTree_[d];
      if (canopyTree == NULL) continue;
      CanopyNode* node = canopyTree->head();
      while(node)
      {
        Canopy* canopy = node->data();
        delete canopy;
        node = node->next();
      }
    }
    canopyTree_.deleteItemsAndClear();
  }

  CanopyTree* canopyTree(const int& dim) const
  {
    if (dim >= canopyTree_.size()) return NULL;
    return canopyTree_[dim];
  }

  //add symbol to symset if it appears at least minSymAppear times
  void populateSymSets(const int& minSymAppear, Array<StringSet>& symSets)
  {
    Array<StringToIntMap> symToCntMaps;
    symToCntMaps.growToSize( clustIdToCombs_.size() );

    Array<CombTree*>& clustIdToCombs = clustIdToCombs_[0];
    for (int c = 0; c < clustIdToCombs.size(); c++)
    {
      if (clustIdToCombs[c] == NULL) continue;
      CombNode* node = clustIdToCombs[c]->head();
      while(node)
      {
        Comb& comb = *node->data();
        assert(comb.size() == symToCntMaps.size());

        for (int d = 0; d < comb.size(); d++)
        {
          Cluster& cluster = *(comb[d]);  assert(cluster.size() == 1);
          string symName = cluster[0]->name();
          StringToIntMap& symToCntMap = symToCntMaps[d];
          StringToIntMap::iterator it = symToCntMap.find( symName );
          if (it == symToCntMap.end())
          {
            symToCntMap[ symName ] = 1;
            if (1 == minSymAppear)
            {
              int typeId = types_[d]->id();
              symSets[typeId].insert(symName);
            }
          }
          else
          {
            (*it).second++;
            if ((*it).second == minSymAppear)
            {
              int typeId = types_[d]->id();
              symSets[typeId].insert(symName);
            }
          }
        }

        node = node->next();
      }
    }
  }

  ClusterTree& getClusterTree(CanopyTree& canopyTree, const Comb& comb,
                              const int& dim)
  {
    Canopy* canopy = getCanopy(canopyTree, comb, dim);
    return canopy->clusters_;
  }

  //0-A-B and 1-A-B merged to become 0-A-B i.e. remove 1-A-B
  void updateCanopyTree(const int& dim, Cluster* const& clust0,
                        Cluster* const& clust1, const Array<int>& typeDims)
  {
    if (types_.size() <= 1) return; //no canopy for arity one relation

    CombTree* combTree1 = clustIdToCombs_[ dim ][ clust1->id() ];
    if (combTree1 == NULL) return;
    CombNode* node = combTree1->head();
    while(node)
    {
      Comb& comb = *node->data(); //1-A-B

      ////////////// first dimension ///////////////
      //for each 1-A-B
      //  find *-A-B,
      //  remove 1 from *,
      //  add    0 to   *

      Canopy* foundCanopy = findCanopy(dim, comb);

      if (foundCanopy)
      {
        //remove 1 from X in X-A-B and add 0
        ClusterTree& clusters = foundCanopy->clusters_;
        bool removed = clusters.remove( clust1 );

        if (removed)
        {
          clusters.insert( clust0 );

          if (typeDims.size() > 1) //if has repeated dims
          {
            int firstRepeatedDim = -1;
            for (int d = 0; d < comb.size(); d++)
            {
              if (d == dim) continue;
              if (comb[d] == clust1) { firstRepeatedDim = d; break; }
            }
            if (firstRepeatedDim > dim) //if not update before
            {
              //remove *-A-1-B canopy, insert *-A-0-B canopy
              Canopy* remCan = removeCanopy(dim, comb);

              Util::assertt(remCan == foundCanopy, "expect remCan == foundCanopy", -1);
              Canopy* newCanopy = foundCanopy;
              Array<Cluster*>& indexClusters = newCanopy->indexClusters_;
              for (int i = 0; i < indexClusters.size(); i++)
                if (indexClusters[i] == clust1) indexClusters[i] = clust0;

              //insert canopy;
              CanopyNode* existingNode = NULL;
              bool inserted = canopyTree_[dim]->insert(newCanopy, existingNode);
              if (!inserted)
              {
                //transfer clusters to existingCanopy
                Canopy* existingCanopy = existingNode->data();
                existingCanopy->add(newCanopy->clusters_);
                delete newCanopy;
              }
            }
          }//if has repeated dims

          //commented out:
          //int numDim = comb.size();
          //for (int d = 2; d <= numDim; d++)
          //  updateCanopyTreeHelper(dim, comb, clust0, clust1, d);

        }//if (removed)

        int numDim = comb.size();
        for (int d = 2; d <= numDim; d++)
          updateCanopyTreeHelper(dim, comb, clust0, clust1, d);
        
      }//if (foundCanopy)

      node = node->next();
    }
  }

  private: void updateCanopyTreeHelper(const int& dim, const Comb& comb,
                                       Cluster* const& clust0,
                                       Cluster* const& clust1,
                                       const int& updateNthDim)
  {
    //for each 1-A-B
    //  remove 1-*-B tree
    //  add    0-*-B tree
    //  transfer contents from first * to second *

    assert(comb.size() > 1);
    assert(updateNthDim > 1);
    Array<int> indexDims( comb.size() - 1 );
    int cnt = 0;
    int canopyDim = -1;
    for (int d = 0; d < comb.size(); d++)
    {
      if (d != dim) cnt++;
      if (d == dim || cnt != updateNthDim-1) indexDims.append(d);
      else                                   canopyDim = d;
    }
    assert(canopyDim >= 0);

    //remove 1-*-B, etc.
    static Canopy target;
    Array<Cluster*>& indexClusters = target.indexClusters_;
    indexClusters.clear();
    for (int i = 0; i < indexDims.size(); i++)
      indexClusters.append( comb[ indexDims[i] ] );

    bool removed;
    Canopy* removedCanopy = NULL;
    CanopyTree* otherCanopyTree = canopyTree_[ canopyDim ];

    //if canopy not created for a type that is not clustered
    if (otherCanopyTree == NULL) { removed = false; }
    else
    {
      removed = otherCanopyTree->remove(&target, removedCanopy);

      //removed can be false because it has been removed, e.g, A-1-1-B.
      //when we look at 2nd 1, try to find canopy A-*-1-B, but it has been removed
      //when we looked at 1st 1
    }

    if (removed)
    {
      //replace all occurrences of clust1 in canopy's index with clust0
      Array<Cluster*> newIndexClusters(indexDims.size());
      for (int i = 0; i < indexDims.size(); i++)
      {
        int d = indexDims[i];
        if (comb[d] == clust1) newIndexClusters.append(clust0);
        else                   newIndexClusters.append(comb[d]);
      }

      Canopy* newCanopy = new Canopy(newIndexClusters,-1);

      CanopyNode* existingNode = NULL;
      bool inserted = otherCanopyTree->insert( newCanopy, existingNode );
      if (!inserted)
      {
        Canopy* delNewCanopy = newCanopy;
        newCanopy = existingNode->data();
        delete delNewCanopy;
      }

      newCanopy->add(removedCanopy->clusters_);//transfer clusters to new canopy
      if (newCanopy->remove(clust1)) newCanopy->add(clust0);
    }
    delete removedCanopy;


  }//updateCanopyTreeHelper

  public: void printCanopyTree(ostream& out)
  {
    int numDim = canopyTree_.size();
    for (int dim = 0; dim < numDim; dim++)
    {
      out << "----------------------- DIMENSION " << dim << endl;
      CanopyTree* canopyTree =  canopyTree_[dim];
      if (canopyTree == NULL) { out << " NULL canopyTree " << endl; continue; }
      CanopyNode* node = canopyTree->head();
      while(node)
      {
        Canopy& canopy = *node->data();
        
        //print "- A B"
        out << "---------- DIM " << dim << "  ";
        printCanopy(out, canopy, numDim, dim);
        node = node->next();
      }    
    }
  }

  private: void printCanopy(ostream& out, const Canopy& canopy, 
                          const int& numDim, const int& dim)
  {
    const Array<Cluster*>& indexClusters = canopy.indexClusters_;
    int idx = 0;
    for (int i = 0; i < numDim; i++)
    {
      if (i == dim) out << "(-)  ";
      else
      {
        Cluster& cluster = *( indexClusters[idx++] );
        out << "(" << cluster.id() << " " << cluster[0]->name() << ")  ";
      }
    }
    out << endl;
    
    const ClusterTree& clusters = canopy.clusters_ ;
    ClusterNode* intNode = clusters.head();
    while(intNode)
    {
      Cluster& cluster = *intNode->data();
      out << cluster.id() << " " << cluster[0]->name() << endl;;
      intNode = intNode->next();
    }    
  }

  private: Canopy* findCanopy(const int& dim, const Comb& comb)
  {
    static Canopy target;
    Array<Cluster*>& indexClusters = target.indexClusters_;
    indexClusters.clear();
    for (int d = 0; d < comb.size(); d++)
      if (d != dim) indexClusters.append( comb[d] );

    Canopy* foundCanopy = NULL;
    CanopyNode* foundNode = NULL;
    canopyTree_[dim]->find(&target, foundCanopy, foundNode);
    return foundCanopy;
  }

  private: Canopy* removeCanopy(const int& dim, const Comb& comb)
  {
    static Canopy target;
    Array<Cluster*>& indexClusters = target.indexClusters_;
    indexClusters.clear();
    for (int d = 0; d < comb.size(); d++)
      if (d != dim) indexClusters.append( comb[d] );

    Canopy* remCanopy = NULL;
    bool removed = canopyTree_[dim]->remove(&target, remCanopy);
    if (removed) assert(remCanopy != NULL);
    else         assert(remCanopy == NULL);
    return remCanopy;
  }

  public: bool appearInComb(const Cluster* const & clustA, const Cluster* const& clustB)
  {
    Util::assertt(types_.size() == 2 && types_[0] == types_[1], "can be called only for R(type,type)", -1);

    static Comb target;
    if (target.size() == 0) { target.appendCluster(NULL); target.appendCluster(NULL); }

    CombTree* combTree = NULL;

    target.setCluster(0, (Cluster*) clustA);
    target.setCluster(1, (Cluster*) clustB);

    combTree = clustIdToCombs_[0][clustA->id()];
    if (combTree && combTree->contains(&target)) return true;

    combTree = clustIdToCombs_[1][clustB->id()];
    if (combTree && combTree->contains(&target)) return true;

    target.setCluster(0, (Cluster*) clustB);
    target.setCluster(1, (Cluster*) clustA);

    combTree = clustIdToCombs_[0][clustB->id()];
    if (combTree && combTree->contains(&target)) return true;

    combTree = clustIdToCombs_[1][clustA->id()];
    if (combTree && combTree->contains(&target)) return true;

    return false;
  }

 private:
  int id_;
  string name_;
  bool closedWorld_;
  double betaTrue_;
  double betaTotal_;

  Array<Type*> types_; //not owned by object; do not delete

  TupleMap trueTuples_;    //store true tuples only
  TupleMap unknownTuples_; //store unknown true tuples only

  //clustIdToCombs_[dim][clustId]: tree combinations
  Array<Array<CombTree*> > clustIdToCombs_;

  //canopyTrees_[dim] : collection of cluster sets indexed by clusters in
  //all dimensions except dim
  Array<CanopyTree*> canopyTree_;

  double defaultFalseProb_;
};

inline
ostream& operator<<(ostream& out, const Relation& r) { return r.print(out); }


#endif

