#ifndef TYPE_H_OCT_02_2008
#define TYPE_H_OCT_02_2008

#include "array.h"
#include "hashstring.h"
#include "rel.h"
#include "cluster.h"
#include "symbol.h"
#include "util.h"

const int MAX_UNIT_CLUST_TO_PRINT = 10; //print unit clusters if no more than this


class Type
{
 public:
  Type() : id_(-1), name_(""), doCluster_(false), minClusters_(-1),
           maxClusters_(-1), repeatedInNaryRelation_(false) {}

  Type(const int& id, const string& name)
    : id_(id), name_(name), doCluster_(false), minClusters_(-1),
      maxClusters_(-1), repeatedInNaryRelation_(false) {}

  ~Type()
  {
    clusters_.deleteItemsAndClear();
    bestClusters_.deleteItemsAndClear();
    symbols_.deleteItemsAndClear();
    rels_.deleteItemsAndClear();
  }

  int id() const            { return id_; }
  void setId(const int& id) { id_ = id; }

  string name() const              { return name_; }
  void setName(const string& name) { name_ = name; }

  bool doCluster() const           { return doCluster_; }
  void setDoCluster(const bool& b) { doCluster_ = b ; }

  int minClusters() const             { return minClusters_; }
  void setMinClusters(const int& min) { minClusters_ = min; }

  int maxClusters() const             { return maxClusters_; }
  void setMaxClusters(const int& max) { maxClusters_ = max; }

  const Array<Cluster*>& clusters() const { return clusters_; }
  const Array<Symbol*>& symbols() const   { return symbols_; }
  const Array<Rel*>& rels() const         { return rels_; }

  int numClusters() const { return clusters_.size(); }
  int numRels() const     { return rels_.size(); }
  int numSymbols() const  { return symbols_.size(); }

  const Rel* rel(const int& idx) const { return rels_[idx]; }
  void addRel(Rel* const& rel)         { rels_.append(rel); }

  bool repeatedInNaryRelation() const { return repeatedInNaryRelation_; }

  void setRepeatedInNaryRelation();

  const Array<Relation*>* binRelationsWithSameTypes() const
  {
    if (binRelationsWithSameTypes_.empty()) return NULL;
    return &binRelationsWithSameTypes_;
  }

  void setBinRelationsWithSameTypes();

  void compress() {clusters_.compress(); symbols_.compress(); rels_.compress();}

  bool addSymbol(const string& name, int& id)
  {
    StringToIntMap::iterator it = symNameToIdMap_.find(name);
    if (it != symNameToIdMap_.end()) { id = (*it).second; return false; }
    id = symbols_.size();
    Symbol* sym = new Symbol(id, name);
    symbols_.append(sym);
    symNameToIdMap_[name] = id;
    return true;
  }

  const Symbol* symbol(const int& id) const
  {
    assert(symbols_[id]->id() == id);
    return symbols_[id];
  }

  const Symbol* symbol(const string& symName) const
  {
    StringToIntMap::const_iterator it = symNameToIdMap_.find( symName );
    assert(it != symNameToIdMap_.end());
    return symbol( (*it).second );
  }

  Cluster* addNewCluster()
  {
    if (clusters_.size() < maxClusters_)
    {
      int newClustId = clusters_.size();
      Cluster* cluster= new Cluster(newClustId);
      clusters_.append(cluster);
      return cluster;
    }
    return NULL;
  }

  Cluster* addNewCluster(Cluster* const & newClust)
  {
    if (clusters_.size() < maxClusters_)
    {
      int newClustId = clusters_.size();
      newClust->setId( newClustId );
      clusters_.append(newClust);
      return newClust;
    }
    return NULL;
  }

  //add cluster to position prevClustIdx of clusters_, move cluster originally
  //at prevClustIdx to end
  void addCluster(Cluster* const& cluster, const int& prevClustIdx)
  {
    if (prevClustIdx == clusters_.size())
    {
      clusters_.append(cluster);
      cluster->setId(prevClustIdx);
    }
    else
    {
      Cluster* clust = clusters_[prevClustIdx];
      clusters_[prevClustIdx] = cluster;
      cluster->setId( prevClustIdx );
      clust->setId(clusters_.size());
      clusters_.append(clust);
    }
  }

  Cluster* removeCluster(const int& clustId, const bool& force=false)
  {
    if (!force && clusters_.size() == minClusters_) return NULL;
    bool remLastClust = (clustId == clusters_.size()-1);
    Cluster* lastCluster = clusters_.lastItem();

    if (!remLastClust) lastCluster->setId( clustId );
    Cluster* remCluster = clusters_.removeItemFastDisorder( clustId );

    assert(remLastClust || clusters_[clustId] == lastCluster);
    return remCluster;
  }

  Cluster* removeCluster(Cluster* const& remCluster, const bool& force=false)
  {
    Cluster* cluster = removeCluster(remCluster->id(), force);
    assert(cluster == remCluster);
    return cluster;
  }

  void replaceCluster(Cluster*const& remCluster, Cluster*const& newCluster)
  {
    Util::assertt(clusters_[ remCluster->id() ] == remCluster,
                  "cannot find remCluster", -1);
    newCluster->setId( remCluster->id() );
    clusters_[ newCluster->id() ] = newCluster;
  }

  void removeAllClusters() { clusters_.deleteItemsAndClear(); }

  void setBestClusters(const Array<Cluster*>& clusters)
  {
    bestClusters_.deleteItemsAndClear();
    bestClusters_.growToSize( clusters.size() );
    for (int i = 0; i < clusters.size(); i++)
      bestClusters_[i] = new Cluster(*(clusters[i]), i);
  }

  bool checkClusterIds() const
  {
    for (int i = 0; i < clusters_.size(); i++)
      if (clusters_[i]->id() != i) return false;
    return true;
  }

  Cluster* cluster(const int& clustId) const { return clusters_[clustId]; }

  void sortClustersAndRenumberIds(const bool& largestFirst)
  {
    for (int c = 0; c < clusters_.size(); c++)
    {
      Array<Symbol*>& syms = clusters_[c]->symbols();
      sortSymbols(syms, 0, syms.size()-1);
    }
    sort(clusters_, 0, clusters_.size()-1, largestFirst);
    for (int c = 0; c < clusters_.size(); c++)
      clusters_[c]->setId(c);
  }

  void initClusters()
  {
    if (minClusters_ == -1) minClusters_ = 1;
    if (maxClusters_ == -1 || maxClusters_ > symbols_.size())
      maxClusters_ = symbols_.size();
    assert(minClusters_ <= maxClusters_);

    int numClust = maxClusters_;

    clusters_.growToSize(numClust);
    for (int i = 0; i < numClust; i++)
      clusters_[i] = new Cluster(i);

    //randomly one assigned symbol to each cluster
    Array<Symbol*> tmpSym(symbols_);
    for (int i = 0; i < clusters_.size(); i++)
    {
      Symbol* s = tmpSym.removeItemFastDisorder( random() % tmpSym.size() );
      s->setCluster( clusters_[i] );
      clusters_[i]->addSymbol(s);
    }

    //randomly assigned remaining symbols to clusters
    for (int i = 0; i < tmpSym.size(); i++)
    {
      Cluster* cluster = clusters_[ random() % clusters_.size() ];
      tmpSym[i]->setCluster(cluster);
      cluster->addSymbol(tmpSym[i]);
    }

    setBestClusters(clusters_);
  }

  void initClusters(const Array<Array<string>*>& initClusts)
  {
    StringSet unseenSymNames; //keep track of symbols that are not in bestClusts
    for (int i = 0; i < symbols_.size(); i++)
      unseenSymNames.insert( symbols_[i]->name() );
    Util::assertt((int)unseenSymNames.size()==symbols_.size(),"wrong #sym",-1);

    clusters_.deleteItemsAndClear();

    //for each initial cluster
    for (int c = 0; c < initClusts.size(); c++)
    {
      int clustId = clusters_.size();
      Cluster* newCluster = new Cluster( clustId );
      clusters_.append( newCluster );

      Array<string>& initClust = *( initClusts[c] );
      for (int i = 0; i < initClust.size(); i++) //for each sym in init cluster
      {
        //remove symbol from unseen symbols,
        string symName = initClust[i];
        StringSet::iterator it = unseenSymNames.find( symName );
        Util::assertt(it != unseenSymNames.end(), "sym not found", symName,-1);
        unseenSymNames.erase(it);

          //find the symbol
        StringToIntMap::iterator iit = symNameToIdMap_.find( symName );
        Util::assertt(iit !=symNameToIdMap_.end(),"sym not found2",symName,-1);

        int symId = (*iit).second;
        Symbol* symbol = symbols_[symId];  assert(symbol);

        newCluster->addSymbol(symbol);  //add symbol to cluster
        symbol->setCluster(newCluster); //point symbol to cluster
      }
    }

    Util::assertt(clusters_.size() == initClusts.size(), "wrong clustSize",-1);

    //create unit clusters for unseen symbols
    StringSet::iterator it = unseenSymNames.begin();
    for (; it != unseenSymNames.end(); it++)
    {
      string symName = (*it);
      StringToIntMap::iterator iit = symNameToIdMap_.find( symName );
      Util::assertt(iit != symNameToIdMap_.end(), "sym not found3",symName,-1);
      int symId = (*iit).second;
      Symbol* symbol = symbols_[symId];  assert(symbol);

      Cluster* newCluster = new Cluster( clusters_.size() );
      newCluster->addSymbol(symbol);
      symbol->setCluster(newCluster);
      clusters_.append(newCluster);
    }

    setBestClusters(clusters_);
  }//initClusters

  void printSymbols(ostream& out) const
  {
    for (int i = 0; i < symbols_.size(); i++)
      out << *(symbols_[i]) << endl;
  }

  void printRels(ostream& out) const;

  void printClusters(ostream& out, const bool& printUnit, int& numNonUnit,
                     const bool& sortClusts, const bool& largestFirst) const
  { printClusters(out, clusters_, printUnit, numNonUnit, sortClusts,
                  largestFirst); }

  void printBestClusters(ostream& out, const bool& printUnit, int& numNonUnit,
                         const bool& sortClusts, const bool& largestFirst) const
  { printClusters(out, bestClusters_, printUnit, numNonUnit, sortClusts,
                  largestFirst); }

  ostream& print(ostream& out) const
  {
    out << id_ << "  " << name_ << " " << (doCluster_?"true":"false") << " "
        << minClusters_ << " " << maxClusters_;
    return out;
  }

 private:
  void printClusters(ostream& out, const Array<Cluster*>& clusters,
                     const bool& printUnit, int& numNonUnit,
                     const bool& sortClusts, const bool& largestFirst) const
  {
    //print unit clusters if no more than this number
    static int maxNum = MAX_UNIT_CLUST_TO_PRINT;

    if (sortClusts)
    {
      Array<Cluster*>* clusts;
      Array<Cluster*> unitClusts( maxNum+1 );
      if (printUnit) clusts = new Array<Cluster*>( clusters );
      else
      {  //not printing unit so only copy the non-unit clusters
        clusts = new Array<Cluster*>( 1000 );
        for (int i = 0; i < clusters.size(); i++)
        {
          if (clusters[i]->size() > 1)           clusts->append( clusters[i]);
          else if (unitClusts.size() < maxNum+1) unitClusts.append( clusters[i]);
        }

        if (unitClusts.size() <= maxNum)
        {
          for (int j = 0; j  < unitClusts.size(); j++)
            clusts->append( unitClusts[j] );
        }
      }

      sort(*clusts, 0, clusts->size()-1, largestFirst);
      int cnt = 0;
      numNonUnit = 0;
      for (int i = 0; i < clusts->size(); i++)
      {
        printCluster(out, (*clusts)[i], cnt++);        
        if ((*clusts)[i]->size() > 1) numNonUnit++;
      }
      delete clusts;
    }
    else
    {  //not sorting clusters
      numNonUnit = 0;
      int cnt = 0;
      Array<Cluster*> unitClusts( maxNum+1 );
      for (int i = 0; i < clusters.size(); i++)
      {
        Cluster* cluster = clusters[i];
        if (cluster->size() > 1)
        {
          printCluster(out, cluster, cnt++);
          numNonUnit++;
        }
        else
        {
          if (printUnit)                         printCluster(out,cluster,cnt++);
          else if (unitClusts.size() < maxNum+1) unitClusts.append( cluster );
        }
      }
      if (unitClusts.size() <= maxNum)
      {
        for (int j = 0; j < unitClusts.size(); j++)
          printCluster(out, unitClusts[j], cnt++);
      }
    }
  }

  void printCluster(ostream& out, Cluster* const& cluster, const int& cnt)const
  {
    out << " CLUS " << cluster->id() << " , cnt " << cnt << endl;
    for (int j = 0; j < cluster->size(); j++)
      out << id_ << "  " << j << " : " << (*cluster)[j]->name() << endl;
    out << endl;
  }

  static void sort(Array<Cluster*>& clusters, const int& l, const int& r,
                   const bool& largestFirst)
  {
    if (l >= r) return;

    Cluster** items = (Cluster**) clusters.getItems();
    Cluster* tmp = items[l];
    items[l] = items[(l+r)/2];
    items[(l+r)/2] = tmp;

    int last = l;
    for (int i = l+1; i <= r; i++)
    {
      //find the two clusters' sizes
      int isize = items[i]->size(), lsize = items[l]->size();

      //compare the two clusters
      bool cmp = false;
      bool eqSize = false;
      if (largestFirst)
      {
        if (isize > lsize)       cmp = true;
        else if (isize < lsize)  cmp = false;
        else                     eqSize = true;
      }
      else
      {
        if (isize < lsize)       cmp = true;
        else if (isize > lsize)  cmp = false;
        else                     eqSize = true;
      }
      if (eqSize) //if clusters have same size, then compare their contents
      {
        Cluster& iclust = *( items[i] );
        Cluster& lclust = *( items[l] );
        cmp = false;

        string iname, lname;
        iname = iclust[0]->name();
        lname = lclust[0]->name();

        int cf = iname.compare( lname );
        if (cf == -1)     cmp = true;
        else if (cf == 1) cmp = false;
      }

      if (cmp)
      {
        ++last;
        Cluster* tmp = items[last];
        items[last] = items[i];
        items[i] = tmp;
      }
    }

    tmp = items[l];
    items[l] = items[last];
    items[last] = tmp;
    sort(clusters, l, last-1, largestFirst);
    sort(clusters, last+1, r, largestFirst);
  }

  static void sortSymbols(Array<Symbol*>& symbols, const int& l, const int& r)
  {
    Symbol** items = (Symbol**) symbols.getItems();

    if (l >= r) return;
    Symbol* tmp = items[l];
    items[l] = items[(l+r)/2];
    items[(l+r)/2] = tmp;

    int last = l;
    for (int i = l+1; i <= r; i++)
    {
      Symbol* isym  = items[i];
      string  iname = isym->name();
      Symbol* lsym  = items[l];
      string  lname = lsym->name();

      if (iname.compare( lname ) < 0)
      {
        ++last;
        Symbol* tmp = items[last];
        items[last] = items[i];
        items[i] = tmp;
      }
    }

    tmp = items[l];
    items[l] = items[last];
    items[last] = tmp;
    sortSymbols(symbols, l, last-1);
    sortSymbols(symbols, last+1, r);
  }

 private:
  int id_;
  string name_;

  bool doCluster_;
  int minClusters_;
  int maxClusters_;
  Array<Cluster*> clusters_;
  Array<Cluster*> bestClusters_;

  Array<Symbol*> symbols_;
  Array<Rel*> rels_;

  StringToIntMap symNameToIdMap_;

  bool repeatedInNaryRelation_;
  Array<Relation*> binRelationsWithSameTypes_;
};

inline
ostream& operator<<(ostream& out, const Type& t) { return t.print(out); }

#endif
