#include <iostream>
using namespace std;
#include <fstream>
#include <sstream>
#include <cstring>
#include <cstdlib>
#include <climits>
#include <cfloat>
#include <unistd.h>
#include <sys/time.h>
#include "array.h"
#include "arraysaccessor.h"
#include "timer.h"
#include "hashint.h"
#include "hashstring.h"
#include "util.h"
#include "candclause.h"

const int    PRUNE1_ACCEPT_ALL_LEN = 2;
const double CLAMP_DEV             = 1e-1;
const string CONV_THRESH           = "-gConvThresh 1e-6";
const bool   NO_EQ_PRED_WT         = false;
const string TMP_MLN_FILE_NAME     = "tmp_mln_prunerules";
const string TMP_OUT_MLN_FILE_NAME = "tmp_out_mln_prunerules";
const string TMP_DB_FILE_NAME      = "tmp_db_prunerules";
const string TMP_LOG_FILE          = "tmp_logfile";
int seed_ = 1;


int pow2(const int& n)
{
  int v = 1;
  for (int i = 0; i < n; i++)
    v *= 2;
  return v;
}

int clauseLen(const string& clauseStr)
{
  int len = 0;
  for (unsigned int i = 0; i < clauseStr.length(); i++)
    if (clauseStr.at(i) == '(') len++;
  return len;
}

double countTotalComb(const Array<CandClause*>& cands)
{
  double total = 0;
  for (int i = 0; i < cands.size(); i++)
    total += pow2(cands[i]->preds().size());
  return total;
}

void readRelStr(string str, string& relStr, Array<string>& typeStrs)
{
  string::size_type ltparen = str.find("(");
  assert(ltparen != string::npos);
  relStr = Util::trim( str.substr(0, ltparen) );
  int i = ltparen+1;
  while (true)
  {
    string::size_type comma = str.find(",",i);
    if (comma == string::npos) break;
    string typeStr = Util::trim( str.substr(i, comma-i) );
    typeStrs.append(typeStr);
    i = comma+1;
  }
  string::size_type rtparen = str.find(")",i);
  assert(rtparen != string::npos);
  string typeStr = Util::trim( str.substr(i, rtparen-i) );
  typeStrs.append(typeStr);
}

void readInfoFile(const char* const& infoFile, StringToStringArrayMap& relToTypesMap)
{
  ifstream in(infoFile);
  Util::assertt(!in.fail(), "failed to open", infoFile, -1);
  StringSet relNames;
  string buf;
  bool start = false;
  while(getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.find("//Relations")  == 0) start = true;
    else if (buf.find("//Types") == 0) start = false;
    if (start)
    {
      if (buf.find("//") == 0) continue;
      string relStr;
      Array<string>* typeStrs = new Array<string>;
      readRelStr(buf, relStr, *typeStrs);

      cout << "Declared relation: " << relStr << "(" << (*typeStrs)[0];
      for (int i = 1; i < typeStrs->size(); i++) cout << "," << (*typeStrs)[i];
      cout << ")" << endl;

      if (relNames.find(relStr) == relNames.end())
      {
        relNames.insert(relStr);
        assert(relToTypesMap.find(relStr) == relToTypesMap.end());
        relToTypesMap[relStr] = typeStrs;
      }
      else
      {
        cout << "ERROR: duplicate declaration of " << relStr << endl; exit(-1);
      }
    }
  }

  in.close();
}

void readPredDeclarations(StringHashArray& predDecl, const char* infoFile)
{
  ifstream inp(infoFile);
  string buf;
  while (getline(inp,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.find("//") == 0) continue;
    predDecl.append(buf);
  }
  inp.close();
}

void createTypeToVarNameMap(StringToStringMap& typeToVarNameMap,
                            const Array<string>& types)
{
  StringSet varNames;
  for (int t = 0; t != types.size(); t++)
  {
    string type = types[t];
    if (typeToVarNameMap.find(type) != typeToVarNameMap.end()) continue;

    bool assignedVarName = false;
    for (unsigned int i = 0; i < type.length()-1; i++)
    {
      string varName = type.substr(0,i+1);
      pair<StringSet::iterator,bool> pr = varNames.insert(varName);
      if (pr.second) //if unique varName
      {
        typeToVarNameMap[type] = varName;
        assignedVarName = true;
        break;
      }
    }

    if (assignedVarName) continue;

    int cnt = 1;
    while (true)
    {
      string varName = "";
      for (int i = 0; i < cnt; i++)
        varName += type[0];
      pair<StringSet::iterator,bool> pr = varNames.insert(varName);
      if (pr.second) //if unique varName
      {
        typeToVarNameMap[type] = varName;
        break;
      }
      cnt++;
    }
  }
}

//create unitClauses with shared vars
void createUnitClauses(StringHashArray& unitClauses,
                       StringToStringArrayMap& relToTypesMap)
{
  StringToStringArrayMap::iterator it = relToTypesMap.begin();
  for (; it != relToTypesMap.end(); it++) //for each relation
  {
    string rel = (*it).first;
    Array<string>& types = *( (*it).second );

    StringToIntMap typeToNumAppear;
    bool haveRepeatedTypes = false;
    for (int t = 0; t < types.size(); t++)
    {
      string type = types[t];
      if (typeToNumAppear.find(type) == typeToNumAppear.end())
        typeToNumAppear[type] = 0;
      typeToNumAppear[type]++;
      if (typeToNumAppear[type] > 1)  haveRepeatedTypes = true;
    }

    if (!haveRepeatedTypes) continue;

    StringToIntMap typeToStartVar;
    StringToIntMap::iterator it = typeToNumAppear.begin();
    int startVar = 0;
    for (; it != typeToNumAppear.end(); it++)
    {
      string type = (*it).first;
      int numAppear = (*it).second;
      typeToStartVar[type] = startVar;
      startVar += numAppear;
    }
    assert(typeToStartVar.size() == typeToNumAppear.size());

    ArraysAccessor<int> acc;
    for (int t = 0; t < types.size(); t++)
    {
      string type = types[t];
      int numAppear = typeToNumAppear[type];
      int startVar = typeToStartVar[type];
      Array<int>* vars = new Array<int>;
      for (int i = 0; i < numAppear; i++) vars->append(startVar+i);
      acc.appendArray(vars);
    }

    StringToStringMap typeToVarNameMap;
    createTypeToVarNameMap(typeToVarNameMap, types);
    Array<int> vars;
    while(acc.getNextCombination(vars))
    {
      assert(vars.size() == types.size());
      IntToIntMap oldToNewVars;
      int cnt = 0;
      for (int t = 0; t < vars.size(); t++)
      {
        if (oldToNewVars.find(vars[t]) == oldToNewVars.end())
          oldToNewVars[vars[t]] = cnt++;
        vars[t] = oldToNewVars[vars[t]];
      }

      Array<string> strVars(vars.size());
      for (int t = 0; t < types.size(); t++)
        strVars.append(typeToVarNameMap[types[t]] + Util::intToString(vars[t]));

      bool haveRepeat = false;
      for (int a = 0; a < strVars.size(); a++)
        for (int b = a+1; b < strVars.size(); b++)
          if (strVars[a].compare(strVars[b])==0) { haveRepeat = true; break; }

      if (haveRepeat)
      {
        string unitClause = rel + "(" + strVars[0];
        for (int v = 1; v < strVars.size(); v++)
          unitClause += "," + strVars[v];
        unitClause += ")";
        unitClauses.append(unitClause);
      }
    }
    acc.deleteArraysAndClear();
  }//for each relation

  //for each relation
  /*for (it = relToTypesMap.begin(); it != relToTypesMap.end(); it++)
  {
    string rel = (*it).first;
    Array<string>& types = *( (*it).second );
    string unitClause = rel + "(v0";
    for (int v =  1; v < types.size(); v++)
      unitClause += ",v" + Util::intToString(v);
    unitClause += ")";
    unitClauses.append(unitClause);
  }*/

  for (int i = 0; i < unitClauses.size(); i++)
    cout << "UNIT_CLAUSE: " << unitClauses[i] << endl;
}

int countNumCombs(const string& semnetFile)
{
  ifstream in(semnetFile.c_str());
  Util::assertt(!in.fail(), "failed to open semnet file", semnetFile, -1);
  string buf;
  int cnt = 0;
  while (getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.find("#START_COMB") != 0) continue;
    string noop; int idx;
    istringstream iss(buf);
    iss >> noop >> noop >> idx;
    Util::assertt(idx == cnt, "expect idx == cnt", idx, cnt, -1);
    cnt++;
  }
  in.close();
  return cnt;
}

void readSemanticNetwork(string semnetFiles,
                         Array<Array<Array<string>*> >& combIdToTuplesMapByDomain)
{
  for (unsigned int i = 0; i < semnetFiles.length(); i++)
    if (semnetFiles.at(i) == ',') semnetFiles.at(i) = ' ';
  semnetFiles.append(" ::");

  Array<string> semnetFilesArr;
  istringstream iss(semnetFiles);
  string buf; iss >> buf;
  while (buf.compare("::") != 0) { semnetFilesArr.append(buf); iss >> buf; }

  int numDom = semnetFilesArr.size();
  combIdToTuplesMapByDomain.growToSize( numDom );
  for (int d = 0; d < numDom; d++)
  {
    string semnetFile = semnetFilesArr[d];
    Array<Array<string>*>& combIdToTuplesMap = combIdToTuplesMapByDomain[d];

    int numOfCombs = countNumCombs(semnetFile);
    combIdToTuplesMap.growToSize(numOfCombs);
    for (int i = 0; i < combIdToTuplesMap.size(); i++)
      combIdToTuplesMap[i] = new Array<string>;

    ifstream in(semnetFile.c_str());
    Util::assertt(!in.fail(), "failed to open semnet file", semnetFile, -1);
    string buf;
    while (getline(in,buf))
    {
      buf = Util::trim(buf);
      if (buf.empty()) continue;
      if (Util::startsWith(buf,"#START_RELATION"))
      {
        while (!Util::startsWith(buf,"#END_RELATION"))
        {
          buf = Util::getLineAndTrim(in);
          int clustCombId = -1;
          if (Util::startsWith(buf,"#START_COMB"))
          {
            Array<int> clustIds;
            istringstream iss(buf);
            string noop; int numClusters;
            iss >> noop >> noop >> clustCombId >> noop >> numClusters;
            for (int i = 0; i < numClusters; i++)
            {
              buf = Util::getLineAndTrim(in); assert(Util::startsWith(buf,"#START_CLUSTER"));
              while (!Util::startsWith(buf,"#END_CLUSTER")) //read all symbol
              buf = Util::getLineAndTrim(in);
            }
            buf = Util::getLineAndTrim(in);  Util::assertt(Util::startsWith(buf,"#END_COMB"),"expect #END_COMB", -1);
            buf = Util::getLineAndTrim(in);  Util::assertt(Util::startsWith(buf,"#START_TRUE_TUPLES"),"expect #START_TRUE_TUPLES", -1);

            //read the true tuples
            buf = Util::getLineAndTrim(in);
            while (buf.find("#END_TRUE_TUPLES")!=0)
            {
              combIdToTuplesMap[clustCombId]->append(buf);
              buf = Util::getLineAndTrim(in);
            }
            combIdToTuplesMap[clustCombId]->compress();
          }
        }//while not end relation
      }//if start relation
    }

    in.close();
  }//for each domain

  /*for (int d = 0; d < combIdToTuplesMapByDomain.size(); d++)
  {
    Array<Array<string>*>& combIdToTuplesMap = combIdToTuplesMapByDomain[d];
    for (int i = 0; i < combIdToTuplesMap.size(); i++)
    {
      Array<string>& tuples = *( combIdToTuplesMap[i] );
      for (int j = 0; j < tuples.size(); j++)
        cout << i << ": " << tuples[j] << endl;
    }
  }*/
}//readSemanticNetwork()

void readCands(string inCandFiles, Array<CandClause*>& candClauses)
{
  Array<Array<CandClause*> > candClausesByLen;
  candClausesByLen.growToSize(20);

  for (unsigned int i = 0; i < inCandFiles.length(); i++)
    if (inCandFiles.at(i) == ',') inCandFiles.at(i) = ' ';
  inCandFiles.append(" ::");

  Array<string> inCandFilesArr;
  istringstream iss(inCandFiles);
  string buf; iss >> buf;
  while (buf.compare("::") != 0) { inCandFilesArr.append(buf); iss >> buf; }

  CandClauseSet ccset;
  int numDom = inCandFilesArr.size();
  for (int d = 0; d < numDom; d++)
  {
    string inCandFile = inCandFilesArr[d];

    ifstream in(inCandFile.c_str());
    Util::assertt(!in.fail(), "failed to open cand file", inCandFile, -1);
    string buf;
    while (getline(in,buf))
    {
      buf = Util::trim(buf);
      if (buf.empty()) continue;
      if (buf.find("#CAND:") != 0) continue;

      int startIdx = strlen("#CAND:");
      string::size_type endIdx = buf.find("#CLUST_COMB_IDS");
      Util::assertt(endIdx != string::npos, "expect #CLUST_COMB_IDS",-1);
      CandClause* cand = new CandClause(numDom);

      string candStr = Util::trim( buf.substr(startIdx, endIdx-startIdx) );
      candStr += " ::";
      istringstream iss(candStr);
      string predStr; iss >> predStr;
      while (predStr.compare("::") != 0) { cand->addPred(predStr); iss>>predStr; }

      startIdx = endIdx + strlen("#CLUST_COMB_IDS");
      string support = Util::trim( buf.substr(startIdx, buf.length()-startIdx) );
      support += " ::";
      istringstream iss2(support);
      string sup; iss2 >> sup;
      while (sup.compare("::") != 0) { cand->addSupport(d,sup); iss2 >> sup; }

      pair<CandClauseSet::iterator,bool> pr = ccset.insert(cand);
      if (pr.second) candClausesByLen[ cand->preds().size() ].append(cand);
      else
      {
        const Array<string>& support = cand->supportByDomain(d);
        CandClause* existingCandClause = *(pr.first);
        existingCandClause->addSupport(d,support);
        delete cand;
      }

      /*const Array<string>& preds = cand->preds();
      for (int i = 0; i < preds.size(); i++)
        cout << preds[i] << endl;
      const Array<string>& sups = cand->support();
      for (int i = 0; i < sups.size(); i++)
        cout << sups[i] << endl;*/
    }

    in.close();
  }//for each candFile

  for (int i = 1; i < candClausesByLen.size(); i++)
    candClauses.append( candClausesByLen[i] );
  candClauses.compress();
}//readCands()

void getVars(Array<int>& vars, Array<string>& varPrefixes,
             const Array<string>& args)
{
  for (int i = 0; i < args.size(); i++)
  {
    string arg = args[i];

    unsigned int j = 0;
    for (; j < arg.length(); j++)
      if (isdigit( arg.at(j) )) break;
    assert(j < arg.length());

    string prefix = arg.substr(0,j);
    int    var    = atoi( arg.substr(j,arg.length()-j).c_str() );

    vars.append(var);
    varPrefixes.append(prefix);
  }
}

void sortPreds(Array<string>& preds, const int&l, const int& r)
{
  string* items = (string*) preds.getItems();

  if (l >= r) return;
  string tmp = items[l];
  items[l] = items[(l+r)/2];
  items[(l+r)/2] = tmp;

  int last = l;
  for (int i = l+1; i <= r; i++)
  {
    string itemi = items[i];
    string iteml = items[l];

    string reli, rell;
    Array<string> argsi, argsl;
    readRelStr(itemi, reli, argsi);
    readRelStr(iteml, rell, argsl);
    if (reli.at(0) == '!') reli = reli.substr(1,reli.length()-1);
    if (rell.at(0) == '!') rell = rell.substr(1,rell.length()-1);
    assert(reli.compare(rell)==0);
    assert(argsi.size() == argsl.size());

    Array<string> varPrefixes;
    Array<int> varsi, varsl;
    getVars(varsi, varPrefixes, argsi);
    getVars(varsl, varPrefixes, argsl);
    assert(varsi.size() == varsl.size());

    bool smaller = false;
    for (int a = 0; a < varsi.size(); a++)
    {
      if (varsi[a] < varsl[a])      { smaller=true;  break; }
      else if (varsi[a] > varsl[a]) { smaller=false; break; }
    }

    if (smaller)
    {
      ++last;
      string tmp = items[last];
      items[last] = items[i];
      items[i] = tmp;
    }
  }

  tmp = items[l];
  items[l] = items[last];
  items[last] = tmp;
  sortPreds(preds, l, last-1);
  sortPreds(preds, last+1, r);
}

void sortPredsBySign(Array<string>& preds, const int&l, const int& r)
{
  string* items = (string*) preds.getItems();

  if (l >= r) return;
  string tmp = items[l];
  items[l] = items[(l+r)/2];
  items[(l+r)/2] = tmp;

  int last = l;
  for (int i = l+1; i <= r; i++)
  {
    string itemi = items[i];
    string iteml = items[l];

    string reli, rell;
    Array<string> argsi, argsl;
    readRelStr(itemi, reli, argsi);
    readRelStr(iteml, rell, argsl);
    bool ifalse = (reli.at(0) == '!');
    bool lfalse = (rell.at(0) == '!');
    bool smaller = false;
    if ((ifalse != lfalse) && ifalse) smaller = true;

    if (smaller)
    {
      ++last;
      string tmp = items[last];
      items[last] = items[i];
      items[i] = tmp;
    }
  }

  tmp = items[l];
  items[l] = items[last];
  items[last] = tmp;
  sortPreds(preds, l, last-1);
  sortPreds(preds, last+1, r);
}

void normalizeVar(Array<string>& preds)
{
  Array<string> ppreds(preds);

  IntToIntMap oldToNewVar;
  int cnt = 0;
  for (int i = 0; i < ppreds.size(); i++)
  {
    string pred = ppreds[i];
    string rel; Array<string> args;
    readRelStr(pred, rel, args);

    Array<int> vars;
    Array<string> varPrefixes;
    getVars(vars, varPrefixes, args);
    assert(vars.size()==varPrefixes.size());

    for (int j = 0; j < vars.size(); j++)
    {
      IntToIntMap::iterator it = oldToNewVar.find( vars[j] );
      int newVar;
      if (it == oldToNewVar.end()) { newVar = cnt; oldToNewVar[vars[j]]=cnt++; }
      else                         { newVar = (*it).second; }
      vars[j] = newVar;
    }

    pred = rel + "(" + varPrefixes[0] + Util::intToString( vars[0] );
    for (int j = 1; j < vars.size(); j++)
      pred += "," + varPrefixes[j] + Util::intToString( vars[j] );
    pred += ")";
    ppreds[i] = pred;
  }

  //canonicalize
  for (int i = 0; i < ppreds.size(); i++)
  {
    string pred = ppreds[i];
    string reli; Array<string> args;
    readRelStr(pred, reli, args);
    if (reli.at(0) == '!')  reli = reli.substr(1, reli.length()-1);

    int j = i;
    for (; j < ppreds.size(); j++)
    {
      string pred = ppreds[j];
      string relj; Array<string> args;
      readRelStr(pred, relj, args);
      if (relj.at(0) == '!')  relj = relj.substr(1, relj.length()-1);
      if (reli.compare(relj) != 0) break;
    }
    j--;

    if (j > i) sortPreds(ppreds, i, j);

    i = j;
  }

  preds.clear();
  preds.append(ppreds);
}

void getLits(Array<string>& lits, string cand)
{
  for (unsigned int i = 0; i < cand.length(); i++)
  {
    if (cand.at(i) == 'v' && (i > 0 && cand.at(i-1) == ' ') &&
        (i < cand.length()-1 && cand.at(i+1) == ' '))
      cand.at(i) = ' ';
  }
  cand.append(" ::");
  istringstream iss(cand);
  string buf;
  iss >> buf;
  while(buf.compare("::") != 0) { lits.append(buf); iss >> buf; }
}

void sortPredsBySign(Array<string>& preds)
{
  for (int i = 0; i < preds.size(); i++)
  {
    string pred = preds[i];
    string signRel; Array<string> args;
    readRelStr(pred, signRel, args);
    string rel;
    if (signRel.at(0) == '!')  rel = signRel.substr(1, signRel.length()-1);
    else                       rel = signRel;

    int j = i;
    for (; j < preds.size(); j++)
    {
      pred = preds[j];
      Array<string> args;
      readRelStr(pred, signRel, args);
      string rel2;
      if (signRel.at(0) == '!')  rel2 = signRel.substr(1, signRel.length()-1);
      else                       rel2 = signRel;
      if (rel.compare(rel2) != 0) break;
    }
    j--;

    if (j > i) sortPredsBySign(preds, i, j);
    i = j;
  }
}

string predsToClause(Array<string>& preds, const bool& sortBySign)
{
  if (sortBySign) sortPredsBySign(preds);
  normalizeVar(preds); //ignores pred signs
  string clause = preds[0];
  for (int i = 1; i < preds.size(); i++)
    clause += " v " +  preds[i];
  return clause;
}

string canonicalize(const string& clause, const bool& sortBySign)
{
  Array<string> lits;
  getLits(lits, clause);
  return predsToClause(lits, sortBySign);
}

void createClauses(Array<string>& clauses, const CandClause* const& candClause,
                   const bool& canon)
{
  const Array<string>& preds = candClause->preds();

  //write the formulas
  ArraysAccessor<bool> acc;
  for (int i = 0; i < preds.size(); i++)
  {
    Array<bool>* barr = new Array<bool>(2);
    barr->append(true); barr->append(false);
    acc.appendArray(barr);
  }

  clauses.clear();
  clauses.growToSize( int(pow(2.0,preds.size())) );
  int numClausesCreated = 0;

  while(acc.hasNextCombination())
  {
    ostringstream oss;
    bool sign;
    int idx = 0;
    while (acc.nextItemInCombination(sign))
    {
      string pred = preds[idx++];
      if (sign == false) oss << "!";
      oss << pred;
      if (idx-1 < preds.size()-1) oss << " v ";
    }

    clauses[ numClausesCreated++ ] = oss.str();
  }

  acc.deleteArraysAndClear();

  if (canon)
  {
    StringSet sset;
    Array<string> copy(clauses);
    clauses.clear();
    for (int i = 0; i < copy.size(); i++)
    {
      pair<StringSet::iterator,bool> pr = sset.insert( canonicalize(copy[i], true) );
      if (pr.second) clauses.append(copy[i]);
    }
  }
}

string createPredDeclarationsHelper(const StringSet& relNames,
                                    StringToStringArrayMap& relToTypesMap)
{
  StringSet uniqueDecl;
  ostringstream oss;
  for (StringSet::iterator it = relNames.begin(); it != relNames.end(); it++)
  {
    string relName = (*it);
    Array<string>& types = *( relToTypesMap[relName] );
    string preddecl = relName + "(" + types[0];
    for (int i = 1; i < types.size(); i++)
      preddecl += "," + types[i];
    preddecl += ")";

    pair<StringSet::iterator, bool> pr = uniqueDecl.insert(preddecl);
    if (pr.second) oss << preddecl << endl;
  }
  return oss.str();
}

string createPredDeclarations(const CandClause* const& candClause,
                              StringToStringArrayMap& relToTypesMap)
{
  StringSet relNames;
  const Array<string>& preds = candClause->preds();
  for (int i = 0; i < preds.size(); i++)
  {
    string relName;
    Array<string> typeStrs;
    readRelStr(preds[i], relName, typeStrs);
    relNames.insert(relName);
  }

  return createPredDeclarationsHelper(relNames,relToTypesMap);
}

void getPredsInClause(StringHashArray& predNames,
                      const CandClause* const& candClause)
{
  const Array<string>& preds = candClause->preds();
  for (int i = 0; i < preds.size(); i++)
  {
    string relName;
    Array<string> typeStrs;
    readRelStr(preds[i], relName, typeStrs);
    predNames.append(relName);
  }
}

string createPredDeclarations(const string& candClause,
                              StringToStringArrayMap& relToTypesMap)
{
  StringSet relNames;
  int startIdx = 0;
  while (true)
  {
    string::size_type lparen = candClause.find('(', startIdx);
    if (lparen == string::npos) break;
    string::size_type beg = candClause.rfind(' ',lparen);
    if (beg == string::npos) beg = 0;
    else                     beg++;
    if (candClause.at(beg) == '!') beg++;
    string relName = candClause.substr(beg, lparen-beg);
    //cout << "relName " << "\"" << relName << "\"" << endl;
    relNames.insert(relName);
    startIdx = lparen+1;
  }
  return createPredDeclarationsHelper(relNames, relToTypesMap);
}

//support is a copy
void getCombIds(Array<int>& combIds, const int& clauseLen, string sup)
{
  for (unsigned int j = 0; j < sup.length(); j++)
    if (sup[j] == '-') sup[j] = ' ';
  istringstream iss(sup);
  for (int j = 0; j < clauseLen; j++)
  {
    int id; iss >> id; combIds.append(id);
  }
}

int countTotalUniqueTuples(const Array<Array<string> >& uniqueTuplesByDomain)
{
  int n = 0;
  for (int d = 0; d < uniqueTuplesByDomain.size(); d++)
    n += uniqueTuplesByDomain[d].size();
  return n;
}

double getUniqueTuples(Array<Array<string> >& uniqueTuplesByDomain,
                       const int& maxTuples,
                       const CandClause* const& candClause,
                       Array<Array<Array<string>*> >& combIdToTuplesMapByDomain)
{
  int numDom = uniqueTuplesByDomain.size();
  const Array<Array<string> >& supportByDomain = candClause->supportByDomain();

  assert(combIdToTuplesMapByDomain.size() == numDom);
  assert(supportByDomain.size() == numDom);

  Array<Array<Array<string>*> > tuplesByClustCombByDomain;
  tuplesByClustCombByDomain.growToSize(numDom);

  //get the uniqueTuples in each domain
  for (int d = 0; d < numDom; d++)
  {
    Array<string>& uniqueTuples = uniqueTuplesByDomain[d];
    Array<Array<string>*>& combIdToTuplesMap = combIdToTuplesMapByDomain[d];
    Array<string>& support = supportByDomain[d];

    Array<Array<string>*>& tuplesByClustComb = tuplesByClustCombByDomain[d];
    tuplesByClustComb.growToSize(support.size(), NULL);

    StringSet sset;
    for (int i = 0; i < support.size(); i++) //for each comb of clustComb
    {
      //read the comb ids (each comb id correspond to a predicate)
      string sup = support[i];

      Array<int> combIds( candClause->preds().size() );
      getCombIds(combIds, candClause->preds().size(), support[i]);

      assert(tuplesByClustComb[i] == NULL);
      tuplesByClustComb[i] = new Array<string>;
      Array<string>* supTuples = tuplesByClustComb[i];

      for (int j = 0; j < combIds.size(); j++)
      {
        Array<string>& tuples = *( combIdToTuplesMap[ combIds[j] ] );
        for (int k = 0; k < tuples.size(); k++)
        {
          string tuple = tuples[k];
          pair<StringSet::iterator,bool> pr = sset.insert(tuple);
          if (pr.second) uniqueTuples.append(tuple);
          supTuples->append(tuple);
        }
      }
    }//for each comb of clustComb (for each support)
  }//for each domain

  //determine if we need to sample the tuples
  const double totalUniqueTuples = countTotalUniqueTuples(uniqueTuplesByDomain);

  if (totalUniqueTuples <= maxTuples)
  {
    for (int d = 0; d < numDom; d++)
      tuplesByClustCombByDomain[d].deleteItemsAndClear();
    cout << "NO SAMPLING OF TUPLES " << totalUniqueTuples << endl;
    return totalUniqueTuples;
  }

  //note which domains have tuples
  Array<int> domWithTuples;
  for (int d = 0; d < numDom; d++)
    if (uniqueTuplesByDomain[d].size() > 0) domWithTuples.append(d);
  Util::assertt(domWithTuples.size() > 0, "expect at least 1 sup tuple", -1);

  //sample from domains with tuples
  IntHashArray chosenDoms;
  double numChosenTuples = 0.0;
  while (numChosenTuples < maxTuples)
  {
    int idx = random() % domWithTuples.size();
    int dom = domWithTuples[ idx ];
    chosenDoms.append(dom);
    domWithTuples.removeItemFastDisorder(idx);
    numChosenTuples += uniqueTuplesByDomain[dom].size();
  }

  //clear the tuples in domains that are not sampled
  for (int d = 0; d < numDom; d++)
    if (!chosenDoms.contains(d)) uniqueTuplesByDomain[d].clear();


  //determine if we need to further sample from tuples from domain
  if (numChosenTuples <= maxTuples)
  {
    for (int d = 0; d < numDom; d++)
      tuplesByClustCombByDomain[d].deleteItemsAndClear();
    cout << "SAMPLED_TUPLES " << numChosenTuples << " / "<<totalUniqueTuples
         << endl;
    assert(numChosenTuples > 0.0);
    return totalUniqueTuples;
  }

  cout << "  SAMPLING DOWN FROM " << numChosenTuples << " TO MAX_TUPLES" <<endl;
  while (true)
  {
    for (int d = 0; d < chosenDoms.size(); d++)
    {
      int dom = chosenDoms[d];
      if (uniqueTuplesByDomain[dom].size() == 0) continue;
      int idx = random() % uniqueTuplesByDomain[dom].size();
      uniqueTuplesByDomain[dom].removeItemFastDisorder(idx);
      numChosenTuples--;
      if (numChosenTuples <= maxTuples) break;
    }
    if (numChosenTuples <= maxTuples) break;
  }

  chosenDoms.clear();
  for (int d = 0; d < uniqueTuplesByDomain.size(); d++)
    if (uniqueTuplesByDomain.size() > 0) chosenDoms.append(d);

  double tmp = 0.0;
  for (int d = 0; d < chosenDoms.size(); d++)
    tmp += uniqueTuplesByDomain[ chosenDoms[d] ].size();
  assert(tmp == numChosenTuples && tmp == maxTuples);


  cout << "SAMPLED_TUPLES " << numChosenTuples << " / " << totalUniqueTuples
       << endl;

  for (int d = 0; d < numDom; d++)
    tuplesByClustCombByDomain[d].deleteItemsAndClear();
  assert(numChosenTuples > 0);
  return totalUniqueTuples;
}

string createDBFiles(Array<string>& dbFiles,
                     const Array<Array<string> >& tuplesByDomain,
                     const string& tmpDir)
{
  string dbFileStr = "";
  assert(tuplesByDomain.size() > 0);
  for (int d = 0; d < tuplesByDomain.size(); d++)
  {
    Array<string>& tuples = tuplesByDomain[d];
    if (tuples.empty()) continue;

    string domDbFile = tmpDir + TMP_DB_FILE_NAME + Util::intToString(d) + ".db";
    dbFiles.append(domDbFile);
    dbFileStr += ((dbFiles.size()==1)?"":",") + domDbFile;

    ofstream dbOut(domDbFile.c_str());
    Util::assertt(!dbOut.fail(), "failed to open dbFile", domDbFile, -1);
    for (int k = 0; k < tuples.size(); k++)
      dbOut << tuples[k] << endl;
    dbOut.close();
  }
  return dbFileStr;
}

void readClauses(StringHashArray& outCandStrs, int& numPruned,
                 Array<double>& absWts, const string& outMLNFile,
                 const int& numClausesCreated, const double& minWt,
                 const bool& pruneAll)
{
  absWts.clear();
  ifstream in(outMLNFile.c_str());
  Util::assertt(!in.fail(), "failed to open out mln file", outMLNFile, -1);
  string buf;

  //read till //function declarations
  while(getline(in, buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.find("//function declarations") == 0) break;
  }

  //start reading the clauses and weights
  int numClausesRead = 0;
  while(getline(in, buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.find("//") != 0) continue;
    cout << "WT_RULE: " << buf << endl;

    if (numClausesRead++ < numClausesCreated)
    {
      string::size_type orr  = buf.find(" v ");
      string::size_type andd = buf.find(" ^ ");
      Util::assertt(orr != string::npos || andd != string::npos, "don't expect unit clause", buf, -1);

      string::size_type ws0 = buf.find(" ");
      string::size_type ws1 = buf.find(" ", ws0+1);
      string wtStr = Util::trim( buf.substr(ws0, ws1-ws0) );
      Util::assertt(isdigit(wtStr[0]) || wtStr[0]=='-', "expect wt", -1);
      double wt = atof(wtStr.c_str());
      double abswt = fabs(wt);
      absWts.append(abswt);

      if (pruneAll) numPruned++;
      else
      {
        if (abswt >= minWt)
        {
          string clauseStr = Util::trim( buf.substr(ws1, buf.length()-ws1) );
          outCandStrs.append(clauseStr);
        }
        else numPruned++;
      }
    }
  }
  cout << endl;

  in.close();
}

void printCand(ostream& out, const CandClause* const& cand)
{
  const Array<string>& preds = cand->preds();
  out << preds[0];
  for (int p = 1; p < preds.size(); p++)
    out << "  " << preds[p];
}

void createMLNFile(const string& mlnFile, const string& predDecl,
                   const string& clause, const StringHashArray& unitClauses,
                   const Array<double>*const& unitWts)
{
  if (unitWts) Util::assertt(unitWts->size() == unitClauses.size(),
                             "expect #wts == #unitClauses",unitWts->size(),
                             unitClauses.size(), -1);

  ofstream mlnOut(mlnFile.c_str());
  Util::assertt(!mlnOut.fail(), "failed to open mln file", mlnFile, -1);

  mlnOut << "//predicate declarations" << endl;
  mlnOut << predDecl << endl; //write predicate declarations

  mlnOut << clause << endl; //write formulas

  if (unitWts)
  {
    for (int i = 0; i < unitClauses.size(); i++)
      mlnOut << (*unitWts)[i] << " " << unitClauses[i] << endl;
  }
  else
  {
    for (int i = 0; i < unitClauses.size(); i++)
      mlnOut << unitClauses[i] << endl;
  }

  mlnOut.close();
}

void createCommand(char* const& command, const char* const& learnWtsExec,
                   const string& mlnFile, const string& outMLNFile,
                   const string& dbFilesStr, const bool& clampUnit,
                   const bool& countNumGndings, const string& logFile)
{
  string noEqualPredWtStr = NO_EQ_PRED_WT ? "-gNoEqualPredWt" : "";
  string countNumGndingsStr = countNumGndings ? "-countGndings" : "";
  string clampUnitStr = "";
  string clampDev     = "";
  if (clampUnit)
  {
    clampUnitStr =  "-clampUnit";
    clampDev     =  "-clampDev " + Util::doubleToString(CLAMP_DEV);
  }
  sprintf(command, "%s -g -i %s -o %s -t %s -multipleDatabases %s %s %s %s %s > %s",
          learnWtsExec, mlnFile.c_str(), outMLNFile.c_str(), dbFilesStr.c_str(),
          CONV_THRESH.c_str(), noEqualPredWtStr.c_str(), clampUnitStr.c_str(),
          clampDev.c_str(), countNumGndingsStr.c_str(), logFile.c_str());
  cout << "command: " << command << endl;
}

double readScore(const string& logFile)
{
  ifstream in(logFile.c_str());
  if (in.fail())
  {
    cout << "ERROR: failed to open " << logFile << endl;
    return -1000000.0;
  }
  //Util::assertt(!in.fail(), "failed to open", logFile, -1);

  string buf;
  while (getline(in,buf))
  {
    if (buf.find("pseudo-log-likelihood = ") == 0)
    {
      string noop; double score;
      istringstream iss(buf); iss >> noop >> noop >> score;
      return score;
    }
  }
  //Util::exit("failed to read score",-1);
  cout << "FAILED TO READ SCORE!!!!!!" << endl;
  in.close();

  return -1000000.0;
}

double runLearnWts(const string& mlnFile, const string& predDecl,
                   const string& clause, const StringHashArray& unitClauses,
                   const Array<double>*const& unitWts, char*const& command,
                   const char* const& learnWtsExec, const string& outMLNFile,
                   const string& dbFilesStr, const bool& clampUnit,
                   const bool& countNumGndings, const string& logFile)
{
  createMLNFile(mlnFile, predDecl, clause, unitClauses, unitWts);
  createCommand(command, learnWtsExec, mlnFile, outMLNFile, dbFilesStr,
                clampUnit, countNumGndings, logFile);
  system(command);
  double loglike = readScore(logFile);
  return loglike;
}

void readWts(Array<double>& unitWts, const string& logFile)
{
  ifstream in(logFile.c_str());
  Util::assertt(!in.fail(), "failed to open", logFile, -1);
  string buf;
  while (getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.at(0) == '-' || isdigit(buf.at(0)))
    {
      istringstream iss(buf);
      double wt; iss >> wt;
      unitWts.append(wt);
    }
  }
  in.close();
}

void findUnitWts(Array<double>& unitWts, const string& mlnFile,
                 const string& predDecl, const StringHashArray& unitClauses,
                 char*const& command, const char* const& learnWtsExec,
                 const string& outMLNFile, const string& dbFilesStr,
                 const string& logFile)
{
  createMLNFile(mlnFile, predDecl, "", unitClauses, NULL);
  bool clampUnit = false;
  bool countNumGndings = false;
  createCommand(command, learnWtsExec, mlnFile, outMLNFile, dbFilesStr,
                clampUnit, countNumGndings, logFile);
  system(command);
  readWts(unitWts, outMLNFile);
  Util::assertt(unitWts.size() == unitClauses.size(),
                "expected #unitWts == #unitClauses", unitWts.size(),
                unitClauses.size(), -1);
  //for (int i = 0; i < unitClauses.size(); i++)
  //  cout << "unitWts " << i << " : " << unitWts[i] << "  "
  //       << unitClauses[i] << endl;
}

void getUnitClauses(StringHashArray& uunitClauses,
                    const StringHashArray& predNames,
                    const StringHashArray& unitClauses)
{
  for (int i = 0; i < unitClauses.size(); i++)
  {
    string clause = unitClauses[i];
    string::size_type paren = clause.find('(');  assert(paren != string::npos);
    string pred = clause.substr(0,paren);
    if (predNames.contains(pred)) uunitClauses.append(clause);
  }
}

void createLenMinus1Cands(Array<string>& lenMinus1Cands,
                          Array<string>& otherCands, const string& cand)
{
  Array<string> lits;
  getLits(lits, cand);  assert(lits.size() == clauseLen(cand));

  for (int i = 0; i < lits.size(); i++)
  {
    Array<string> preds, others;
    for (int j = 0; j < lits.size(); j++)
    {
      if (j != i) preds.append(lits[j]);
      else        others.append(lits[j]);
    }

    lenMinus1Cands.append( predsToClause(preds, false)  );
    otherCands.append(     predsToClause(others,false) );
  }
}

void createLenMinus2Cands(Array<string>& lenMinus2Cands,
                          Array<string>& otherCands, const string& cand)
{
  Array<string> lits;
  getLits(lits, cand);  assert(lits.size() == clauseLen(cand));

  if (lits.size() <= 2) return;

  for (int i = 0; i < lits.size(); i++)
    for (int j = i+1; j < lits.size(); j++)
    {
      Array<string> preds, others;
      for (int k = 0; k < lits.size(); k++)
      {
        if (k != i && k != j) preds.append(lits[k]);
        else                  others.append(lits[k]);
      }

      lenMinus2Cands.append( predsToClause(preds, false) );
      otherCands.append(     predsToClause(others,false));
    }
}

void createLenMinus3Cands(Array<string>& lenMinusCands,
                          Array<string>& otherCands, const string& cand)
{
  Array<string> lits;
  getLits(lits, cand);  assert(lits.size() == clauseLen(cand));

  if (lits.size() <= 3) return;

  for (int a = 0; a < lits.size(); a++)
    for (int b = a+1; b < lits.size(); b++)
      for (int c = b+1; c < lits.size(); c++)
      {
        Array<string> preds, others;
        for (int k = 0; k < lits.size(); k++)
        {
          if (k != a && k != b && k != c) preds.append(lits[k]);
          else                            others.append(lits[k]);
        }

        lenMinusCands.append( predsToClause(preds, false) );
        otherCands.append(    predsToClause(others,false));
      }
}

void createLenMinus4Cands(Array<string>& lenMinusCands,
                          Array<string>& otherCands, const string& cand)
{
  Array<string> lits;
  getLits(lits, cand);  assert(lits.size() == clauseLen(cand));

  if (lits.size() <= 4) return;

  for (int a = 0; a < lits.size(); a++)
    for (int b = a+1; b < lits.size(); b++)
      for (int c = b+1; c < lits.size(); c++)
        for (int d = c+1; d < lits.size(); d++)
        {
          Array<string> preds, others;
          for (int k = 0; k < lits.size(); k++)
          {
            if (k != a && k != b && k != c && k != d) preds.append(lits[k]);
            else                                      others.append(lits[k]);
          }

          lenMinusCands.append( predsToClause(preds, false) );
          otherCands.append(    predsToClause(others,false));
        }
}

void createSubClauses(Array<string>& subClauses, const string& cand)
{
  int candLen = clauseLen(cand);
  Util::assertt(candLen <= 8, "expect clauseLen to be <= 8", -1);

  int halfLen = candLen/2;
  Array<Array<string> > others; others.growToSize(halfLen);
  for (int i = 1; i <= halfLen; i++)
  {
    if (i == 1 && candLen >1)      createLenMinus1Cands(subClauses,others[i-1],cand);
    else if (i == 2 && candLen >2) createLenMinus2Cands(subClauses,others[i-1],cand);
    else if (i == 3 && candLen >3) createLenMinus3Cands(subClauses,others[i-1],cand);
    else if (i == 4 && candLen >4) createLenMinus4Cands(subClauses,others[i-1],cand);
  }

  int startIdx = candLen%2==0 ? halfLen-1-1 : halfLen-1;
  for (int i = startIdx; i >0; i--) //don't add in unit clauses
    subClauses.append(others[i]);

  /*StringHashArray unique;
  for (int i = 0; i < subClauses.size(); i++)
  {
    if (i > 0)
      Util::assertt(clauseLen(subClauses[i-1]) >= clauseLen(subClauses[i]),
                    "expect subClause to be arranged longest to shortest",
                    subClauses[i-1], subClauses[i], -1);
    cout << "subClause " << i << ": " << subClauses[i] << endl;
    unique.append(subClauses[i]);
  }
  cout << "#unique = " << unique.size() << endl;*/
}

void pruneCandsHelper(StringHashArray& outCandStrs,
                      StringHashArray& prunedCandStrs,
                      int& numPruned, int& numIgnored,
                      const int& minTuples, const int& maxTuples,
                      const double& minWt,  const char* const& learnWtsExec,
                      const double& L1c,    const CandClause* const& candClause,
                      const string& tmpDir, const double& lenPenalty,
                      Array<Array<Array<string>*> >& combIdToTuplesMapByDomain,
                      StringToStringArrayMap& relToTypesMap,
                      const bool& evalRules, const StringHashArray& unitClauses)
{
  //numPruned = 0; //commented out: should not reset to 0

  //create all combinations of positive/negative literals
  Array<string> clauses;
  createClauses(clauses, candClause, false);

  Array<Array<string> > uniqueTuplesByDomain;
  uniqueTuplesByDomain.growToSize( combIdToTuplesMapByDomain.size() );
  double numUniqueTuples = getUniqueTuples(uniqueTuplesByDomain,
                                           maxTuples, candClause,
                                           combIdToTuplesMapByDomain);
  int candClauseLen = candClause->preds().size();
  cout << "#SUPPORTING TUPLES\t" << numUniqueTuples << "\t"
       << numUniqueTuples / candClauseLen << "\t";
  printCand(cout, candClause); cout << endl;

  //accept all clauses of this length or less
  if (candClauseLen <= PRUNE1_ACCEPT_ALL_LEN)
  {
    for (int i = 0; i < clauses.size(); i++)
      outCandStrs.append( clauses[i] );
    return;
  }

  int min= minTuples * candClauseLen;
  if (numUniqueTuples < min)
  {
    cout << "IGNORE CANDS, TOO FEW TUPLES " << numUniqueTuples << " < " << min
         << " : " << clauses[0] << endl;
    //cout << "ignore" << endl; char zz; cin >> zz;
    int num = pow2(candClauseLen);  assert(num == clauses.size());
    numIgnored += num;
    numPruned  += num;
    for (int i = 0; i < clauses.size(); i++)
      prunedCandStrs.append( clauses[i] );
    return;
  }

  //if not evaluating rules by their loglikelihoods in cluster combination
  if (!evalRules)
  {
    for (int i = 0; i < clauses.size(); i++)
      outCandStrs.append( clauses[i] );
    return;
  }

  Array<string> dbFiles;
  string dbFilesStr = createDBFiles(dbFiles, uniqueTuplesByDomain, tmpDir);
  string mlnFile    = tmpDir + TMP_MLN_FILE_NAME     + ".mln";
  string outMLNFile = tmpDir + TMP_OUT_MLN_FILE_NAME + ".mln";
  string logFile    = tmpDir + TMP_LOG_FILE;

  char* command = new char[2000 + dbFilesStr.length()];
  string predDecl = createPredDeclarations(candClause, relToTypesMap);

  //get the preds and unitClauses that appear in candClause
  StringHashArray predNames;
  getPredsInClause(predNames, candClause);
  StringHashArray uunitClauses;
  getUnitClauses(uunitClauses, predNames, unitClauses);

  //learn wts due to unit clauses
  //Array<double> unitWts;
  //findUnitWts(unitWts, mlnFile, predDecl, uunitClauses, command,
  //            learnWtsExec, outMLNFile, dbFilesStr, logFile);

  //find score due to unit clauses
  bool clampUnit = false;
  bool countNumGndings = false;
  double prevScore = runLearnWts(mlnFile, predDecl, "", uunitClauses, NULL,
                                 command, learnWtsExec, outMLNFile, dbFilesStr,
                                 clampUnit, countNumGndings, logFile);
  //find score of each new clause
  for (int i = 0; i < clauses.size(); i++)
  {
    string clause = clauses[i];
    cout << "    TRY_PRUNE1: " << clause << endl;
    double newScore = runLearnWts(mlnFile, predDecl, clause, uunitClauses,
                                  NULL, command, learnWtsExec, outMLNFile,
                                  dbFilesStr,clampUnit,countNumGndings,logFile);

    double diffScore = newScore - prevScore - candClauseLen * lenPenalty;
    cout << "Score = " << newScore << " - " << prevScore << " - "
         << candClauseLen*lenPenalty << " = " << diffScore << endl;

    if (diffScore <= 0.0) //if prune cand because change in score is too low
    {
      //read this to print out the weights; numPruned will be incremented
      //outCandStrs will not be changed
      bool prune = true;
      Array<double> absWts;
      readClauses(outCandStrs, numPruned, absWts, outMLNFile, 1, minWt, prune);
      prunedCandStrs.append(clause);
    }
    else
    { //check whether weights large enough
      StringHashArray tmpOutCandStrs;
      int tmpNumPruned = 0;
      Array<double> absWts;
      readClauses(tmpOutCandStrs,tmpNumPruned,absWts,outMLNFile,1,minWt,false);

      if (tmpNumPruned > 0) //if clause's wt is too small
      {
        assert(tmpNumPruned == 1);
        numPruned++;
        prunedCandStrs.append(clause);
      }
      else //if clause's wt >= minWt
      {
        outCandStrs.append(clause);
      }
    }//check whether clause's wt large enough
  }//for each new clause with a particular assignment of pos/neg literals

  delete [] command;
  for (int d = 0; d < dbFiles.size(); d++)
    unlink(dbFiles[d].c_str());
  unlink(mlnFile.c_str());
  unlink(outMLNFile.c_str());
  unlink(logFile.c_str());
}

int pruneCands(StringHashArray& outCandStrs,
               const int& minTuples, const int& maxTuples,
               const double& minWt,  const char* const& learnWtsExec,
               const double& L1c,    const Array<CandClause*>& cands,
               const string& tmpDir, const double& lenPenalty,
               Array<Array<Array<string>*> >& combIdToTuplesMapByDomain,
               StringToStringArrayMap& relToTypesMap, const bool& evalRules,
               const StringHashArray& unitClauses)
{
  //for each cand
  //  create the multiple databases
  //  create the mln file
  //  run learnwts
  //  get the clauses

  srandom(seed_); //for sampling tuples
  struct timeval tvA; struct timezone tzpA; //learnwts is external process
  gettimeofday(&tvA,&tzpA);

  StringHashArray prunedCandStrs;
  int numPruned = 0;
  int numIgnored = 0;
  for (int i = 0; i < cands.size(); i++)
  {
    struct timeval tv; struct timezone tzp; //learnwts is external process
    gettimeofday(&tv,&tzp);

    pruneCandsHelper(outCandStrs, prunedCandStrs, numPruned, numIgnored,
                     minTuples, maxTuples, minWt, learnWtsExec, L1c, cands[i],
                     tmpDir, lenPenalty, combIdToTuplesMapByDomain,
                     relToTypesMap, evalRules, unitClauses);

    struct timeval tv2; struct timezone tzp2;
    gettimeofday(&tv2,&tzp2);
    cout << "PROCESS CAND " << i+1 << " / " << cands.size() << " took ";
    Timer::printTime(cout, tv2.tv_sec-tv.tv_sec); cout <<  " : ";
    printCand(cout, cands[i]); cout << endl;

    cout << "pruneCands elapsed time = ";
    Timer::printTime(cout, tv2.tv_sec-tvA.tv_sec); cout << endl;
  }

  struct timeval tvB; struct timezone tzpB;
  gettimeofday(&tvB,&tzpB);
  cout << "PRUNE_CAND1 took "; Timer::printTime(cout, tvB.tv_sec-tvA.tv_sec);
  cout << endl;

  double total = numPruned + outCandStrs.size();
  cout << "1_NUM_IGNORED   " << numIgnored           << " / " << total << endl;
  cout << "1_NUM_PRUNED    " << numPruned-numIgnored << " / " << total << endl;
  cout << "1_TOTAL_PRUNED  " << numPruned            << " / " << total << endl;
  cout << "1_NUM_REMAINING " << outCandStrs.size()   << " / " << total << endl;
  cout << endl;

  for (int i = 0; i < outCandStrs.size(); i++)
    cout << "P1CANDS " << i << ": " << outCandStrs[i] << endl;
  cout << endl;

  return numPruned;
}

void createCommand2(char* const& command, const char* const& learnWtsExec,
                    const string& mlnFile, const string& outMLNFile,
                    const string& dbFiles, const bool& sampleAtoms,
                    const double& fractAtoms, const string& logFile,
                    const int& numDBs, const bool& clampUnit)
{
  string noEqualPredWtStr = NO_EQ_PRED_WT ? "-gNoEqualPredWt" : "";
  string clampUnitStr = "";
  string clampDev     = "";
  if (clampUnit)
  {
    clampUnitStr =  "-clampUnit";
    clampDev     =  "-clampDev " + Util::doubleToString(CLAMP_DEV);
  }
  sprintf(command,"%s -g -i %s -o %s -t %s %s%s%s %s %s %s %s %s > %s",
          learnWtsExec, mlnFile.c_str(), outMLNFile.c_str(), dbFiles.c_str(),
          (sampleAtoms?" -sampleAtoms":""), (sampleAtoms?" -fractAtoms ":""),
          (sampleAtoms?(Util::doubleToString(fractAtoms).c_str()):""),
          ((numDBs>1)?"-multipleDatabases":""), CONV_THRESH.c_str(),
          noEqualPredWtStr.c_str(), clampUnitStr.c_str(),
          clampDev.c_str(), logFile.c_str());
  cout << "command: " << command << endl;
}

void getNumTuplesInDbs(Array<double>& dbNumTuples,
                       const Array<string>& dbFilesArr,
                       const string& tmpDir, const string& predDecl,
                       const char* const& learnWtsExec)
{
  string dbFilesStr = dbFilesArr[0];
  for (int i = 1; i < dbFilesArr.size(); i++)
    dbFilesStr += "," + dbFilesArr[i];

  string mlnFile    = tmpDir + TMP_MLN_FILE_NAME     + ".mln";
  string outMLNFile = tmpDir + TMP_OUT_MLN_FILE_NAME + ".mln";
  string logFile    = tmpDir + TMP_LOG_FILE;

  char* command = new char[2000 + dbFilesStr.length()];

  //find score due to unit clauses
  StringHashArray noop;
  bool countNumGndings = true;
  runLearnWts(mlnFile, predDecl, "", noop, NULL, command, learnWtsExec,
              outMLNFile, dbFilesStr, false, countNumGndings, logFile);

  dbNumTuples.clear();
  ifstream in(logFile.c_str());
  Util::assertt(!in.fail(), "failed to open", logFile, -1);
  string buf;
  while(getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.find("numTotalGndings_ in domain ") == 0)
    {
      istringstream iss(buf);
      string tmp; int domainIdx; double numTuples;
      iss >> tmp >> tmp >> tmp >> domainIdx >> tmp >> numTuples;
      dbNumTuples.append(numTuples);
    }
  }

  for(int i = 0; i < dbNumTuples.size(); i++)
    cout << "NUM_TUPLES_DOMAIN_" << i << "  " << dbNumTuples[i] << endl;

  in.close();
  unlink(mlnFile.c_str());
  unlink(outMLNFile.c_str());
  unlink(logFile.c_str());
}

void getTrainingDBs(string& dbFiles, int& numDBs, bool& ssampleAtoms,
                    double& ffractAtoms, const Array<string>& dbFilesArr,
                    const bool& sampleAtoms, const double& fractAtoms,
                    const string& tmpDir, const string& predDecl,
                    const char* const& learnWtsExec)
{
  dbFiles      = "";
  numDBs       = 0;
  ssampleAtoms = false;
  ffractAtoms  = 0.0;

  if (sampleAtoms)
  {
    if (dbFilesArr.size() > 1) //if multiple DBs
    {
      Array<double> dbNumTuples;
      getNumTuplesInDbs(dbNumTuples, dbFilesArr, tmpDir, predDecl,learnWtsExec);

      double totalNumTuples = 0.0;
      for (int i = 0; i < dbNumTuples.size(); i++)
        totalNumTuples += dbNumTuples[i];

      //then sample by databases
      Array<int> idxs;
      for (int j = 0; j < dbFilesArr.size(); j++) idxs.append(j);
      double numTuplesSelected = 0.0;
      while (!idxs.empty() && numTuplesSelected < fractAtoms * totalNumTuples)
      {
        int pos = random() % idxs.size();
        int dbIdx = idxs[ pos ];
        idxs.removeItemFastDisorder(pos);
        if (!dbFiles.empty()) dbFiles += ",";
        dbFiles += dbFilesArr[dbIdx];
        cout << "sampled DB: " << dbFilesArr[dbIdx] << endl;
        numTuplesSelected += dbNumTuples[dbIdx];
        numDBs++;
      }
      ssampleAtoms = false;
      ffractAtoms  = 0.0;
    }
    else
    { //sampling atoms but only one DB, so sample atoms in DB
      dbFiles = dbFilesArr[0];
      numDBs = 1;
      ssampleAtoms = sampleAtoms;
      ffractAtoms  = fractAtoms;
    }
  }
  else
  {  //not sampling atoms
    numDBs = dbFilesArr.size();
    dbFiles = dbFilesArr[0];
    for (int j = 1; j < dbFilesArr.size(); j++)
      dbFiles += "," + dbFilesArr[j];
    ssampleAtoms = false;
    ffractAtoms  = 0.0;
  }
}

double runLearnWts2(const string& mlnFile, const string& predDecl,
                    const string& clauseStr, const StringHashArray& unitClauses,
                    const Array<double>*const& unitWts, char* const& command,
                    const char* const& learnWtsExec, const string& outMLNFile,
                    const string& dbFiles, const bool& ssampleAtoms,
                    const double& ffractAtoms, const string& logFile,
                    const int& numDBs, const bool& clampUnit)
{
  createMLNFile(mlnFile, predDecl, clauseStr, unitClauses, unitWts);
  createCommand2(command, learnWtsExec, mlnFile, outMLNFile, dbFiles,
                 ssampleAtoms, ffractAtoms, logFile, numDBs, clampUnit);
  system(command);
  double logLike = readScore(logFile);
  return logLike;
}

bool lowerLoglikeThanSubClauses(const string& cand, const double& diffLoglike,
                                Array<double>& diffLoglikes,
                                const StringHashArray& candStrs,
                                const string& mlnFile, const string& predDecl,
                                const StringHashArray& unitClauses,
                                const Array<double>*const& unitWts,
                                char* const& command,
                                const char* const& learnWtsExec,
                                const string& outMLNFile, const string& dbFiles,
                                const bool& ssampleAtoms,
                                const double& ffractAtoms, const string& logFile,
                                const int& numDBs, const double& lenPenalty,
                                const double& unitScore)
{
  int candLen = clauseLen(cand);

  Array<string> subClauses;
  createSubClauses(subClauses, cand);

  for (int k = 0; k < subClauses.size(); k++)
  {
    string subClause = subClauses[k];
    int    subLen    = clauseLen(subClause);    assert(subLen < candLen);
    cout << "\tcomparing against subClause: " << subClause << " , ";

    int idx = candStrs.find( subClause );
    if (idx < 0)
    {
      cout << "ignored because doesn't exist" << endl;
      continue; //ignore subclause since it is not created
    }
    double subDiffLoglike;
    if (diffLoglikes[idx] != DBL_MAX)  subDiffLoglike = diffLoglikes[idx];
    else
    { //loglike of subclause not computed
      bool clampUnit = false;
      double subScore = runLearnWts2(mlnFile, predDecl, subClause, unitClauses,
                                     unitWts, command, learnWtsExec, outMLNFile,
                                     dbFiles, ssampleAtoms, ffractAtoms,
                                     logFile, numDBs, clampUnit);
      diffLoglikes[idx] = subScore - unitScore;
      subDiffLoglike = diffLoglikes[idx];
    }

    int    diffLen   = candLen - subLen;
    double diffScore = diffLoglike - subDiffLoglike - diffLen*lenPenalty;
    cout << "diffLogLike = "    << diffLoglike << " , "
         << "subDiffLogLike = " << subDiffLoglike << " , "
         << "diffScore = "      <<  diffScore << endl;

    if (diffScore <= 0.0)
    {
      cout << "LESS_THAN_SUB: " << cand
           << " diffLoglike - subDiffLoglike - diffLen*lenPenalty = "
           << diffLoglike  << " - " << subDiffLoglike << " - "
           << diffLen*lenPenalty << " <= 0 of " << subClause << endl;
      return true;
    }
  }
  return false;
}

//need to run learnwts
void pruneCands2Helper(const int& candStrIdx, const StringHashArray& candStrs,
                       Array<double>& diffLoglikes, Array<double>& absWts,
                       StringHashArray& outCandStrs,
                       int& numPruned, const string& predDecl,
                       const string& mlnFile, const string& outMLNFile,
                       const string& logFile, char* const& command,
                       const StringHashArray& unitClauses,
                       const Array<double>*const& unitWts, const string& dbFiles,
                       const int& numDBs, const bool& ssampleAtoms,
                       const double& ffractAtoms, const double& minWt,
                       const char* const& learnWtsExec,const double& lenPenalty,
                       const double& unitScore)
{
  string candStr = candStrs[ candStrIdx ];
  int    candLen = clauseLen(candStr);

  double newScore, diffScore, absWt;
  bool clampUnit = false;

  if (diffLoglikes[ candStrIdx ] == DBL_MAX) //if value have not been computed
  {
    newScore = runLearnWts2(mlnFile, predDecl, candStr, unitClauses, unitWts,
                            command, learnWtsExec, outMLNFile, dbFiles,
                            ssampleAtoms,ffractAtoms,logFile,numDBs,clampUnit);
    diffLoglikes[ candStrIdx ] = newScore - unitScore;
    diffScore = newScore - unitScore - candLen * lenPenalty;

    int npruned = 0;
    Array<double> tmpAbsWts;
    StringHashArray outCands;
    readClauses(outCands, npruned, tmpAbsWts, outMLNFile, 1, minWt, false);
    assert(tmpAbsWts.size() == 1);
    absWt = tmpAbsWts[0];
    absWts[ candStrIdx ] = absWt;

    if (absWt < minWt) assert(outCands.empty() && npruned == 1);
  }
  else
  {
    newScore  = diffLoglikes[ candStrIdx ] + unitScore;
    diffScore = diffLoglikes[ candStrIdx ] - candLen * lenPenalty;
    absWt     = absWts[ candStrIdx ];
  }

  cout << "diffLogLike = " << diffLoglikes[ candStrIdx ] << endl;
  cout << "DiffScore = " << newScore << " - " << unitScore << " - "
       << candLen*lenPenalty << " = " << diffScore << " ; absWt = " << absWt
       << endl;

  if (diffScore > 0.0 && absWt >= minWt) //if better than unit clauses
  {
    bool lowerll = lowerLoglikeThanSubClauses(candStr, diffLoglikes[candStrIdx],
                                              diffLoglikes, candStrs, mlnFile,
                                              predDecl, unitClauses, unitWts,
                                              command, learnWtsExec, outMLNFile,
                                              dbFiles, ssampleAtoms, ffractAtoms,
                                              logFile, numDBs, lenPenalty,
                                              unitScore);
    if (!lowerll)  outCandStrs.append(candStr);
    else           { cout << "PRUNE2: " << candStr << endl;  numPruned++; }
  }
  else
  {
    cout <<"PRUNE2: " << candStr << endl;
    numPruned++;
  }
}


//second round of pruning
void pruneCands2(StringHashArray& outCandStrs, const int& prevPruned,
                 const StringHashArray& predDeclarations,
                 const bool& sampleAtoms, const double& fractAtoms,
                 string dbFiles, const double& minWt,
                 const char* const& learnWtsExec, const double& L1c,
                 const string& tmpDir, const double& lenPenalty,
                 StringToStringArrayMap& relToTypesMap,
                 const StringHashArray& unitClauses)
{
  struct timeval tvA; struct timezone tzpA;
  gettimeofday(&tvA,&tzpA);

  for (unsigned int i = 0; i < dbFiles.length(); i++)
    if (dbFiles.at(i) == ',') dbFiles.at(i) = ' ';
  dbFiles.append(" ::");

  Array<string> dbFilesArr;
  istringstream iss(dbFiles);
  string buf; iss >> buf;
  while (buf.compare("::") != 0) { dbFilesArr.append(buf); iss >> buf; }

  string predDecl = "//predicate declarations\n";
  for (int j = 0; j < predDeclarations.size(); j++)
    predDecl += predDeclarations[j] + "\n";

  srandom(seed_); //for sampling DBs
  string chosenDBFiles; int numDBs; bool ssampleAtoms; double ffractAtoms;
  getTrainingDBs(chosenDBFiles, numDBs, ssampleAtoms, ffractAtoms, dbFilesArr,
                 sampleAtoms, fractAtoms, tmpDir, predDecl, learnWtsExec);

  StringHashArray candStrs( outCandStrs.size() );
  for (int i = 0; i < outCandStrs.size(); i++)
    candStrs.append( outCandStrs[i] );
  outCandStrs.clear();
  Array<double> diffLoglikes; //with respect to MLN with unitClausees
  Array<double> absWts;
  diffLoglikes.growToSize(candStrs.size(), DBL_MAX);
  absWts.growToSize(candStrs.size(), -1);


  string mlnFile    = tmpDir + TMP_MLN_FILE_NAME     + ".mln";
  string outMLNFile = tmpDir + TMP_OUT_MLN_FILE_NAME + ".mln";
  string logFile    = tmpDir + TMP_LOG_FILE;
  int    numPruned  = 0;

  char* command = new char[2000 + dbFilesArr[0].length()];

  //Array<double> unitWts;
  //findUnitWts(unitWts, mlnFile, predDecl, unitClauses, command, learnWtsExec,
  //            outMLNFile, chosenDBFiles, logFile);
  //for (int i = 0; i < unitWts.size(); i++)
  //  cout << "UNIT_WTS " << i << " : " << unitWts[i] << " "
  //       << unitClauses[i] << endl;

  //compute score due to unit clauses
  bool clampUnit = false;
  double unitScore = runLearnWts2(mlnFile, predDecl, "", unitClauses,
                                  NULL, command, learnWtsExec, outMLNFile,
                                  chosenDBFiles, ssampleAtoms, ffractAtoms,
                                  logFile, numDBs, clampUnit);

  for (int i = 1; i < candStrs.size(); i++)
    Util::assertt(clauseLen(candStrs[i-1]) <= clauseLen(candStrs[i]),
                  "expect clauses to be sorted shortest to longest", -1);

  for (int i = 0; i < candStrs.size(); i++)
  {
    cout << "    TRY_PRUNE2: " << candStrs[i] << endl;
    struct timeval tv; struct timezone tzp; //learnwts is external process
    gettimeofday(&tv,&tzp);

    pruneCands2Helper(i, candStrs, diffLoglikes, absWts, outCandStrs, numPruned,
                      predDecl, mlnFile, outMLNFile, logFile, command,
                      unitClauses, NULL, chosenDBFiles, numDBs, ssampleAtoms,
                      ffractAtoms, minWt, learnWtsExec, lenPenalty, unitScore);

    struct timeval tv2; struct timezone tzp2;
    gettimeofday(&tv2,&tzp2);
    cout << "PRUNE CANDSTR " << i+1 << " / " << candStrs.size() << " took ";
    Timer::printTime(cout, tv2.tv_sec-tv.tv_sec); cout <<  " : ";
    cout << candStrs[i] << endl;

    cout << "pruneCands2 elapsed time = ";
    Timer::printTime(cout, tv2.tv_sec-tvA.tv_sec); cout << endl;
  }

  delete [] command;
  unlink(mlnFile.c_str());
  unlink(outMLNFile.c_str());
  unlink(logFile.c_str());

  struct timeval tvB; struct timezone tzpB;
  gettimeofday(&tvB,&tzpB);
  cout << "PRUNE_CAND2 took "; Timer::printTime(cout, tvB.tv_sec-tvA.tv_sec);
  cout << endl;

  int total = numPruned + prevPruned + outCandStrs.size();
  cout << "2_NUM_PRUNED    " << numPruned            << " / " << total << endl;
  cout << "2_TOTAL_PRUNED  " << numPruned+prevPruned << " / " << total << endl;
  cout << "2_NUM_REMAINING " << outCandStrs.size()   << " / " << total << endl;
  cout << endl;
}

void writeMLN(const char* const& outMLNFile, const StringHashArray& outCandStrs,
              const StringHashArray& predDeclarations,
              StringToStringArrayMap& relToTypesMap,
              const StringHashArray& unitClauses)
{
  ofstream out(outMLNFile);
  Util::assertt(!out.fail(), "failed to open out mln file", outMLNFile, -1);

  //write the predicate declarations
  out << "//predicate declarations" << endl;
  for (int i = 0; i < predDeclarations.size(); i++)
    out << predDeclarations[i] << endl;
  out << endl;

  //write the clauses
  for (int i = 0; i < outCandStrs.size(); i++)
    out << outCandStrs[i] << endl;

  for (int i = 0; i < unitClauses.size(); i++)
    out << unitClauses[i] << endl;

  out.close();
}

void checkFiles(string inCandFiles, string semnetFiles, string dbFiles)
{
  int num = 0;
  for (unsigned int i = 0; i < inCandFiles.length(); i++)
    if (inCandFiles.at(i) == ',') num++;

  int num2 = 0;
  for (unsigned int i = 0; i < semnetFiles.length(); i++)
    if (semnetFiles.at(i) == ',') num2++;

  int num3 = 0;
  for (unsigned int i = 0; i < dbFiles.length(); i++)
    if (dbFiles.at(i) == ',') num3++;

  Util::assertt(num == num2 || num == num3, "expect same num of inCandFiles, semnetFiles and dbFiles", -1);
}

void pruneCands3(StringHashArray& cands, Array<double>& loglikes,
                 const double& lenPenalty, const int& prevPruned)
{
  assert(cands.size() == loglikes.size());

  int origNumCands = cands.size();
  Timer timer;
  double bsec = timer.time();

  int numPruned = 0;

  //check that candidates are sorted in increasing order of length
  //and arrange the cands by length
  Array<StringHashArray*> candsByLen;
  Array<Array<double>*> loglikesByLen;
  for (int i = 0; i < cands.size(); i++)
  {
    string cand      = cands[i];
    int    len       = clauseLen(cand);
    double loglike   = loglikes[i];

    if (i < cands.size()-1 && len > clauseLen(cands[i+1]))
      Util::exit("expect cands to sorted in increasing order of len", cand, -1);

    if (len >= candsByLen.size())
    {
      candsByLen.growToSize(len+1,NULL);
      loglikesByLen.growToSize(len+1,NULL);
      assert(candsByLen.size() == loglikesByLen.size());
    }
    if (candsByLen[len] == NULL)
    {
      assert(loglikesByLen[len] == NULL);
      candsByLen[len] = new StringHashArray;
      loglikesByLen[len] = new Array<double>;
    }
    int idx  = candsByLen[len]->append(cand);  assert(idx >= 0);
    int idx2 = loglikesByLen[len]->append(loglike);  assert(idx2 >= 0);
    assert(idx == idx2);
  }

  cands.clear();
  loglikes.clear();

  for (int i = 1; i < candsByLen.size(); i++)
  {
    if (candsByLen[i] == NULL) continue;
    StringHashArray& ccands    = *( candsByLen[i] );
    Array<double>&   lloglikes = *( loglikesByLen[i] );
    assert(ccands.size() == lloglikes.size());

    for (int j = 0; j < ccands.size(); j++) //for each cand
    {
      string ccand      = ccands[j];
      double lloglike   = lloglikes[j];
      int    cclauseLen = clauseLen(ccand);

      Array<string> shorterCands;
      //Array<string> others;
      //if (cclauseLen >= 3) createLenMinus1Cands(shorterCands, others, ccand);
      //if (cclauseLen >= 4) createLenMinus2Cands(shorterCands, others, ccand);
      createSubClauses(shorterCands, ccand);

      //check whether cand is better than sub-clauses
      bool smallerLogLikeThanPrev = false;
      for (int k = 0; k < shorterCands.size(); k++)
      {
        string shorterCand = shorterCands[k];
        int    shorterLen  = clauseLen(shorterCand);
        assert(shorterLen<cclauseLen);
        cout << "\tcomparing against shorter cand: " << shorterCand << endl;

        if (candsByLen[ shorterLen ])
        {
          StringHashArray& prevCands    = *( candsByLen[ shorterLen ] );
          Array<double>&   prevLoglikes = *( loglikesByLen[ shorterLen ] );
          assert(prevCands.size() == prevLoglikes.size());

          int idx = prevCands.find(shorterCand);
          if (idx < 0) continue;
          double prevLoglike = prevLoglikes[idx];
          double diffLen = cclauseLen - shorterLen;
          double diffLoglike = lloglike - prevLoglike - diffLen*lenPenalty;

          if (diffLoglike <= 0.0)
          {
            cout << "PRUNE3 " << ccand << " loglike - prevLoglike - diffLen*lenPenalty = "
                 << lloglike  << " - " << prevLoglike << " - " << diffLen*lenPenalty
                 << " <= 0 of " << shorterCand << endl;
            smallerLogLikeThanPrev = true;
            break;
          }
        }
      }

      if (!smallerLogLikeThanPrev)  cands.append(ccand);
      else                          numPruned++;
    }
  }

  candsByLen.deleteItemsAndClear();
  loglikesByLen.deleteItemsAndClear();

  assert(numPruned + cands.size() == origNumCands);

  cout << "PRUNE_CAND3 took "; Timer::printTime(cout, timer.time()-bsec);
  cout << endl;

  int total = numPruned + prevPruned + cands.size();
  cout << "3_NUM_PRUNED    " << numPruned            << " / " << total << endl;
  cout << "3_TOTAL_PRUNED  " << numPruned+prevPruned << " / " << total << endl;
  cout << "3_NUM_REMAINING " << cands.size()         << " / " << total << endl;
  cout << endl << endl;
}

void readCandStrs(StringHashArray& outCandStrs, const char* candStrsFile)
{
  ifstream in(candStrsFile); assert(!in.fail());
  string buf;
  while(getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    outCandStrs.append(buf);
  }
  in.close();
}

int main(int argc, char* argv[])
{
  if (argc != 19+1)
  {
    cout << "usage: " << argv[0] << " <inCandFiles> <infoFile> <semnetFiles> "
         << "<outMLNFile> <tmpDir> <learnWtsExec> <minWt> <minTuplesPerPredPerDomain> "
         << "<maxTuples> <sampleAtoms> <fractAtoms> <dbFiles> <minWt2> <lenPenalty1> "
         << "<pruneRules1> <candStrsFile> <evalRules1> <lenPenalty2> <seed>" << endl;
    exit(-1);
  }

  const char*  inCandFiles  = argv[1];
  const char*  infoFile     = argv[2];
  const char*  semnetFiles  = argv[3];
  const char*  outMLNFile   = argv[4];
  const char*  ttmpDir      = argv[5];
  const char*  learnWtsExec = argv[6];
  const double minWt        = atof( argv[7] );
  const int    minTuples    = atoi( argv[8] ); //minTuples PER PRED PER DOMAIN
  const int    maxTuples    = atoi( argv[9] );
  const bool   sampleAtoms  = (strcmp(argv[10],"true")==0);
  const double fractAtoms   = atof( argv[11] );
  const char*  dbFiles      = argv[12];
  const double minWt2       = atof( argv[13] );
  const double lenPenalty1  = atof( argv[14] );
  const bool   pruneRules1  = (strcmp(argv[15],"true")==0);
  const char*  candStrsFile = argv[16];
  const bool   evalRules    = (strcmp(argv[17],"true")==0);
  const double lenPenalty2  = atof( argv[18] );
  const int    seed         = atoi( argv[19] );
  const double L1c          = 0.0;
  seed_ = seed;

  cout << "----------------------------------" << endl;
  cout << "inCandFile   " << inCandFiles  << endl;
  cout << "infoFile     " << infoFile     << endl;
  cout << "semnetFile   " << semnetFiles  << endl;
  cout << "outMLNFile   " << outMLNFile   << endl;
  cout << "tmpDir       " << ttmpDir       << endl;
  cout << "learnWtsExec " << learnWtsExec << endl;
  cout << "minWt        " << minWt        << endl;
  cout << "minTuples    " << minTuples    << endl;
  cout << "maxTuples    " << maxTuples    << endl;
  cout << "sampleAtoms  " << (sampleAtoms?"true":"false") << endl;
  cout << "fractAtoms   " << fractAtoms   << endl;
  cout << "dbFile       " << dbFiles      << endl;
  cout << "minWt2       " << minWt2       << endl;
  cout << "lenPenalty1  " << lenPenalty1  << endl;
  cout << "pruneRules1  " << (pruneRules1?"true":"false") << endl;
  cout << "candStrsFile " << candStrsFile << endl;
  cout << "lenPenalty2  " << lenPenalty2  << endl;
  cout << "seed         " << seed         << endl;
  //cout << "L1c          " << L1c          << endl;
  cout << "evalRules    " << (evalRules?"true":"false")   << endl;
  cout << "----------------------------------" << endl;

  checkFiles(inCandFiles, semnetFiles, dbFiles);
  if (sampleAtoms) {Util::assertt(fractAtoms>0.0,"expect fractAtoms > 0.0",-1);}

  string tmpDir = ttmpDir;
  if (tmpDir.at(tmpDir.length()-1) != '/') tmpDir += '/';

  //using wall clock time because weight learning is a separate process called
  //via system()
  struct timeval tv; struct timezone tzp;
  gettimeofday(&tv,&tzp);

  cout << "reading info file..." << endl;
  StringToStringArrayMap relToTypesMap;
  readInfoFile(infoFile, relToTypesMap);

  StringHashArray predDeclarations;
  readPredDeclarations(predDeclarations, infoFile);

  StringHashArray unitClauses;
  createUnitClauses(unitClauses, relToTypesMap);

  Array<Array<Array<string>*> > combIdToTuplesMapByDomain;
  StringHashArray outCandStrs;
  int numPruned = 0;

  if (pruneRules1)
  {
    cout << "reading semantic network..." << endl;
    readSemanticNetwork(semnetFiles, combIdToTuplesMapByDomain);

    cout << "reading candidates..." << endl;
    Array<CandClause*> cands;
    readCands(inCandFiles, cands);

    cout << "pruning candidates..." << endl;
    numPruned = pruneCands(outCandStrs, minTuples, maxTuples, minWt,
                           learnWtsExec, L1c, cands, tmpDir, lenPenalty1,
                           combIdToTuplesMapByDomain, relToTypesMap,
                           evalRules, unitClauses);
  }
  else
  {
    readCandStrs(outCandStrs, candStrsFile);
    cout << "#candStrs read = " << outCandStrs.size() << endl;
  }

  Array<double> loglikes;
  pruneCands2(outCandStrs, numPruned, predDeclarations,
              sampleAtoms, fractAtoms, dbFiles, minWt2, learnWtsExec,
              L1c, tmpDir, lenPenalty2, relToTypesMap, unitClauses);

  //commented out: check improvement over shorter clauses in pruneCands
  //if use this method, must change it so that clauses to be compared are
  //trained on the same sampled data, and add lenMinus2 cands
  //pruneCands3(outCandStrs,  loglikes, lenPenalty, numPruned);

  cout << "creating MLN file..." << endl;
  writeMLN(outMLNFile, outCandStrs, predDeclarations, relToTypesMap, unitClauses);

  struct timeval tv2; struct timezone tzp2;
  gettimeofday(&tv2,&tzp2);
  cout << "PRUNE_RULES took "; Timer::printTime(cout, tv2.tv_sec-tv.tv_sec);
  cout << endl;

  for (int d = 0; d < combIdToTuplesMapByDomain.size(); d++)
    combIdToTuplesMapByDomain[d].deleteItemsAndClear();
}
